/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AvatarManipulationWidget.h"

#include "Debug.h"

#include <QPainter>
#include <cmath>

const double precision = 5;
const double pointerPrecision = 10;
const double minSize = 30;

//________________________________________________________________________
template<typename T> T square( T x ) { return x*x; }

//________________________________________________________________________
AvatarManipulationWidget::AvatarManipulationWidget( QWidget* parent, const QPixmap& source ):
    QWidget( parent ),
    Counter( "Private::AvatarManipulationWidget" ),
    source_( source )
{

    setMouseTracking( true );
    setMinimumSize( minSize, minSize );
    _updateImageRect();
    resetSelection();

}

//________________________________________________________________________
QSize AvatarManipulationWidget::sizeHint( void ) const
{
    return source_.isNull() ?
        QWidget::sizeHint():
        QSize( qMin( source_.width(), 512 ), qMin( source_.height(), 512 ) );
}

//________________________________________________________________________
QPixmap AvatarManipulationWidget::pixmap( int size ) const
{

    QPixmap pixmap( size, size );
    pixmap.fill( Qt::transparent );

    QPainter painter( &pixmap );
    painter.setRenderHints(QPainter::SmoothPixmapTransform|QPainter::Antialiasing);
    painter.setWindow( selectionRect_.toRect() );

    painter.setPen( Qt::NoPen );
    painter.setBrush( Qt::black );
    painter.drawEllipse( selectionRect_ );

    painter.setCompositionMode( QPainter::CompositionMode_SourceIn );
    painter.drawPixmap( source_.rect(), source_ );

    return pixmap;

}

//________________________________________________________________________
void AvatarManipulationWidget::resetSelection( void )
{
    selectionRect_ = source_.rect();
    if( selectionRect_.height() < selectionRect_.width() ) selectionRect_.setWidth( selectionRect_.height() );
    else if( selectionRect_.width() < selectionRect_.height() ) selectionRect_.setHeight( selectionRect_.width() );

    selectionRect_.moveCenter( QRectF(source_.rect()).center() );
    update();
}

//________________________________________________________________________
void AvatarManipulationWidget::resizeEvent( QResizeEvent* event )
{
    _updateImageRect();
    return QWidget::resizeEvent( event );
}

//________________________________________________________________________
void AvatarManipulationWidget::paintEvent( QPaintEvent* event )
{

    QPainter painter( this );
    painter.setClipRegion( event->region() );
    painter.setRenderHints(QPainter::SmoothPixmapTransform);
    painter.drawPixmap( imageRect_, source_ );

    // create mask
    QPixmap pixmap( imageRect_.size() );
    {
        pixmap.fill( QColor( 0, 0, 0, 100 ) );

        QPainter painter( &pixmap );
        painter.setRenderHints( QPainter::Antialiasing );
        painter.setCompositionMode( QPainter::CompositionMode_DestinationOut );
        painter.setBrush( Qt::black );
        painter.setPen( Qt::NoPen );

        const double scale = double( imageRect_.width() ) / source_.width();
        const QRectF selectionRect( selectionRect_.topLeft()*scale, selectionRect_.bottomRight()*scale );
        painter.drawEllipse( selectionRect );

        painter.setCompositionMode( QPainter::CompositionMode_SourceOver );
        painter.setBrush( Qt::white );
        painter.setPen( Qt::black );

        // draw corners
        auto innerRect( selectionRect.adjusted( precision, precision, -precision, -precision ) );
        QVector<QPointF> corners =
        {
            innerRect.topLeft(),
            innerRect.topRight(),
            innerRect.bottomLeft(),
            innerRect.bottomRight()
        };

        QRectF rect( 0, 0, 2*precision, 2*precision );
        for( const auto& corner:corners )
        {
            rect.moveCenter( corner );
            painter.drawRect( rect.adjusted( 0.5, 0.5, -0.5, -0.5 ) );
        }

    }

    painter.drawPixmap( imageRect_, pixmap );

}

//________________________________________________________________________
void AvatarManipulationWidget::mousePressEvent( QMouseEvent* event )
{

    if( dragMode_ == DragNone && event->button() == Qt::LeftButton )
    {
        dragMode_ = _dragMode( event->pos() );
        switch( dragMode_ )
        {

            case DragTopLeft:
            dragOffset_ = selectionRect_.topLeft() - _toImageCoordinates( event->pos() );
            break;

            case DragTopRight:
            dragOffset_ = selectionRect_.topRight() - _toImageCoordinates( event->pos() );
            break;

            case DragBottomLeft:
            dragOffset_ = selectionRect_.bottomLeft() - _toImageCoordinates( event->pos() );
            break;

            case DragBottomRight:
            dragOffset_ = selectionRect_.bottomRight() - _toImageCoordinates( event->pos() );
            break;

            case DragCenter:
            dragOffset_ = _toImageCoordinates( event->pos() );
            break;

            default:
            break;

        }

    }

}

//________________________________________________________________________
void AvatarManipulationWidget::mouseMoveEvent( QMouseEvent* event )
{
    if( dragMode_ == DragNone && !event->buttons() )
    {

        auto dragMode( _dragMode( event->pos() ) );
        switch( dragMode )
        {
            case DragTopLeft:
            case DragBottomRight:
            setCursor( Qt::SizeFDiagCursor );
            break;

            case DragTopRight:
            case DragBottomLeft:
            setCursor( Qt::SizeBDiagCursor );
            break;

            case DragCenter:
            setCursor( Qt::SizeAllCursor );
            break;

            default:
            setCursor( Qt::ArrowCursor );
            break;
        }

    }

    // color calculation delta function
    auto bind = [](qreal& value, qreal min, qreal max )
    {
        if( value < min ) value = min;
        if( value > max ) value = max;
        return value;
    };

    if( event->buttons() & Qt::LeftButton )
    {

        const QRectF sourceRect( source_.rect() );
        switch( dragMode_ )
        {
            default:
            case DragNone: break;

            case DragTopLeft:
            {
                auto dragPoint( _toImageCoordinates( event->pos() ) + dragOffset_ );
                bind( dragPoint.rx(), sourceRect.left(), selectionRect_.right() - minSize );
                bind( dragPoint.ry(), sourceRect.top(), selectionRect_.bottom() - minSize );

                auto offset = selectionRect_.bottomRight() - dragPoint;
                offset.setX( qMin( offset.x(), offset.y() ) );
                offset.setY( qMin( offset.x(), offset.y() ) );
                selectionRect_.setTopLeft( selectionRect_.bottomRight()-offset );
                update();
                break;
            }

            case DragTopRight:
            {
                auto dragPoint( _toImageCoordinates( event->pos() ) + dragOffset_ );
                bind( dragPoint.rx(), selectionRect_.left() + minSize, sourceRect.right() );
                bind( dragPoint.ry(), sourceRect.top(), selectionRect_.bottom() - minSize );

                auto offset = selectionRect_.bottomLeft() - dragPoint;
                offset.setX( - qMin( -offset.x(), offset.y() ) );
                offset.setY( qMin( -offset.x(), offset.y() ) );
                selectionRect_.setTopRight( selectionRect_.bottomLeft()-offset );
                update();
                break;
            }

            case DragBottomLeft:
            {
                auto dragPoint( _toImageCoordinates( event->pos() ) + dragOffset_ );
                bind( dragPoint.rx(), sourceRect.left(), selectionRect_.right() - minSize );
                bind( dragPoint.ry(), selectionRect_.top() + minSize, sourceRect.bottom() );

                auto offset = selectionRect_.topRight() - dragPoint;
                offset.setX( qMin( offset.x(), -offset.y() ) );
                offset.setY( -qMin( offset.x(), -offset.y() ) );
                selectionRect_.setBottomLeft( selectionRect_.topRight() - offset );
                update();
                break;
            }

            case DragBottomRight:
            {
                auto dragPoint( _toImageCoordinates( event->pos() ) + dragOffset_ );
                bind( dragPoint.rx(), selectionRect_.left() + minSize, sourceRect.right() );
                bind( dragPoint.ry(), selectionRect_.top() + minSize, sourceRect.bottom() );

                auto offset = selectionRect_.topLeft() - dragPoint;
                offset.setX( qMax( offset.x(), offset.y() ) );
                offset.setY( qMax( offset.x(), offset.y() ) );
                selectionRect_.setBottomRight( selectionRect_.topLeft() - offset );
                update();
                break;
            }

            case DragCenter:
            {
                auto dragPoint( _toImageCoordinates( event->pos() ) );
                bind( dragPoint.rx(), dragOffset_.x() + sourceRect.left() - selectionRect_.left(), dragOffset_.x() + sourceRect.right() - selectionRect_.right() );
                bind( dragPoint.ry(), dragOffset_.y() + sourceRect.top() - selectionRect_.top(), dragOffset_.y() + sourceRect.bottom() - selectionRect_.bottom() );

                selectionRect_.moveCenter( selectionRect_.center() + dragPoint - dragOffset_ );
                dragOffset_ = dragPoint;
                update();
                break;
            }

        }

    }

}

//________________________________________________________________________
void AvatarManipulationWidget::mouseReleaseEvent( QMouseEvent* event )
{ dragMode_ = DragNone; }

//________________________________________________________________________
AvatarManipulationWidget::DragMode AvatarManipulationWidget::_dragMode( QPoint position ) const
{

    const QRectF sourceRect( _toWidgetCoordinates( selectionRect_.topLeft() ), _toWidgetCoordinates( selectionRect_.bottomRight() ) );
    const QRectF innerRect( sourceRect.adjusted( precision, precision, -precision, -precision ) );

    if( (position - innerRect.topLeft()).manhattanLength() < pointerPrecision ) return DragTopLeft;
    else if( (position - innerRect.topRight()).manhattanLength() < pointerPrecision ) return DragTopRight;
    else if( (position - innerRect.bottomLeft()).manhattanLength() < pointerPrecision ) return DragBottomLeft;
    else if( (position - innerRect.bottomRight()).manhattanLength() < pointerPrecision ) return DragBottomRight;
    else if( sourceRect.contains( position ) ) return DragCenter;
    else return DragNone;

}

//________________________________________________________________________
QPointF AvatarManipulationWidget::_toWidgetCoordinates( QPointF source ) const
{
    const double scale = double( imageRect_.width() ) / source_.width();
    return source*scale + imageRect_.topLeft();
}

//________________________________________________________________________
QPointF AvatarManipulationWidget::_toImageCoordinates( QPointF source ) const
{
    const double scale = double( imageRect_.width() ) / source_.width();
    return (source - imageRect_.topLeft()) / scale;
}

//________________________________________________________________________
void AvatarManipulationWidget::_updateImageRect( void )
{
    auto rect( this->rect() );
    auto imageRect( source_.rect() );

    imageRect_ = rect;
    imageRect_.setWidth( rect.height() * imageRect.width()/imageRect.height() );
    if( imageRect_.width() > rect.width() )
    {
        double scale = double(imageRect_.width())/rect.width();
        imageRect_.setWidth( rect.width() );
        imageRect_.setHeight( imageRect_.height()/scale );
    }

    imageRect_.moveTopLeft( QPoint(
        (rect.width() - imageRect_.width())/2,
        (rect.height() - imageRect_.height())/2 ) );

}

