/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AvatarSelectionWidget.h"

#include "AvatarManipulationDialog.h"
#include "AvatarManipulationWidget.h"
#include "AvatarPixmap.h"
#include "ImageFileDialog.h"
#include "InformationDialog.h"
#include "QtUtil.h"
#include "XmlOptions.h"

#include <QLayout>
#include <QMimeData>
#include <QUrl>

const int imageSize = 128;

//________________________________________________________________________
AvatarSelectionWidget::AvatarSelectionWidget( QWidget* parent ):
    QLabel( parent ),
    Counter( "AvatarSelectionWidget" )
{
    setAlignment( Qt::AlignCenter );
    setWordWrap( true );
    setAcceptDrops( true );

    _setAvatar( QPixmap() );
    _updateButtons();
}

//________________________________________________________________________
void AvatarSelectionWidget::read( void )
{
    Debug::Throw( "AvatarSelectionWidget::read.\n" );

    if( XmlOptions::get().contains( "USERNAME" ) )
    {
        AvatarPixmap pixmap( XmlOptions::get().raw( "USERNAME" ) );
        _setAvatar( pixmap );
        _updateButtons();
    }

}

//________________________________________________________________________
void AvatarSelectionWidget::write( void ) const
{
    Debug::Throw( "AvatarSelectionWidget::write.\n" );

    if( XmlOptions::get().contains( "USERNAME" ) )
    {
        if( avatar_.isNull() )  AvatarPixmap( XmlOptions::get().raw( "USERNAME" ), avatar_ ).remove();
        else AvatarPixmap( XmlOptions::get().raw( "USERNAME" ), avatar_ ).save();
    }

}

//________________________________________________________________________
QPushButton& AvatarSelectionWidget::changeAvatarButton( QWidget* parent )
{
    if( changeAvatarButton_ )
    {

        if( changeAvatarButton_.data()->parent() != parent )
        { changeAvatarButton_.data()->setParent( parent ); }

        return *changeAvatarButton_.data();

    } else {

        changeAvatarButton_ = new QPushButton( tr( "Change Avatar Image" ), parent );
        connect( changeAvatarButton_.data(), SIGNAL(clicked()), SLOT(_changeAvatar()) );
        _updateButtons();
        return *changeAvatarButton_.data();

    }

}

//________________________________________________________________________
QPushButton& AvatarSelectionWidget::resetAvatarButton( QWidget* parent )
{
    if( resetAvatarButton_ )
    {

        if( resetAvatarButton_.data()->parent() != parent )
        { resetAvatarButton_.data()->setParent( parent ); }

        return *resetAvatarButton_.data();

    } else {

        resetAvatarButton_ = new QPushButton( tr( "Unset Avatar Image" ), parent );
        connect( resetAvatarButton_.data(), SIGNAL(clicked()), SLOT(_reset()) );
        _updateButtons();
        return *resetAvatarButton_.data();

    }

}


//______________________________________________________________________
void AvatarSelectionWidget::dragEnterEvent( QDragEnterEvent *event )
{
    Debug::Throw( "AvatarSelectionWidget::dragEnterEvent.\n" );
    if( event->mimeData()->hasUrls() && event->mimeData()->urls().size() == 1 )
    { event->acceptProposedAction(); }
}

//______________________________________________________________________
void AvatarSelectionWidget::dropEvent( QDropEvent *event )
{

    Debug::Throw( "ImageFileDialog::Label::dropEvent.\n" );

    // check if event is valid
    if( !event->mimeData()->hasUrls() ) return;

    // loop over event URLs
    const auto urlList( event->mimeData()->urls() );
    if( urlList.size() != 1 ) return;

    // load image from file
    const auto url( urlList.front() );
    QFileInfo fileInfo( url.toLocalFile() );
    QPixmap pixmap( fileInfo.filePath() );
    if( pixmap.isNull() )
    {
        InformationDialog( this, tr( "Image is invalid." ) ).exec();
        return;
    }

    {
        // open manipulation dialog
        AvatarManipulationDialog dialog( this, pixmap );
        if( !dialog.exec() ) return;

        // update avatar and buttons
        _setAvatar( dialog.manipulationWidget().pixmap( imageSize ) );
        _updateButtons();
    }

}

//________________________________________________________________________
void AvatarSelectionWidget::_changeAvatar( void )
{

    ImageFileDialog dialog( this );
    dialog.setFileMode( QFileDialog::AnyFile );
    dialog.setAcceptMode( QFileDialog::AcceptOpen );
    dialog.setWindowTitle( tr( "Open" ) );
    QtUtil::centerOnParent( &dialog );

    dialog.show();
    if( dialog.exec() == QDialog::Rejected ) return;

    // retrieve selected files
    QStringList files( dialog.selectedFiles() );

    // check file size
    if( files.size() > 1 )
    {
        InformationDialog( this, tr( "Too many files selected." ) ).exec();
        return;
    }

    if( files.size() < 1 )
    {
        InformationDialog( this, tr( "No file selected." ) ).exec();
        return;
    }

    // load image from file
    QPixmap pixmap( files.front() );
    if( pixmap.isNull() )
    {
        InformationDialog( this, tr( "Image is invalid." ) ).exec();
        return;
    }

    {
        // open manipulation dialog
        AvatarManipulationDialog dialog( this, pixmap );
        if( !dialog.exec() ) return;

        // update avatar and buttons
        _setAvatar( dialog.manipulationWidget().pixmap( imageSize ) );
        _updateButtons();
    }

    return;

}


//________________________________________________________________________
void AvatarSelectionWidget::_reset( void )
{
    _setAvatar( QPixmap() );
    _updateButtons();
}

//________________________________________________________________________
void AvatarSelectionWidget::_setAvatar( const QPixmap& pixmap )
{
    Debug::Throw("AvatarSelectionWidget::_setAvatar.\n" );
    avatar_ = pixmap;
    setPixmap( avatar_ );

    if( avatar_.isNull() )
    {

        setFrameStyle( QFrame::Raised|QFrame::StyledPanel );
        setFixedSize( imageSize, imageSize );
        setText( tr( "No image set" ) );

    } else {

        setFrameStyle( QFrame::NoFrame );
        setFixedSize( avatar_.size() );
        setText( QString() );

    }

}

//______________________________________________________________
void AvatarSelectionWidget::_updateButtons( void )
{
    if( avatar_.isNull() )
    {

        if( changeAvatarButton_ ) changeAvatarButton_.data()->setText( tr( "Set Avatar Image" ) );
        if( resetAvatarButton_ ) resetAvatarButton_.data()->setEnabled( false );

    } else {

        if( changeAvatarButton_ ) changeAvatarButton_.data()->setText( tr( "Change Avatar Image" ) );
        if( resetAvatarButton_ ) resetAvatarButton_.data()->setEnabled( true );

    }

}

