#ifndef ChatCommand_h
#define ChatCommand_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "BaseChatCommand.h"
#include "ConnectionId.h"
#include "CommandLineArguments.h"
#include "TimeStamp.h"

#include <QDataStream>
#include <QString>
#include <QHash>

//_________________________________________
//* command to be send over the server
class ChatCommand:public BaseChatCommand
{

    public:

    //* command type
    enum CommandType
    {
        None,
        Message,
        ConnectionRequest,
        ConnectionAccepted,
        FetchMessages,
        ContactOnline,
        ContactOffline,
        FetchContacts,
        Contact,
        PingServer,
        PongServer,
        PingClient,
        PongClient,
        ClearContactList,
        ClearMessageList,
        UserStateChanged,
        StartMessageLog,
        StopMessageLog,
        UserTyping,
        UserNameChanged,
        FileTransferRequest,
        FileTransferAccepted,
        FileTransferDenied,
        FileTransferCanceled,
        ConnectionDenied
    };

    //* list
    using List = QList<ChatCommand>;

    //* constructor
    ChatCommand( const ConnectionId& = ConnectionId(), const CommandType& = None );

    //* equal-to operator
    bool operator == ( const ChatCommand& other ) const
    {
        return creationTime_ == other.creationTime_ &&
            connectionId() == other.connectionId() &&
            command_ == other.command_;
    }

    //* less-than operator
    bool operator < ( const ChatCommand& other ) const
    {
        if( creationTime_ != other.creationTime_ ) return creationTime_ < other.creationTime_;
        if( connectionId() != other.connectionId() ) return connectionId() < other.connectionId();

        if( command_ != other.command_ ) return command_ < other.command_;
        return false;
    }

    //*@name accessors
    //@{

    //* validity
    bool isValid( void ) const
    { return BaseChatCommand::isValid() && command_ != None && creationTime_.isValid(); }

    //* true if this command is a child of argument
    /*! this is needed to have commands appear in Tree view */
    bool isChild( const ChatCommand& command ) const
    { return creationTime_ >= command.creationTime_ && connectionId() == command.connectionId(); }

    //* time stamp
    const TimeStamp& creationTime( void ) const
    { return creationTime_; }

    //* time stamp
    const TimeStamp& receptionTime( void ) const
    { return receptionTime_; }

    //* command
    CommandType command( void ) const
    { return command_; }

    //* command name
    QString commandName( void ) const
    {
        _initializeCommandNames();
        return _commandNames()[ command() ];
    }

    //* raw text
    const QString& text( void ) const
    { return text_; }

    //* arguments
    const CommandLineArguments& arguments( void ) const
    { return arguments_; }

    //@}

    //*@name modifiers
    //@{

    //* time stamp
    void setCreationTime( const TimeStamp& value )
    { creationTime_ = value; }

    //* time stamp
    void setReceptionTime( const TimeStamp& value )
    { receptionTime_ = value; }

    //* command
    void setCommand( const CommandType& command )
    { command_ = command; }

    //* raw text
    void setText( const QString& value )
    { text_ = value; }

    //* argument
    void setArguments( const CommandLineArguments& arguments )
    { arguments_ = arguments; }

    //@}

    private:

    //* command names
    using CommandMap = QHash<CommandType, QString >;

    //* command names
    static CommandMap& _commandNames( void );

    //* map command types to names
    void _initializeCommandNames( void ) const;

    //* time stamp
    TimeStamp creationTime_;

    //* time stamp
    TimeStamp receptionTime_;

    //* command
    CommandType command_ = None;

    //* arguments
    CommandLineArguments arguments_;

    //* raw text
    QString text_;

    //*@name serializers
    //@{
    friend QDataStream& operator << (QDataStream&, const ChatCommand& );
    friend QDataStream& operator >> (QDataStream&, ChatCommand& );
    //@}

};

#endif
