#ifndef ChatMessage_h
#define ChatMessage_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ChatCommand.h"
#include "ConnectionId.h"
#include "Counter.h"
#include "Debug.h"
#include "TimeStamp.h"
#include "QOrderedSet.h"

#include <QDomElement>
#include <QDomDocument>
#include <QPair>
#include <QString>

//* chat message
class ChatMessage:public Counter
{

    public:

    //* set
    class Set: public QOrderedSet<ChatMessage>
    {
        public:

        //* constructor
        explicit Set( void )
        {}

        //* constructor
        Set( const QOrderedSet<ChatMessage>& other ):
            QOrderedSet<ChatMessage>(other)
        {}

        //* constructor
        Set( const QDomElement& );

        //* dom
        QDomElement domElement( QDomDocument& parent ) const;

    };

    //* list
    class List: public QList<ChatMessage>
    {

        public:

        //* constructor
        explicit List( void )
        {}

        //* constructor from QList
        explicit List( const QList<ChatMessage>& other ):
            QList<ChatMessage>( other )
        {}

        //* assignment from QList
        const List& operator = (const QList<ChatMessage>& other )
        { *this = List( other ); return *this; }

        //* insert chat message in list
        /** returns true if order of the list was altered */
        bool insert( const ChatMessage& message );

        //* insert chat message in list
        /** returns true if order of the list was altered */
        bool insertUnique( const ChatMessage& message );

    };

    //* constructor
    ChatMessage( const ConnectionId& = ConnectionId(), const QString& = QString() );

    //* constructor from chat command
    ChatMessage( const ChatCommand& );

    //* constructor
    ChatMessage( const QDomElement& );

    //* equal-to operator
    bool operator == ( const ChatMessage& command ) const
    {
        return
            creationTime() == command.creationTime() &&
            connectionId() == command.connectionId();
    }

    //* less-than operator
    bool operator < ( const ChatMessage& command ) const
    {
        if( creationTime() != command.creationTime() ) return creationTime() < command.creationTime();
        else if( connectionId() != command.connectionId() ) return connectionId() < command.connectionId();
        else return false;
    }

    //*@name accessors
    //@{

    //* dom element
    QDomElement domElement( QDomDocument& ) const;

    //* matching chat command
    ChatCommand toChatCommand( void ) const;

    //* message type
    enum MessageType
    {
        NormalMessage = 0,
        TransferNotificationMessage = 1
    };

    //* message type
    MessageType messageType( void ) const
    { return messageType_; }

    //* message type
    bool isTransferNotificationMessage( void ) const
    { return messageType_ == TransferNotificationMessage; }

    //* message type
    bool isNormalMessage( void ) const
    { return messageType_ == NormalMessage; }

    //* validity
    bool isValid( void ) const
    { return connectionId_.isValid() && creationTime_.isValid(); }

    //* time stamp
    const TimeStamp& creationTime( void ) const
    { return creationTime_; }

    //* time stamp
    const TimeStamp& receptionTime( void ) const
    { return receptionTime_; }

    //* connection id
    const ConnectionId& connectionId( void ) const
    { return connectionId_; }

    //* list of recipients
    QStringList recipients( void ) const
    { return recipients_; }

    //* raw text
    const QString& text( void ) const
    { return text_; }

    //* block range corresponding to this message in editor
    using BlockRange = QPair<int,int>;
    const BlockRange& blockRange( void ) const
    { return blockRange_; }

    //@}

    //*@name modifiers
    //@{

    //* message type
    void setMessageType( MessageType value )
    { messageType_ = value; }

    //* timestamp
    void setCreationTime( const TimeStamp& value )
    { creationTime_ = value; }

    //* timestamp
    void setReceptionTime( const TimeStamp& value )
    { receptionTime_ = value; }

    //* connection id
    void setConnectionId( const ConnectionId& connectionId )
    { connectionId_ = connectionId; }

    //* add recipient
    void addRecipient( const QString& recipient )
    { recipients_.append( recipient ); }

    //* set recipients
    void setRecipients( QStringList recipients )
    { recipients_ = recipients; }

    //* raw text
    void setText( const QString& value )
    { text_ = value; }

    //* block range
    void setBlockRange( int first, int second )
    { blockRange_ = qMakePair( first, second ); }

    //* block range
    void setBlockRange( const BlockRange& value )
    { blockRange_ = value; }

    //@}

    class SortReceptionTimeFTor
    {
        public:

        //* predicate
        bool operator()( const ChatMessage& first, const ChatMessage& second )
        { return first.receptionTime() < second.receptionTime(); }

    };

    private:

    //* message type
    MessageType messageType_ = NormalMessage;

    //* time stamp
    TimeStamp creationTime_;

    //* time stamp
    TimeStamp receptionTime_;

    //* connection id
    ConnectionId connectionId_;

    //* list of recipient user names
    QStringList recipients_;

    //* raw text
    QString text_;

    //* block range in editor
    BlockRange blockRange_;

    //* streamer
    friend QTextStream& operator << (QTextStream&, const ChatMessage& );

};

#endif
