/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ChatMessageListDialog.h"

#include "ChatMessage.h"
#include "ChatMessageSet.h"
#include "ContextMenu.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "QuestionDialog.h"
#include "TextEditor.h"
#include "TreeView.h"
#include "XmlOptions.h"

#include <QHeaderView>
#include <QLayout>
#include <QShortcut>

//______________________________________________________________________
ChatMessageListDialog::ChatMessageListDialog( QWidget* parent, int debugLevel ):
    CustomDialog( parent, OkButton|CancelButton ),
    debugLevel_( debugLevel )
{

    Debug::Throw( "ChatMessageListDialog::ChatMessageListDialog.\n" );
    setOptionName( "CHAT_MESSAGE_LIST_DIALOG" );

    QHBoxLayout* hLayout( new QHBoxLayout() );
    hLayout->setSpacing(5);
    hLayout->setMargin(0);
    mainLayout().addLayout( hLayout );

    // setup list
    QVBoxLayout* vLayout = new QVBoxLayout();
    vLayout->setSpacing(5);
    vLayout->setMargin(0);
    hLayout->addLayout( vLayout );

    vLayout->addWidget( list_ = new TreeView( this ) );
    list_->setModel( &model_ );
    list_->setSortingEnabled( false );
    list_->setSelectionMode( QAbstractItemView::ContiguousSelection );
    list_->setOptionName( "CHAT_MESSAGE_LIST" );

    // label
    if( QBoxLayout* boxLayout = qobject_cast<QBoxLayout*>( buttonBox().layout() ) )
    {

        boxLayout->insertWidget( 0, statusLabel_ = new QLabel( &buttonBox() ) );

    } else {

        vLayout->addWidget( statusLabel_ = new QLabel( this ) );

    }

    // buttons
    vLayout = new QVBoxLayout();
    vLayout->setSpacing(5);
    vLayout->setMargin(0);
    hLayout->addLayout( vLayout );

    vLayout->addWidget( editButton_ = new QPushButton( IconEngine::get( IconNames::Edit ), tr( "Edit" ), this ) );
    connect( editButton_, SIGNAL(clicked()), SLOT(_edit()) );

    vLayout->addWidget( removeButton_ = new QPushButton( IconEngine::get( IconNames::Delete ), tr( "Delete" ), this ) );
    connect( removeButton_, SIGNAL(clicked()), SLOT(_remove()) );

    vLayout->addWidget( reloadButton_ = new QPushButton( IconEngine::get( IconNames::Reload ), tr( "Reload" ), this ) );
    connect( reloadButton_, SIGNAL(clicked()), SLOT(_reload()) );

    vLayout->addStretch( 1 );

    // actions
    QMenu* menu( new ContextMenu( list_ ) );

    addAction( editAction_ = new QAction( IconEngine::get( IconNames::Edit ), tr( "Edit" ), this ) );
    connect( editAction_, SIGNAL(triggered()), SLOT(_edit()) );
    menu->addAction( editAction_ );

    addAction( removeAction_ = new QAction( IconEngine::get( IconNames::Delete ), tr( "Delete" ), this ) );
    connect( removeAction_, SIGNAL(triggered()), SLOT(_remove()) );
    removeAction_->setShortcut( QKeySequence::Delete );
    menu->addAction( removeAction_ );

    addAction( reloadAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Reload" ), this ) );
    connect( reloadAction_, SIGNAL(triggered()), SLOT(_reload()) );
    reloadAction_->setShortcut( QKeySequence::Refresh );

    menu->addAction( reloadAction_ );

    // updates
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateButtons()) );
    connect( list_->selectionModel(), SIGNAL(currentChanged(QModelIndex,QModelIndex)), SLOT(_updateButtons()) );
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateStatusLabelDelayed()) );
    connect( list_, SIGNAL(activated(QModelIndex)), SLOT(_edit()) );

    // save messages
    connect( &okButton(), SIGNAL(clicked()), SLOT(_saveModel()) );

    // shortcuts
    cancelButton().setShortcut( Qt::Key_Escape );
    new QShortcut( QKeySequence::Close, this, SLOT(reject()) );
    new QShortcut( QKeySequence::Quit, this, SLOT(reject()) );

    _setModified( false );
    _reload();
    _updateButtons();
    _updateStatusLabel();

    // scroll down to the bottom of the list
    list_->scrollTo( model_.index( model_.rowCount()-1, 0 ) );

}

//______________________________________________________________________
void ChatMessageListDialog::_reload( void )
{

    Debug::Throw( "DocumentClassManager::Dialog::_reload.\n" );

    // read messages from file
    ChatMessageSet messageSet;
    messageSet.setDebugLevel( debugLevel_ );
    messageSet.setUseCompression( XmlOptions::get().get<bool>( "COMPRESS_HISTORY" ) );

    if( messageSet.setFilename( File( XmlOptions::get().raw("HISTORY_FILE") ) ) )
    {
        // read all messages
        messageSet.read();

        // get corresponding filenames
        // this is needed to rewrite all, in case of modifications
        filenames_ = messageSet.filenames();
    }

    // assign to model
    model_.set( messageSet.messages().toList() );
    _updateStatusLabel();

    // reset modified flags
    _setModified( false );

}

//______________________________________________________________________
void ChatMessageListDialog::_edit( void )
{
    Debug::Throw( "ChatMessageListDialog::_edit.\n" );

    // loop over selected items
    const QList<QModelIndex> indexes( list_->selectionModel()->selectedRows() );
    if( indexes.size() != 1 ) return;

    ChatMessage message( model_.get( indexes.front() ) );

    // create dialog
    CustomDialog dialog( this );
    dialog.setWindowTitle( tr( "Edit message - Homechat-server" ) );
    dialog.okButton().setText( tr( "Save" ) );
    dialog.okButton().setIcon( IconEngine::get( IconNames::Save ) );
    dialog.setOptionName( "EDIT_MESSAGE" );

    // editor
    TextEditor* editor = new TextEditor( &dialog );
    dialog.mainLayout().addWidget( editor );
    editor->setText( message.text() );

    if( !dialog.exec() ) return;

    const QString text( editor->toPlainText() );
    if( message.text() != text )
    {
        message.setText( text );
        model_.add( message );
        _setModified( true );
    }

}

//______________________________________________________________________
void ChatMessageListDialog::_remove( void )
{
    Debug::Throw( "ChatMessageListDialog::_remove.\n" );

    // loop over selected items
    const ChatMessage::List removedMessages( model_.get( list_->selectionModel()->selectedRows() ) );

    if( removedMessages.empty() ) return;

    QuestionDialog* dialog( _removeMessagesQuestionDialog( removedMessages ) );

    // ask confirmation
    QString buffer = (removedMessages.size() == 1 ) ?
        tr( "Delete the selected message ?" ):
        QString( tr( "Delete the selected %1 messages ?" ) ).arg( removedMessages.size() );
    dialog->setText( buffer );

    dialog->okButton().setText( tr( "Delete" ) );
    dialog->okButton().setIcon( IconEngine::get( IconNames::Delete ) );

    // remove
    if( dialog->exec() )
    {
        model_.remove( removedMessages );
        _setModified( true );
    }

    // delete dialog and return
    dialog->deleteLater();
    _updateStatusLabel();

    return;

}

//______________________________________________________________________
void ChatMessageListDialog::_saveModel( void )
{
    Debug::Throw( "ChatMessageListDialog::_saveModel.\n" );
    if( !modified_ ) return;

    QuestionDialog dialog( this, tr( "Save message list modifications to disk ?" ) );
    dialog.okButton().setText( tr( "Save" ) );
    dialog.okButton().setIcon( IconEngine::get( IconNames::Save ) );

    if( dialog.exec() )
    {

        // get list of messages from set
        ChatMessage::Set messages = ChatMessage::Set().fromList( model_.get() );

        // create chat message set class to do the writing
        ChatMessageSet messageSet;
        messageSet.setDebugLevel( debugLevel_ );
        messageSet.setUseCompression( XmlOptions::get().get<bool>( "COMPRESS_HISTORY" ) );

        QRegExp dateRegExp( "-(\\d{4,4})-(\\d{2,2})$" );
        for( const auto& filename:filenames_ )
        {

            // parse time stamp
            if( filename.localName().indexOf( dateRegExp ) < 0 ) continue;

            const int year = dateRegExp.cap(1).toInt();
            const int month = dateRegExp.cap(2).toInt();

            QDateTime first( QDate( year, month, 1 ) );
            first.setTime( QTime( 0, 0 ) );

            const QDateTime second = first.addMonths(1);
            ChatMessage::Set local;
            for( const auto& message:messages )
            {
	      if( message.creationTime().unixTime() >= (time_t) first.toTime_t() && message.creationTime().unixTime() < (time_t) second.toTime_t() )
                {
                    local.insert( message );
                    messages.remove( message );
                }
            }

            // write
            messageSet.write( filename, local );

        }

        Q_ASSERT( messages.empty() );

    }

    _updateStatusLabel();
    return;

}

//____________________________________________________________________________________________
void ChatMessageListDialog::_updateStatusLabelDelayed( void )
{ statusLabelUpdateTimer_.start( 200, this ); }

//____________________________________________________________________________________________
void ChatMessageListDialog::_updateStatusLabel( void )
{
    const QList<QModelIndex> indexes( list_->selectionModel()->selectedRows() );
    int rowCount( indexes.size() );
    const bool selection( rowCount > 0 );
    if( !rowCount ) rowCount = model_.rowCount();

    QString buffer;
    if( rowCount == 0 ) buffer = QString();
    else if( rowCount == 1 ) buffer = tr( "One message" );
    else buffer = tr( "%1 messages" ).arg( rowCount );

    if( selection ) buffer += tr(" selected");

    statusLabel_->setText( buffer );
    statusLabel_->update();

}

//______________________________________________________________________
void ChatMessageListDialog::_updateButtons( void )
{

    const bool editEnabled( list_->selectionModel()->selectedRows().size() == 1 );
    editAction_->setEnabled( editEnabled );
    editButton_->setEnabled( editEnabled );

    const bool removeEnabled( !list_->selectionModel()->selectedRows().isEmpty() );
    removeAction_->setEnabled( removeEnabled );
    removeButton_->setEnabled( removeEnabled );
}

//____________________________________
void ChatMessageListDialog::timerEvent( QTimerEvent* event )
{
    if( event->timerId() == statusLabelUpdateTimer_.timerId() )
    {

        statusLabelUpdateTimer_.stop();
        _updateStatusLabel();

    } else return CustomDialog::timerEvent( event );
}

//______________________________________________________________________
void ChatMessageListDialog::_setModified( bool value )
{
    modified_ = value;
    setWindowTitle( modified_ ?
        tr( "Message list (modified) - Homechat-server" ):
        tr( "Message list - Homechat-server" ) );
}

//______________________________________________________________________
QuestionDialog* ChatMessageListDialog::_removeMessagesQuestionDialog( const ChatMessage::List& messages )
{
    QuestionDialog* dialog = new QuestionDialog( this );
    dialog->setOptionName( "REMOVE_MESSAGES" );

    // create details tree view
    TreeView* list = new TreeView();
    ChatMessageModel* model = new ChatMessageModel( dialog );
    list->setModel( model );
    model->set( messages );

    // assign
    dialog->setDetails( list );

    return dialog;
}

