/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ChatMessageModel.h"
#include "Debug.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "Singleton.h"
#include "XmlColor.h"
#include "XmlOptions.h"

#include <QPalette>

// static members
QVector<QColor> ChatMessageModel::colors_;
int ChatMessageModel::colorIndex_ = 0;
ChatMessageModel::UserMap ChatMessageModel::userMap_;

//_______________________________________________
const QString ChatMessageModel::columnTitles_[ ChatMessageModel::nColumns ] =
{
    tr( "Sender" ),
    tr( "Recipients" ),
    tr( "Created" ),
    tr( "Recieved" ),
    tr( "Content" )
};

//______________________________________________________________________
ChatMessageModel::ChatMessageModel( QObject* parent ):
    ListModel<ChatMessage>( parent ),
    Counter( "ChatMessageModel" )
{

    // configuration
    connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();

}

//______________________________________________________________________
QVariant ChatMessageModel::data( const QModelIndex& index, int role ) const
{

    // check index
    if( !contains( index ) ) return QVariant();

    // retrieve associated file info
    const ChatMessage& command( get()[index.row()] );

    // return text associated to file and column
    switch( role )
    {

        case Qt::DisplayRole:
        {
            switch( index.column() )
            {
                case Sender: return command.connectionId().userString();
                case Recipients:
                {
                    QStringList recipients;
                    for( auto recipient:command.recipients() )
                    {
                        if( !recipient.isEmpty() ) recipient[0] = recipient[0].toUpper();
                        recipients.append( recipient );
                    }
                    return recipients.join( " " );
                }

                case Creation: return command.creationTime().toString();
                case Reception: return command.receptionTime().toString();
                case Text: return command.text();
                default: return QVariant();
            }
            break;
        }

        case Qt::ForegroundRole:
        {
            switch( index.column() )
            {
                case Sender:
                {
                    if( colors_.isEmpty() ) return QVariant();
                    const QString user( command.connectionId().user() );
                    QColor color;
                    if( !userMap_.contains( user ) )
                    {
                        // reset color index if needed
                        if( colorIndex_ >= colors_.size() ) colorIndex_ = 0;
                        color = colors_[colorIndex_];
                        userMap_.insert( user, color );
                        ++colorIndex_;

                    } else color = userMap_[user];
                    return color;

                }

                default: return QVariant();

            }

        }

        case Qt::TextAlignmentRole:
        {

            switch( index.column() )
            {
                case Recipients:
                case Sender: return static_cast<int>(Qt::AlignRight|Qt::AlignVCenter);
                case Creation:
                case Reception: return Qt::AlignCenter;
                default: return QVariant();
            }

            break;
        }

        default: return QVariant();

    }

    return QVariant();

}

//______________________________________________________________________
QVariant ChatMessageModel::headerData(int section, Qt::Orientation orientation, int role ) const
{

   switch( role )
    {

        case Qt::DisplayRole: return columnTitles_[section];
        case Qt::TextAlignmentRole:
        {

            switch( section )
            {
                case Recipients:
                case Sender: return static_cast<int>(Qt::AlignRight|Qt::AlignVCenter);
                case Creation:
                case Reception: return Qt::AlignCenter;
                default: return QVariant();
            }

            break;
        }

        default: return QVariant();

    }


    // return empty
    return QVariant();

}

//__________________________________________________________
void ChatMessageModel::_updateConfiguration( void )
{

    Debug::Throw( "ChatMessageModel::_updateConfiguration.\n" );

    colors_.clear();
    for( const auto& color:XmlOptions::get().specialOptions( "USER_COLOR" ) )
    { colors_.append( color.get<Base::Color>() ); }

}
