
/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ChatMessageSet.h"
#include "Debug.h"
#include "TimeStamp.h"
#include "XmlDef.h"

#include <QDir>

//________________________________________________________
ChatMessageSet::ChatMessageSet( void ):
    Counter( "ChatMessageSet" )
{}

//________________________________________________________
bool ChatMessageSet::needsReading( const TimeStamp& timestamp ) const
{

    Debug::Throw( "ChatMessageSet::needsReading.\n" );
    if( filename_.isEmpty() ) return false;
    else return !_filenames( timestamp ).isEmpty();

}

//________________________________________________________
ChatMessage::List ChatMessageSet::fetch( const TimeStamp& timeStamp, const QStringList& sourceRecipients ) const
{
    // find all matching messages
    ChatMessage::List messages;
    ChatMessage last;
    for( const auto& message:messages_ )
    {
        bool accepted = false;

        // to accept a message, both the sender and at least one of the recipients must match
        // source recipients
        if( sourceRecipients.contains( message.connectionId().user() ) )
        {
            for( const auto& recipient:message.recipients() )
            {
                if( sourceRecipients.contains( recipient ) )
                { accepted = true; break; }
            }
        }
        if( !accepted ) continue;

        // store last message
        if( !last.isValid() || message.receptionTime() > last.receptionTime() )
        { last = message; }

        // skip message if too old
        if( timeStamp.isValid() && message.receptionTime() < timeStamp ) continue;

        // store in list
        messages.append( message );

    }

    if( messages.isEmpty() && last.isValid() ) messages.append( last );
    qSort( messages.begin(), messages.end(), ChatMessage::SortReceptionTimeFTor() );

    return messages;

}

//________________________________________________________
bool ChatMessageSet::setFilename( File filename )
{

    Debug::Throw() << "ChatMessageSet::setFilename - " << filename << endl;
    if( filename != filename_ )
    {

        // make sure path exists
        QDir().mkpath( filename.path() );

        // update filename
        filename_ = filename;
        filenameNew_ = filename_+".new";

        // clear messages and list of opened filenames
        messages_.clear();
        filenames_.clear();
        return true;

    } else return false;

}

//________________________________________________________
bool ChatMessageSet::read( const TimeStamp& timestamp )
{
    if( filename_.isEmpty() ) return false;

    // at first call, merge messages
    if( messages_.empty() )
    { mergeMessages(); }

    // read messages from selected files
    for( const auto& filename:_filenames( timestamp ) )
    {

        // append to list of openned files
        filenames_.insert( filename );

        // read
        ChatMessage::Set local;
        read( filename, local );

        // merge
        for( const auto& message:local )
        { _append( message, messages_ ); }

    }

    return true;
}

//____________________________________________
bool ChatMessageSet::write()
{
    Debug::Throw( "ChatMessageSet::write.\n" );
    if( filenameNew_.isEmpty() )
    {
        Debug::Throw(0) << "ChatMessageSet::write - invalid file." << endl;
        return false;
    }

    if( !modified_ ) return true;
    modified_ = false;
    return write( filenameNew_, newMessages_ );
}

//________________________________________________________________________________________
bool ChatMessageSet::read( const File& file, ChatMessage::Set& messages ) const
{

    // parse the file
    XmlDocument document;
    if( !_readXmlDocument( file, document ) )
    { return false; }

    // look for relevant element
    QDomNodeList topNodes = document.elementsByTagName( Xml::MessageList );
    if( topNodes.isEmpty() ) return false;

    // load
    messages = ChatMessage::Set( topNodes.at(0).toElement() );

    if( messages.size() )
    { Debug::Throw( debugLevel_ ) << "ChatMessageSet::read - read " << messages.size() << " messages from file " << file << endl; }

    return true;

}

//________________________________________________________________________________________
bool ChatMessageSet::write( const File& file, const ChatMessage::Set& messages ) const
{
    Debug::Throw() << "ChatMessageSet::write - file: " << file << endl;

    // create document and read
    XmlDocument document;
    _readXmlDocument( file, document );

    // append truncated list
    QDomElement top( messages.domElement( document ) );
    document.replaceChild( top );
    {
        QFile qfile( file );
        if( !qfile.open( QIODevice::WriteOnly ) )
        {
            Debug::Throw(0) << "ChatMessageSet::write - cannot open file: " << file << " for writting." << endl;
            return false;
        }

        if( useCompression_ ) qfile.write( qCompress( document.toByteArray() ) );
        else qfile.write( document.toByteArray() );
    }

    if( messages.size() )
    { Debug::Throw(  debugLevel_ ) << "ChatMessageSet::write - wrote " << messages.size() << " messages to file " << file << endl; }
    return true;

}

//_________________________________________________
void ChatMessageSet::append( const ChatMessage& message )
{

    // add to local
    if( message.receptionTime().isValid() )
    {

        _append( message, messages_ );
        _append( message, newMessages_ );

    } else {

        ChatMessage copy( message );
        copy.setReceptionTime( TimeStamp::now() );
        _append( copy, messages_ );
        _append( copy, newMessages_ );

    }

    modified_ = true;

}

//________________________________________________________________________________________
void ChatMessageSet::compress( void ) const
{
    Debug::Throw( "ChatMessageSet::compress.\n" );
    const auto filenames( _filenames( TimeStamp() ) );
    for( const auto& filename:filenames )
    {
        QFile qfile( filename );
        if( !qfile.open( QIODevice::ReadOnly ) )
        {
            Debug::Throw(0) << "ChatMessageSet::compress - cannot open file: " << filename << endl;
            continue;
        }

        // read
        const QByteArray content( qfile.readAll() );
        QByteArray uncompressed;
        if( useCompression_ ) uncompressed = qUncompress( content );
        if( !uncompressed.size() ) uncompressed = content;

        // close
        qfile.close();

        // write
        if( !qfile.open( QIODevice::WriteOnly ) )
        {
            Debug::Throw(0) << "ChatMessageSet::compress - cannot open file: " << filename << endl;
            continue;
        }

        Debug::Throw(debugLevel_) << "ChatMessageSet::compress - compressing " << filename << endl;
        qfile.write( qCompress( uncompressed ) );

    }

}

//________________________________________________________________________________________
void ChatMessageSet::uncompress( void ) const
{
    Debug::Throw( 0, "ChatMessageSet::uncompress.\n" );
    const auto filenames( _filenames( TimeStamp() ) );
    for( const auto& filename:filenames )
    {
        QFile qfile( filename );
        if( !qfile.open( QIODevice::ReadOnly ) )
        {
            Debug::Throw(0) << "ChatMessageSet::uncompress - cannot open file: " << filename << endl;
            continue;
        }

        // read
        const QByteArray content( qfile.readAll() );
        QByteArray uncompressed;
        if( useCompression_ ) uncompressed = qUncompress( content );
        if( !uncompressed.size() ) uncompressed = content;

        // close
        qfile.close();

        // write
        if( !qfile.open( QIODevice::WriteOnly ) )
        {
            Debug::Throw(0) << "ChatMessageSet::uncompress - cannot open file: " << filename << endl;
            continue;
        }

        Debug::Throw(debugLevel_) << "ChatMessageSet::uncompress - uncompressing " << filename << endl;
        qfile.write( uncompressed );

    }

}

//________________________________________________________
bool ChatMessageSet::mergeMessages( void ) const
{

    Debug::Throw( "ChatMessageSet::mergeMessages.\n" );
    if( filename_.isEmpty() )
    {
        Debug::Throw(0) << "ChatMessageSet::read - invalid file" << endl;
        return false;
    }

    // read messages from new message file, and merge
    ChatMessage::Set newMessages;
    read( filenameNew_, newMessages );

    // for migration also read, merge, and erase messages from old file
    if( filename_.exists() )
    {

        // read from file
        ChatMessage::Set local;
        read( filename_, local );

        // merge
        for( const auto& message:local )
        { _append( message, newMessages ); }

        // remove (this file is not used anymore)
        filename_.remove();

    }

    // get the list of filenames to which they should be merged
    QList<TimeStamp> timestamps;
    for( const auto& message:newMessages )
    {
        if(
            timestamps.empty() ||
            message.creationTime().month() != timestamps.back().month() ||
            message.creationTime().year() != timestamps.back().year() )
        { timestamps.append( message.creationTime() ); }

    }

    // read all
    ChatMessage::Set messages;
    for( const auto& timestamp:timestamps )
    {

        // read from file
        ChatMessage::Set local;
        read( _filename( timestamp ), local );

        // merge
        for( const auto& message:local )
        { _append( message, messages ); }

    }

    // merge new messages
    for( const auto& message:newMessages )
    { _append( message, messages ); }

    // clear new messages
    write( filenameNew_, ChatMessage::Set() );

    // rewrite all
    for( const auto& timestamp:timestamps )
    {

        // get time limits
        QDateTime first;
        first.setTime_t( timestamp );
        first.setDate( QDate( first.date().year(), first.date().month(), 1 ) );
        first.setTime( QTime( 0, 0 ) );

        QDateTime second = first.addMonths(1);

        ChatMessage::Set local;
        for( auto iter = messages.begin(); iter != messages.end(); )
        {
            const auto& message( *iter );
            if( message.creationTime().unixTime() >= (time_t) first.toTime_t() && message.creationTime().unixTime() < (time_t) second.toTime_t() )
            {

                local.insert( message );
                iter = messages.erase( iter );

            } else ++iter;

        }

        // write
        write( _filename( timestamp ), local );

    }

    // make sure messages is now empty
    if( !messages.empty() )
    { Debug::Throw(0) << "ChatMessageSet::mergeMessages - something wrong. messages not empty" << endl; }

    return true;
}

//________________________________________________________
QString ChatMessageSet::_filename( const TimeStamp& timestamp ) const
{ return filename_ + timestamp.toString( "-yyyy-MM" ); }

//________________________________________________________
QStringList ChatMessageSet::_filenames( const TimeStamp& timestamp ) const
{

    Debug::Throw( "ChatMessageSet::_filenames.\n" );
    if( filename_.isEmpty() ) return QStringList();

    // get timestamp matching the first day of the timeStamp's month
    QDateTime first;
    if( timestamp.isValid() )
    {
        first.setTime_t( timestamp );
        first.setDate( QDate( first.date().year(), first.date().month(), 1 ) );
        first.setTime( QTime( 0, 0 ) );
    }

    // generate filename list
    QStringList filenames;
    const QStringList filenameCandidates = QDir( filename_.path() ).entryList(
        QStringList() << QString( filename_.localName() + "-*" ),
        QDir::Hidden|QDir::Files,
        QDir::Name );

    QRegExp dateRegExp( "-(\\d{4,4})-(\\d{2,2})$" );
    for( const auto& filename:filenameCandidates )
    {

        // parse date
        if( filename.indexOf( dateRegExp ) < 0 ) continue;
        if( timestamp.isValid() )
        {

            const int year = dateRegExp.cap(1).toInt();
            const int month = dateRegExp.cap(2).toInt();

            QDateTime time( QDate( year, month, 1 ) );
            time.setTime( QTime( 0, 0 ) );
            if( time < first ) continue;

        }

        // add path
        File fullname = File( filename ).addPath( filename_.path() );
        if( !filenames_.contains( fullname ) )
        { filenames.append( fullname ); }

    }

    return filenames;

}

//_________________________________________________
bool ChatMessageSet::_readXmlDocument( const File& file, XmlDocument& document ) const
{

    QFile qfile( file );
    if( !qfile.open( QIODevice::ReadOnly ) )
    {
        Debug::Throw(0) << "ChatMessageSet::_readXmlDocument - cannot open file: " << file << endl;
        return false;
    }

    if( useCompression_ )
    {

        Debug::Throw( "ChatMessageSet::_readXmlDocument - trying compressed.\n" );

        // try read compressed and try uncompress
        const QByteArray content( qfile.readAll() );
        QByteArray uncompressed = qUncompress( content );

        // try read raw if failed
        if( !uncompressed.size() ) uncompressed = content;

        // pass contents to XmlDocument, after trying to uncompress
        if( !document.setContent( uncompressed ) )
        {
            Debug::Throw(0) << document.error() << endl;
            return false;
        }

    } else {

        // read raw
        Debug::Throw( "ChatMessageSet::_readXmlDocument - trying uncompressed.\n" );
        QByteArray uncompressed( qfile.readAll() );
        bool success( false );
        if( !( success = document.setContent( uncompressed ) ) )
        {

            // try read compressed and uncompress
            uncompressed = qUncompress( uncompressed );
            success = document.setContent( uncompressed );
        }

        // pass contents to XmlDocument, after trying to uncompress
        if( !success )
        {
            Debug::Throw(0) << document.error() << endl;
            return false;
        }

    }

    return true;

}

//_________________________________________________
void ChatMessageSet::_append( const ChatMessage& message, ChatMessage::Set& messages ) const
{

    // first check if message is in set
    auto iter( messages.find( message ) );
    if( iter != messages.end() )
    {

        ChatMessage copy( message );
        copy.setReceptionTime( iter->receptionTime() );
        messages.erase( iter );
        messages.insert( copy );

    } else messages.insert( message );

    return;

}
