/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ClientApplication.h"

#include "AvatarSelectionWidget.h"
#include "BaseFileInfo.h"
#include "ChatMessage.h"
#include "ClientInterface.h"
#include "ClientConfigurationDialog.h"
#include "ContactListWindow.h"
#include "CustomDialog.h"
#include "File.h"
#include "FileTransferManager.h"
#include "HistoryMenu.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "LineEditor.h"
#include "ConversationWindow.h"
#include "ProgressStatusBar.h"
#include "SshConnection.h"
#include "SshInterface.h"
#include "SystemTrayIcon.h"
#include "SystemNotifications.h"
#include "XmlOptions.h"

#if USE_ASPELL
#include "AutoSpellConfiguration.h"
#include "SpellCheckConfiguration.h"
#include "SpellInterface.h"
#endif

#include <QApplication>
#include <QTimer>

//____________________________________________
ClientApplication::ClientApplication( CommandLineArguments arguments ) :
    BaseApplication( nullptr, arguments ),
    Counter( "ClientApplication" )
{}

//____________________________________________
bool ClientApplication::realizeWidget( void )
{
    Debug::Throw( "ClientApplication::realizeWidget.\n" );

    // check if the method has already been called.
    if( !BaseApplication::realizeWidget() ) return false;

    // need to redirect closeAction to proper exit
    closeAction().disconnect();
    connect( &closeAction(), SIGNAL(triggered()), SLOT(_exit()) );

    // actions
    _installActions();

    // ssh interface
    sshInterface_ = new Ssh::Interface( this, commandLineParser( _arguments() ) );

    // tcp client interface
    interface_ = new ClientInterface( this, commandLineParser( _arguments() ) );
    interface_->readContacts();

    // system notifications
    systemNotifications_ = new SystemNotifications( this );
    systemNotifications_->initialize();
    systemNotifications_->setApplicationName( applicationName() );
    systemNotifications_->setApplicationIcon( applicationIcon() );
    connect( systemNotifications_, SIGNAL(actionInvoked(quint32,QString)), this, SLOT(_raiseContactList()) );

    // sound notifications
    soundNotifications_ = new SoundNotifications( this );

    // contact list window
    contactListWindow_ = new ContactListWindow();
    contactListWindow_->setConnectionId( interface_->connectionId() );
    contactListWindow_->setContacts( interface_->contacts() );

    if( !soundNotifications_->isSupported() )
    { toggleSoundAction_->setEnabled( false ); }

    // configuration
    connect( this, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    emit configurationChanged();

    // center
    contactListWindow_->centerOnDesktop();

    // show, unless specified by options
    if( !(
        XmlOptions::get().get<bool>( "USE_SYSTEMTRAY" ) &&
        QSystemTrayIcon::isSystemTrayAvailable() &&
        XmlOptions::get().get<bool>( "START_HIDDEN" ) ) )
    { _toggleContactListVisibility(); }

    // contact list connection
    connect( contactListWindow_, SIGNAL(requestConnection()), this, SLOT(_connect()) );
    connect( contactListWindow_, SIGNAL(requestDisconnection()), this, SLOT(_disconnect()) );
    connect( contactListWindow_, SIGNAL(requestPingServer(bool)), interface_, SLOT(pingServer(bool)) );
    connect( contactListWindow_, SIGNAL(requestPingClients(ConnectionId::List,bool)), interface_, SLOT(pingClients(ConnectionId::List,bool)) );
    connect( contactListWindow_, SIGNAL(contactChanged(ConnectionId,ConnectionId)), interface_, SLOT(updateContact(ConnectionId,ConnectionId)) );
    connect( contactListWindow_, SIGNAL(contactRemoved(ConnectionId)), interface_, SLOT(removeContact(ConnectionId)) );
    connect( contactListWindow_, SIGNAL(newTextNotification(QString)), this, SLOT(_notify(QString)) );
    connect( contactListWindow_, SIGNAL(newSoundNotification(SoundNotifications::Type)), this, SLOT(_notify(SoundNotifications::Type)) );

    connect( &contactListWindow_->fetchContactsAction(), SIGNAL(triggered()), interface_, SLOT(fetchContacts()) );

    // interface connections to contact list
    connect( interface_, SIGNAL(contactChanged(ConnectionId,ConnectionId)), contactListWindow_, SLOT(updateContact(ConnectionId,ConnectionId)) );
    connect( interface_, SIGNAL(avatarChanged(ConnectionId)), contactListWindow_, SLOT(updateAvatar(ConnectionId)) );
    connect( interface_, SIGNAL(clearContactList()), contactListWindow_, SLOT(clearContactList()) );
    connect( interface_, SIGNAL(busyStateChanged(bool,int)), contactListWindow_, SLOT(updateBusyState(bool,int)) );
    connect( interface_, SIGNAL(messageRecieved(ChatMessage)), contactListWindow_, SLOT(notifyNewMessage(ChatMessage)) );
    connect( interface_, SIGNAL(fileTransferRequested(ConnectionId,BaseFileInfo)), contactListWindow_, SLOT(notifyFileTransferRequest(ConnectionId,BaseFileInfo)) );
    connect( interface_, SIGNAL(fileTransferCancelled(ConnectionId,BaseFileInfo)), contactListWindow_, SLOT(notifyFileTransferCancelled(ConnectionId,BaseFileInfo)) );
    connect( interface_, SIGNAL(error(QString,bool)), contactListWindow_, SLOT(notifyError(QString,bool)) );
    connect( interface_, SIGNAL(error(QString,bool)), this, SLOT(_notifyError(QString)) );

    connect( &sshInterface_->connection(), SIGNAL(message(QString)), &contactListWindow_->statusBar().label(0), SLOT(setTextAndUpdate(QString)) );
    connect( sshInterface_, SIGNAL(error(QString)), this, SLOT(_notifyError(QString)) );

    // fetch local messages
    interface_->fetchLocalMessages();

    // connect interfaces
    if( XmlOptions::get().get<bool>( "AUTO_CONNECT" ) )
    { _connect(); }

    Debug::Throw( "ClientApplication::realizeWidget - done.\n" );
    return true;

}

//____________________________________________
ConversationWindow* ClientApplication::createConversationWindow( ConnectionId id, ConnectionId::List contacts )
{

    ConversationWindow* conversationWindow = new ConversationWindow();
    conversationWindow->setConnectionId( id );
    conversationWindow->setContacts( contacts );

    conversationWindow->centerOnDesktop();
    conversationWindow->installEventFilter( contactListWindow_ );

    // conversationWindow connections
    connect( conversationWindow, SIGNAL(newMessage(ChatMessage)), interface_, SLOT(sendMessage(ChatMessage)) );
    connect( conversationWindow, SIGNAL(newTextNotification(QString)), this, SLOT(_notify(QString)) );
    connect( conversationWindow, SIGNAL(newSoundNotification(SoundNotifications::Type)), this, SLOT(_notify(SoundNotifications::Type)) );
    connect( conversationWindow, SIGNAL(userTyping(QStringList)), interface_, SLOT(sendUserTyping(QStringList)) );

    // messages
    connect( conversationWindow, SIGNAL(requestMessages(QStringList)), interface_, SLOT(fetchMessages(QStringList)) );
    connect( conversationWindow, SIGNAL(requestMessages(QStringList,HistoryMenu::Delay)), interface_, SLOT(fetchMessages(QStringList,HistoryMenu::Delay)) );

    connect( conversationWindow, SIGNAL(requestLocalMessages(QStringList)), interface_, SLOT(fetchLocalMessages(QStringList)) );
    connect( conversationWindow, SIGNAL(requestLocalMessages(QStringList,HistoryMenu::Delay)), interface_, SLOT(fetchLocalMessages(QStringList,HistoryMenu::Delay)) );

    // file transfer
    connect( conversationWindow, SIGNAL(requestFileTransfer(QStringList,BaseFileInfo)), interface_, SLOT(requestFileTransfer(QStringList,BaseFileInfo)) );
    connect( conversationWindow, SIGNAL(acceptFileTransfer(ConnectionId,BaseFileInfo,File)), interface_, SLOT(acceptFileTransfer(ConnectionId,BaseFileInfo,File)) );
    connect( conversationWindow, SIGNAL(denyFileTransfer(ConnectionId,BaseFileInfo)), interface_, SLOT(denyFileTransfer(ConnectionId,BaseFileInfo)) );
    connect( &conversationWindow->fileTransferManager(), SIGNAL(fileTransferCancelled(BaseFileInfo)), interface_, SLOT(cancelFileTransfer(BaseFileInfo)) );

    // interface connections
    connect( interface_, SIGNAL(contactChanged(ConnectionId,ConnectionId)), conversationWindow, SLOT(updateContact(ConnectionId,ConnectionId)) );
    connect( interface_, SIGNAL(avatarChanged(ConnectionId)), conversationWindow, SLOT(updateAvatar(ConnectionId)) );
    connect( interface_, SIGNAL(clearContactList()), conversationWindow, SLOT(clearContactList()) );
    connect( interface_, SIGNAL(userTyping(ConnectionId)), conversationWindow, SLOT(notifyUserTyping(ConnectionId)));
    connect( interface_, SIGNAL(clearMessageList(QStringList)), conversationWindow, SLOT(clearMessageList(QStringList)) );
    connect( interface_, SIGNAL(messageRecieved(ChatMessage)), conversationWindow, SLOT(printMessage(ChatMessage)) );
    connect( interface_, SIGNAL(busyStateChanged(bool,int)), conversationWindow, SLOT(updateBusyState(bool,int)) );
    connect( &sshInterface_->connection(), SIGNAL(message(QString)), &conversationWindow->statusBar().label(0), SLOT(setTextAndUpdate(QString)) );

    // contact list connections
    connect( contactListWindow_, SIGNAL(contactChanged(ConnectionId,ConnectionId)), conversationWindow, SLOT(updateContact(ConnectionId,ConnectionId)) );

    // associate to this using base key
    Base::Key::associate( conversationWindow, this );

    // setup recipients
    QStringList recipients;
    for( const auto& contact:contacts )
    { recipients.append( contact.user() ); }

    // setup delay
    const HistoryMenu::Delay delay = (HistoryMenu::Delay) XmlOptions::get().get<int>( "DEFAULT_HISTORY_SPAN" );

    // fetch messages
    if( id.connectionState() == ConnectionId::Connected ) interface_->fetchMessages( recipients, delay );
    else interface_->fetchLocalMessages( recipients, delay );

    return conversationWindow;

}

//____________________________________________
ConversationWindow* ClientApplication::findConversationWindow( ConnectionId id, ConnectionId::List contacts ) const
{
    for( const auto& conversationWindow:Base::KeySet<ConversationWindow>(this) )
    {
        if( conversationWindow->connectionId() != id ) continue;

        bool valid( true );
        for( const auto& contact:contacts )
        {
            if( !conversationWindow->hasContact( contact ) )
            {
                valid = false;
                break;
            }
        }

        if( valid ) return conversationWindow;
    }

    return nullptr;
}

//____________________________________________
ConversationWindow* ClientApplication::findConversationWindow( ConnectionId id, QStringList contacts ) const
{
    for( const auto& conversationWindow:Base::KeySet<ConversationWindow>(this) )
    {
        if( conversationWindow->connectionId() != id ) continue;

        bool valid( true );
        for( const auto& contact:contacts )
        {
            if( !conversationWindow->hasContact( contact ) )
            {
                valid = false;
                break;
            }
        }

        if( valid ) return conversationWindow;
    }

    return nullptr;
}

//____________________________________________
CommandLineParser ClientApplication::commandLineParser( CommandLineArguments arguments, bool ignoreWarnings ) const
{
    Debug::Throw( "ClientApplication::commandLineParser.\n" );
    CommandLineParser out( BaseApplication::commandLineParser() );
    out.setGroup( CommandLineParser::applicationGroupName );
    out.registerFlag( CommandLineParser::Tag( "--no-ssh" ), tr( "disable internal SSH tunneling to connect to server" ) );
    out.registerOption( CommandLineParser::Tag( "--chat-host" ), tr( "string" ), tr( "use specified host for chat communication" ) );
    out.registerOption( CommandLineParser::Tag( "--chat-port" ), tr( "integer" ), tr( "use specified port for server communication" ) );
    out.registerOption( CommandLineParser::Tag( "--user", "-u" ), tr( "string" ), tr( "user name" ) );
    out.registerFlag( CommandLineParser::Tag( "--verbose" ), tr( "enable debug messages" ) );

    if( !arguments.isEmpty() ) out.parse( arguments, ignoreWarnings );
    return out;

}

//____________________________________________
void ClientApplication::usage( void ) const
{
    _usage( "homechat" );
    commandLineParser().usage();
    return;
}

//_______________________________________________
void ClientApplication::_configuration( void )
{
    Debug::Throw( "ClientApplication::_configuration.\n" );
    emit saveConfiguration();
    ClientConfigurationDialog dialog( nullptr );
    dialog.setTextNotificationsSupported( systemNotifications_->isSupported() );
    dialog.setSoundNotificationsSupported( soundNotifications_->isSupported() );
    dialog.realizeWidget();
    connect( &dialog, SIGNAL(configurationChanged()), SIGNAL(configurationChanged()) );
    dialog.centerOnWidget( qApp->activeWindow() );

    if( dialog.exec() )
    {
        dialog.avatarSelectionWidget().write();
        emit avatarChanged();
    }

}

//_______________________________________________
void ClientApplication::_spellCheckConfiguration( void )
{

    #if USE_ASPELL

    Debug::Throw( "Application::_spellCheckConfiguration.\n" );

    // create dialog
    CustomDialog dialog;
    dialog.setWindowTitle( "Spell Check Configuration - Qedit" );

    SpellCheck::SpellCheckConfiguration* spellConfig = new SpellCheck::SpellCheckConfiguration( &dialog, SpellCheck::SpellCheckConfiguration::ASpell | SpellCheck::SpellCheckConfiguration::Dictionary );
    dialog.mainLayout().addWidget( spellConfig );

    SpellCheck::AutoSpellConfiguration* autospellConfig = new SpellCheck::AutoSpellConfiguration( &dialog );
    dialog.mainLayout().addWidget( autospellConfig );
    dialog.centerOnWidget( qApp->activeWindow() );

    if( dialog.exec() == QDialog::Rejected ) return;
    spellConfig->write( XmlOptions::get() );
    autospellConfig->write( XmlOptions::get() );
    XmlOptions::write();

    emit spellCheckConfigurationChanged();

    #endif

}

//________________________________________________
void ClientApplication::_notifyError( QString message )
{ Debug::Throw(0) << "ClientApplication::_notifyError - " << message << endl; }

//________________________________________________
bool ClientApplication::_processCommand( Server::ServerCommand command )
{

    Debug::Throw( "ClientApplication::_processCommand.\n" );
    if( BaseApplication::_processCommand( command ) ) return true;
    if( command.command() == Server::ServerCommand::Raise )
    {
        contactListWindow_->uniconify();
        return true;

    } else return false;

}

//_________________________________________________
void ClientApplication::_exit( void )
{
    Debug::Throw( "ClientApplication::_exit.\n" );
    interface_->writeContacts();
    qApp->quit();
}

//_________________________________________________
void ClientApplication::_toggleSoundNotifications( bool checked )
{ XmlOptions::get().set<bool>( "USE_SOUND_NOTIFICATIONS", checked ); }

//_______________________________________________
void ClientApplication::_updateConfiguration( void )
{

    Debug::Throw( "ClientApplication::_updateConfiguration.\n" );

    // test notifications
    useSystemNotifications_ = XmlOptions::get().get<bool>( "USE_NOTIFICATIONS" );

    // sound notification mask
    toggleSoundAction_->setChecked( XmlOptions::get().get<bool>( "USE_SOUND_NOTIFICATIONS" ) );
    soundNotifications_->setNotificationMask( XmlOptions::get().get<int>( "SOUND_NOTIFICATION_MASK" ) );

    // system tray icon
    if( XmlOptions::get().get<bool>( "USE_SYSTEMTRAY" ) )
    {

        qApp->setQuitOnLastWindowClosed( false );
        if( !systemTrayIcon_ )
        {

            systemTrayIcon_ = new SystemTrayIcon();
            systemTrayIcon_->setConnectionId( interface_->connectionId() );
            systemTrayIcon_->setContacts( interface_->contacts() );

            // connections
            connect( systemTrayIcon_, SIGNAL( activated(QSystemTrayIcon::ActivationReason) ), SLOT( _systemTrayActivated(QSystemTrayIcon::ActivationReason)) );
            connect( contactListWindow_, SIGNAL(contactChanged(ConnectionId,ConnectionId)), systemTrayIcon_, SLOT(updateContact(ConnectionId,ConnectionId)) );
            connect( contactListWindow_, SIGNAL(contactRemoved(ConnectionId)), systemTrayIcon_, SLOT(removeContact(ConnectionId)) );
            connect( interface_, SIGNAL(contactChanged(ConnectionId,ConnectionId)), systemTrayIcon_, SLOT(updateContact(ConnectionId,ConnectionId)) );
            connect( contactListWindow_, SIGNAL(hasPendingMessages(bool)), systemTrayIcon_, SLOT(updatePendingMessages(bool)) );

        }

        systemTrayIcon_->show();

    } else {

        qApp->setQuitOnLastWindowClosed( true );
        if( systemTrayIcon_ ) systemTrayIcon_->deleteLater();
        systemTrayIcon_ = nullptr;

    }

    // ssh connection
    if( XmlOptions::get().get<bool>( "USE_SSH" ) && !commandLineParser( _arguments() ).hasFlag( "--no-ssh" ) )
    {

        connect( sshInterface_, SIGNAL(connected()), interface_, SLOT(connect()) );

    } else {

        disconnect( sshInterface_, SIGNAL(connected()), 0, 0 );

    }

}

//_______________________________________________
void ClientApplication::_connect( void )
{

    Debug::Throw( "ClientApplication::_connect.\n" );

    // unset try reconnect flag
    interface_->setTryReconnect(false);

    // ssh connection
    if( XmlOptions::get().get<bool>( "USE_SSH" ) && !commandLineParser( _arguments() ).hasFlag( "--no-ssh" ) )
    {

        interface_->disconnect();
        sshInterface_->disconnect();
        sshInterface_->connect();

    } else {

        // make sure ssh interface is disconnected
        sshInterface_->disconnect();

        // re-connect tcp interface
        interface_->disconnect();
        interface_->connect();

    }

}

//_______________________________________________
void ClientApplication::_disconnect( void )
{

    Debug::Throw( "ClientApplication::_disconnect.\n" );

    // unset try reconnect flag
    interface_->setTryReconnect(false);

    // disconnect tcp interface
    interface_->disconnect();

    // disconnect ssh interface
    if( XmlOptions::get().get<bool>( "USE_SSH" ) && !commandLineParser( _arguments() ).hasFlag( "--no-ssh" ) )
    { sshInterface_->disconnect(); }

}

//_______________________________________________
void ClientApplication::_notify( QString message )
{

    Debug::Throw( "ClientApplication::_notify (text).\n" );
    if( !useSystemNotifications_ ) return;
    else if( systemNotifications_->isSupported() )
    {

        // try system notifications
        Notification notification( "Homechat", message );
        notification.setActionList( QStringList() << "open" << tr( "Open" ) );
        systemNotifications_->sendNotification( notification );

    } else if( QSystemTrayIcon::supportsMessages() && QSystemTrayIcon::isSystemTrayAvailable() && systemTrayIcon_ ) {

        // try system tray messages
        systemTrayIcon_->showMessage( "Homechat", message );

    }

}

//_______________________________________________
void ClientApplication::_notify( SoundNotifications::Type type )
{

    Debug::Throw( "ClientApplication::_notify (sound).\n" );
    if( toggleSoundAction_->isEnabled() && toggleSoundAction_->isChecked() )
    { soundNotifications_->play( type ); }

}

//_______________________________________________
void ClientApplication::_systemTrayActivated( QSystemTrayIcon::ActivationReason reason )
{
    Debug::Throw( "ClientApplication::_systemTrayActivated.\n" );
    switch( reason )
    {

        default: break;

        case QSystemTrayIcon::Trigger:
        _toggleContactListVisibility();
        break;

    }
}

//________________________________________________
void ClientApplication::_raiseContactList( void )
{
    Debug::Throw() << "ClientApplication::_raiseContactList." << endl;
    contactListAction_->setText( tr( "Minimize" ) );
    contactListWindow_->uniconify();
}

//________________________________________________
void ClientApplication::_toggleContactListVisibility( void )
{
    Debug::Throw( "ClientApplication::_toggleContactListVisibility.\n" );

    if( contactListWindow_->isHidden() || contactListWindow_->isMinimized() )
    {

        contactListAction_->setText( tr( "Minimize" ) );
        contactListWindow_->uniconify();

    } else {

        contactListAction_->setText( tr( "Restore" ) );
        contactListWindow_->hide();

    }

}

//_______________________________________________
void ClientApplication::_installActions( void )
{
    Debug::Throw( "ClientApplication::_installActions.\n" );

    // spell check
    spellCheckConfigurationAction_ = new QAction( IconEngine::get( IconNames::SpellCheck ), tr( "Configure Spell Checking..." ), this );
    connect( spellCheckConfigurationAction_, SIGNAL(triggered()), SLOT(_spellCheckConfiguration()) );

    // raise main window
    contactListAction_ = new QAction( tr( "Restore" ), this );
    connect( contactListAction_, SIGNAL(triggered()), SLOT(_toggleContactListVisibility()) );

    // toggle sound
    toggleSoundAction_ = new QAction( IconEngine::get( IconNames::Sound ), tr( "Use Sound Notifications" ), this );
    connect( toggleSoundAction_, SIGNAL(toggled(bool)), this, SLOT(_toggleSoundNotifications(bool)) );
    toggleSoundAction_->setCheckable( true );
    toggleSoundAction_->setChecked( false );


}
