#ifndef ClientApplication_h
#define ClientApplication_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "BaseApplication.h"
#include "Config.h"
#include "CommandLineArguments.h"
#include "ConnectionId.h"
#include "Counter.h"
#include "Key.h"
#include "PixmapEngine.h"
#include "SoundNotifications.h"

#include <QSystemTrayIcon>

class ClientInterface;
class ContactListWindow;
class FileTransferManager;
class ConversationWindow;
class SystemNotifications;
class SystemTrayIcon;

namespace Ssh
{
    class Interface;
}

//* Main Window singleton object
class ClientApplication: public BaseApplication, public Counter, public Base::Key
{

    //* Qt meta object declaration
    Q_OBJECT

    public:

    //* constructor
    ClientApplication( CommandLineArguments );

    //*@name modifier
    //@{

    //* create all widgets
    virtual bool realizeWidget( void );

    //* create a new Main window
    ConversationWindow* createConversationWindow( ConnectionId, ConnectionId::List );

    //@}

    //*@name accessors
    //@{

    //* contact list window
    ContactListWindow& contactListWindow( void ) const
    { return *contactListWindow_; }

    //* find main window matching given connection id and contacts
    ConversationWindow* findConversationWindow( ConnectionId, ConnectionId::List ) const;

    //* find main window matching given connection id and contacts
    ConversationWindow* findConversationWindow( ConnectionId, QStringList ) const;

    //* command line parser
    CommandLineParser commandLineParser( CommandLineArguments = CommandLineArguments(), bool ignoreWarnings = true ) const;

    //* usage
    void usage( void ) const;

    //* application name
    virtual QString applicationName( void ) const
    { return "Homechat"; }

    //* application icon
    virtual QIcon applicationIcon( void ) const
    {
        QIcon icon;
        icon.addPixmap( PixmapEngine::get( ":/homechat.png" ) );
        icon.addPixmap( PixmapEngine::get( ":/homechat-large.png" ) );
        return icon;
    }

    // application version
    virtual QString applicationVersion( void ) const
    { return VERSION; }

    //@}

    //*@name actions
    //@{

    //* main window action
    /* used to raise main window from system tray icon */
    QAction& contactListAction( void ) const
    { return *contactListAction_; }

    //* spellcheck configuration
    QAction& spellCheckConfigurationAction( void ) const
    { return *spellCheckConfigurationAction_; }

    //* sound notification action
    QAction& toggleSoundAction( void ) const
    { return *toggleSoundAction_; }

    //@}

    Q_SIGNALS:

    //* spellcheck configuration modified
    void spellCheckConfigurationChanged( void );

    //* avatar modified
    void avatarChanged( void );

    protected Q_SLOTS:

    //* configuration
    virtual void _configuration( void );

    //* spellcheck configuration
    void _spellCheckConfiguration( void );

    //* error notification
    virtual void _notifyError( QString );

    //* process request from application manager
    virtual bool _processCommand( Server::ServerCommand );

    //* exit safely
    void _exit( void );

    //* toggle sound notifications
    void _toggleSoundNotifications( bool );

    private Q_SLOTS:

    //* configuration
    void _updateConfiguration( void );

    //* connect
    virtual void _connect( void );

    //* disconnect
    virtual void _disconnect( void );

    //* text notification
    void _notify( QString );

    //* sound notification
    void _notify( SoundNotifications::Type );

    //* system tray icon activated
    void _systemTrayActivated( QSystemTrayIcon::ActivationReason );

    //* raise main window
    void _raiseContactList( void );

    //* toggle main window visibility
    void _toggleContactListVisibility( void );

    private:

    //* install actions
    void _installActions( void );

    //* ssh connection
    Ssh::Interface* sshInterface_ = nullptr;

    //* communication interface
    ClientInterface* interface_ = nullptr;

    //* contact list window
    ContactListWindow* contactListWindow_ = nullptr;

    //* system notifications
    bool useSystemNotifications_ = true;

    //* system notifications
    SystemNotifications* systemNotifications_ = nullptr;

    //* sound notifications
    SoundNotifications* soundNotifications_ = nullptr;

    //* system tray icon
    SystemTrayIcon* systemTrayIcon_ = nullptr;

    //*@name actions
    //@{

    //* spellcheck configuration
    QAction* spellCheckConfigurationAction_ = nullptr;

    //* main window visibility action
    QAction* contactListAction_ = nullptr;

    //* sound notification actions
    QAction* toggleSoundAction_ = nullptr;

    //@}

};

#endif
