/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ClientConfigurationDialog.h"

#include "AvatarSelectionWidget.h"
#include "AvatarPixmap.h"
#include "Debug.h"
#include "GridLayout.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "IconSize.h"
#include "OptionBrowsedLineEditor.h"
#include "OptionCheckBox.h"
#include "OptionComboBox.h"
#include "OptionLineEditor.h"
#include "OptionRadioButton.h"
#include "OptionSpinBox.h"
#include "SoundNotificationsOptionWidget.h"
#include "SshConnection.h"
#include "SshConnectionWidget.h"
#include "SystemNotifications.h"

#include <QButtonGroup>
#include <QGroupBox>
#include <QLabel>
#include <QLayout>
#include <QPushButton>
#include <QSystemTrayIcon>

namespace Ssh
{
    class OptionConnectionWidget: public ConnectionWidget, public OptionWidget
    {

        public:

        //* constructor
        OptionConnectionWidget( QWidget* parent ):
            ConnectionWidget( parent ),
            OptionWidget( "generic" )
        { hidePasswordEditor(); }

        //* read
        void read( const Options& options )
        {

            // read attributes from options and assign
            ConnectionAttributes attributes;
            attributes.setHost( options.raw( "SSH_HOST" ) );
            attributes.setPort( options.get<int>( "SSH_PORT" ) );
            attributes.setUserName( options.raw( "SSH_USERNAME" ) );
            attributes.setRememberPassword( options.get<bool>("SSH_REMEMBER_PASSWORD" ) );

            setAttributes( attributes );
        }

        //* write
        void write( Options& options ) const
        {
            auto attributes( this->attributes() );
            options.setRaw( "SSH_HOST", attributes.host() );
            options.set<int>( "SSH_PORT", attributes.port() );
            options.setRaw( "SSH_USERNAME", attributes.userName() );
            options.set<bool>( "SSH_REMEMBER_PASSWORD", attributes.rememberPassword() );
        }

    };

}

//_________________________________________________________
ClientConfigurationDialog::ClientConfigurationDialog( QWidget* parent ):
    BaseConfigurationDialog( parent )
{

    Debug::Throw( "ClientConfigurationDialog::ClientConfigurationDialog.\n" );
    setWindowTitle( tr( "Configuration - Homechat" ) );
}

//_________________________________________________________
void ClientConfigurationDialog::realizeWidget( void )
{

    // base configuration
    baseConfiguration( nullptr, Base);
    textEditConfiguration( nullptr, ParagraphHighlight|TextEditionFlags );

    QWidget *box = nullptr;
    OptionLineEditor* editor = nullptr;
    OptionCheckBox* checkbox = nullptr;
    OptionSpinBox* spinbox = nullptr;

    {
        // identity
        QWidget* page = &addPage( IconEngine::get( IconNames::PreferencesUser ), tr( "Identity" ) );

        {
            page->layout()->addWidget( box = new QWidget( page ) );

            QGridLayout* gridLayout = new QGridLayout();
            gridLayout->setSpacing(5);
            gridLayout->setMargin(5);
            box->setLayout( gridLayout );

            gridLayout->setColumnStretch(0, 0);
            gridLayout->addWidget( avatarSelectionWidget_ = new AvatarSelectionWidget( box ), 0, 0, 5, 1 );
            avatarSelectionWidget_->read();

            QLabel* label = nullptr;
            gridLayout->addWidget(  label = new QLabel( tr( "User name:" ), box ), 0, 1, 1, 1 );
            gridLayout->addWidget( editor = new OptionLineEditor( box, "USERNAME" ), 1, 1, 1, 1 );
            label->setBuddy( editor );
            addOptionWidget( editor );

            gridLayout->addWidget( &avatarSelectionWidget_->changeAvatarButton( box ), 2, 1, 1, 1 );
            gridLayout->addWidget( &avatarSelectionWidget_->resetAvatarButton( box ), 3, 1, 1, 1 );

        }

    }

    {
        // notifications
        QWidget* page = &addPage( IconEngine::get( IconNames::PreferencesNotifications ), tr( "Notifications" ), tr( "Event notifications" ) );
        page->layout()->addWidget( box = new QWidget( page ) );
        box->setLayout( new QVBoxLayout() );
        box->layout()->setMargin(0);

        OptionCheckBox* checkbox;
        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Use text notifications" ), box, "USE_NOTIFICATIONS" ) );
        if( textNotificationsSupported_ ) addOptionWidget( checkbox );
        else checkbox->setEnabled( false );

        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Print a message when remote user is typing" ), page, "NOTIFY_USER_TYPING" ) );
        addOptionWidget( checkbox );

        box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Use sound notifications" ), box, "USE_SOUND_NOTIFICATIONS" ) );
        if( soundNotificationsSupported_ ) addOptionWidget( checkbox );
        else checkbox->setEnabled( false );

        QFrame* frame = new QFrame();
        frame->setFrameStyle( QFrame::HLine );
        box->layout()->addWidget( frame );

        box->layout()->addWidget( new QLabel( tr( "Select the events for which a sound notification should be sent:" ), box ) );
        SoundNotificationsOptionWidget* widget = new SoundNotificationsOptionWidget( box, "SOUND_NOTIFICATION_MASK" );
        box->layout()->addWidget( widget );
        addOptionWidget( widget );
    }

    {
        // appearance
        QWidget* page = &addPage( IconEngine::get( IconNames::PreferencesAppearance ), tr( "Appearance" ), tr( "Application's appearance" ) );

        {

            page->layout()->addWidget( box = new QGroupBox( tr( "Contact List" ), page ) );
            QVBoxLayout* vLayout = new QVBoxLayout();
            box->setLayout( vLayout );

            box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Use flat contact list" ), page, "FLAT_CONTACT_LIST" ) );
            addOptionWidget( checkbox );

            box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Show tooltips" ), page, "SHOW_TOOLTIPS" ) );
            addOptionWidget( checkbox );

            QHBoxLayout* hLayout = new QHBoxLayout();
            vLayout->addItem( hLayout );

            QLabel* label;
            hLayout->addWidget( label = new QLabel( tr( "Contact list orientation: " ), box ) );
            label->setAlignment( Qt::AlignVCenter|Qt::AlignRight );
            QButtonGroup* buttonGroup = new QButtonGroup( this );
            buttonGroup->setExclusive(true);

            OptionRadioButton* radioButton = new OptionRadioButton( tr( "Horizontal" ), box, "HORIZONTAL_CONTACT_LIST" );
            buttonGroup->addButton( radioButton );
            hLayout->addWidget( radioButton );
            addOptionWidget( radioButton );

            QRadioButton* radioButton2 = new QRadioButton( tr( "Vertical" ), box );
            buttonGroup->addButton( radioButton2 );
            hLayout->addWidget( radioButton2 );
            radioButton2->setChecked(true);
            hLayout->addStretch();

        }

        {
            page->layout()->addWidget( box = new QGroupBox( tr( "Emoticons" ), page ) );
            QVBoxLayout* vLayout = new QVBoxLayout();
            box->setLayout( vLayout );

            OptionCheckBox* checkbox;
            vLayout->addWidget( checkbox = new OptionCheckBox( tr( "Use emoticons" ), page, "USE_EMOTICONS" ) );
            addOptionWidget( checkbox );

            vLayout->addWidget( checkbox = new OptionCheckBox( tr( "Use colored emoticons" ), page, "COLORIZE_EMOTICONS" ) );
            checkbox->setNegative( true );
            addOptionWidget( checkbox );

            QHBoxLayout* hLayout = new QHBoxLayout();
            hLayout->setMargin( 0 );
            vLayout->addItem( hLayout );

            QLabel* label;
            hLayout->addWidget( label = new QLabel( tr( "Emoticon icon size:" ), box ) );
            hLayout->addWidget( spinbox = new OptionSpinBox( box, "EMOTICON_SIZE" ) );
            spinbox->setMinimum(8);
            spinbox->setMaximum(IconSize::Huge);
            spinbox->setSuffix( tr( " px" ) );
            label->setBuddy( spinbox );
            addOptionWidget( spinbox );

        }

        {

            page->layout()->addWidget( box = new QWidget( page ) );
            QVBoxLayout* vLayout = new QVBoxLayout();
            vLayout->setMargin(0);
            box->setLayout( vLayout );

            OptionCheckBox* checkbox;

            box->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Outline html links" ), page, "SHOW_LINKS" ) );
            addOptionWidget( checkbox );
        }

    }

    {
        // behavior
        QWidget* page = &addPage( IconEngine::get( IconNames::PreferencesAppearance ), tr( "Behavior" ), tr( "Application's behavior" ) );

        {
            page->layout()->addWidget( box = new QGroupBox( tr( "Startup" ), page ) );

            QVBoxLayout* vLayout = new QVBoxLayout();
            box->setLayout( vLayout );
            QLabel* label = new QLabel( tr( "At startup, fetch history for:" ), box );
            OptionComboBox* comboBox;
            vLayout->addWidget( label );
            vLayout->addWidget( comboBox = new OptionComboBox( box, "DEFAULT_HISTORY_SPAN" ) );
            comboBox->setUseValue( false );

            // must be kept in sync with HistoryMenu::Delay enumeration
            comboBox->addItems( QStringList()
                << tr( "Today" )
                << tr( "Yesterday" )
                << tr( "One Week" )
                << tr( "Two Weeks" )
                << tr( "On Month" )
                << tr( "Six Months" )
                << tr( "One Year" )
                << tr( "From Beginning" ) );
            label->setBuddy( comboBox );
            addOptionWidget( comboBox );

            vLayout->addWidget( checkbox = new OptionCheckBox( tr( "Auto-connect at startup" ), box, "AUTO_CONNECT" ) );
            addOptionWidget( checkbox );

            OptionCheckBox* sysTrayCheckBox = nullptr;
            vLayout->addWidget( sysTrayCheckBox = new OptionCheckBox( tr( "Use system tray" ), box, "USE_SYSTEMTRAY" ) );
            if( QSystemTrayIcon::isSystemTrayAvailable() ) addOptionWidget( sysTrayCheckBox );
            else sysTrayCheckBox->setEnabled( false );

            if( QSystemTrayIcon::isSystemTrayAvailable() )
            {
                vLayout->addWidget( checkbox = new OptionCheckBox( tr( "Hide window on startup" ), box, "START_HIDDEN" ) );
                addOptionWidget( checkbox );
                checkbox->setEnabled( false );
                connect( sysTrayCheckBox, SIGNAL(toggled(bool)), checkbox, SLOT(setEnabled(bool)) );
            }

        }

        {
            page->layout()->addWidget( box = new QGroupBox( tr( "File transfer" ), page ) );

            QVBoxLayout* vLayout = new QVBoxLayout();
            box->setLayout( vLayout );

            QLabel* label;
            vLayout->addWidget( label = new QLabel( tr( "Directory for saving transfered files:" ), box ) );

            OptionBrowsedLineEditor* editor;
            vLayout->addWidget( editor = new OptionBrowsedLineEditor( box, "DOWNLOAD_LOCATION" ) );
            editor->setFileMode( QFileDialog::Directory );
            editor->editor().setReadOnly( true );
            addOptionWidget( editor );
            label->setBuddy( editor );
        }

        {
            page->layout()->addWidget( box = new QGroupBox( tr( "Timeouts" ), box ) );

            // history span
            GridLayout* gridLayout = new GridLayout();
            gridLayout->setMaxCount(2);
            gridLayout->setSpacing(5);
            gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
            box->setLayout( gridLayout );

            QLabel* label;
            gridLayout->addWidget( label = new QLabel( tr( "Mark away when inactive for:" ), box ) );
            gridLayout->addWidget( spinbox = new OptionSpinBox( box, "IDLE_TIMEOUT" ) );
            spinbox->setMinimum(1);
            spinbox->setMaximum(60);
            spinbox->setSuffix( tr( " minutes" ) );
            label->setBuddy( spinbox );
            addOptionWidget( spinbox );

            gridLayout->addWidget( label = new QLabel( tr( "Check connections to server every:" ), box ) );
            gridLayout->addWidget( spinbox = new OptionSpinBox( box, "CHECK_CONNECTIONS_TIMEOUT" ) );
            spinbox->setMinimum(0);
            spinbox->setMaximum(60);
            spinbox->setSuffix( tr( " minutes" ) );
            spinbox->setSpecialValueText( tr( "Never" ) );
            label->setBuddy( spinbox );
            addOptionWidget( spinbox );

            gridLayout->addWidget( label = new QLabel( tr( "Try reconnect to server every:" ), box ) );
            gridLayout->addWidget( spinbox = new OptionSpinBox( box, "RECONNECT_TIMEOUT" ) );
            spinbox->setMinimum(0);
            spinbox->setMaximum(60);
            spinbox->setSuffix( tr( " minutes" ) );
            spinbox->setSpecialValueText( tr( "Never" ) );
            label->setBuddy( spinbox );
            addOptionWidget( spinbox );

        }

    }

    {
        // server configuration
        QWidget* page = &addPage( IconEngine::get( IconNames::PreferencesServer ), tr( "Server" ), tr( "Server configuration" ) );
        GridLayout* gridLayout;
        QLabel* label;

        // use ssh
        if( Ssh::Connection::isSupported() )
        {
            OptionCheckBox* checkbox;
            page->layout()->addWidget( checkbox = new OptionCheckBox( tr( "Connect to server using ssh tunnel" ), box, "USE_SSH" ) );
            addOptionWidget( checkbox );

            // ssh connection
            page->layout()->addWidget( box = new QGroupBox( tr( "Ssh Connection Settings" ), page ) );
            box->setLayout( new QVBoxLayout() );
            Ssh::OptionConnectionWidget* connectionWidget = new Ssh::OptionConnectionWidget( box );
            box->layout()->addWidget( connectionWidget );
            addOptionWidget( connectionWidget );

            box->setEnabled( false );
            connect( checkbox, SIGNAL(toggled(bool)), box, SLOT(setEnabled(bool)) );
        }

        // server connection
        page->layout()->addWidget( box = new QGroupBox( tr( "Server Connection Settings" ), page ) );
        gridLayout = new GridLayout();
        gridLayout->setMaxCount(4);
        gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
        box->setLayout( gridLayout );

        // host
        gridLayout->addWidget( label = new QLabel( tr( "Host:" ), box ) );
        gridLayout->addWidget( editor = new OptionLineEditor( box, "CHAT_HOST" ) );
        label->setBuddy( editor );
        addOptionWidget( editor );

        // port
        gridLayout->addWidget( label = new QLabel( tr( "Port:" ), box ) );
        gridLayout->addWidget( spinbox = new OptionSpinBox( box, "CHAT_PORT", false ) );
        spinbox->setMinimum(0);
        spinbox->setMaximum(10000);
        label->setBuddy( spinbox );
        addOptionWidget( spinbox );

        // timeout
        page->layout()->addWidget( box = new QGroupBox( tr( "Timeouts" ), page ) );
        gridLayout = new GridLayout();
        gridLayout->setMaxCount(2);
        gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
        box->setLayout( gridLayout );

        gridLayout->addWidget( label = new QLabel( tr( "Authorization timeout:" ), box ) );
        gridLayout->addWidget( spinbox = new OptionSpinBox( box, "AUTHORIZATION_TIMEOUT" ) );
        spinbox->setSuffix( tr( "s" ) );
        spinbox->setSpecialValueText( tr( " Never" ) );
        spinbox->setMinimum( 0 );
        spinbox->setMaximum( 60 );
        label->setBuddy( spinbox );
        addOptionWidget( spinbox );

        gridLayout->addWidget( label = new QLabel( tr( "Ping timeout:" ), box ) );
        gridLayout->addWidget( spinbox = new OptionSpinBox( box, "PING_TIMEOUT" ) );
        spinbox->setSuffix( tr( "s" ) );
        spinbox->setSpecialValueText( tr( " Never" ) );
        spinbox->setMinimum( 0 );
        spinbox->setMaximum( 60 );
        label->setBuddy( spinbox );
        addOptionWidget( spinbox );

    }

    // load initial configuration
    read();

}
