/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ClientInterface.h"

#include "Avatar.h"
#include "AvatarImage.h"
#include "ChatClient.h"
#include "ChatMessageThread.h"
#include "ClientApplication.h"
#include "FileEmissionClient.h"
#include "FileReceptionClient.h"
#include "FileTransferManager.h"
#include "FileTransferWidget.h"
#include "ConversationWindow.h"
#include "Singleton.h"
#include "XmlOptions.h"
#include "XmlDef.h"
#include "Util.h"

#include <QApplication>
#include <QDateTime>
#include <QDir>
#include <QHostInfo>
#include <QNetworkInterface>
#include <QTimerEvent>
#include <QTimer>

//____________________________________________
bool ClientInterface::SameFileInfoFTor::operator () (FileEmissionClient* client ) const
{ return client->fileInfo() == fileInfo_; }

//____________________________________________
bool ClientInterface::SameFileInfoFTor::operator () (FileReceptionClient* client ) const
{ return client->fileInfo() == fileInfo_; }

//____________________________________________
ClientInterface::ClientInterface( QObject* parent, const CommandLineParser& parser ):
    QObject( parent ),
    Counter( "ClientInterface" ),
    parser_( parser ),
    client_( new ChatClient( this ) ),
    messageThread_( new ChatMessageThread( this ) )
{

    // debug level
    if( parser_.hasFlag( "--verbose" ) ) debugLevel_ = 0;
    else debugLevel_ = 1;

    messages_.setDebugLevel( debugLevel_ );

    // connections
    connect( &client_->socket(), SIGNAL(error(QAbstractSocket::SocketError)), SLOT(_error(QAbstractSocket::SocketError)) );
    connect( &client_->socket(), SIGNAL(disconnected()), SLOT(_disconnected()) );
    connect( &client_->socket(), SIGNAL(connected()), SLOT(_connected()) );
    connect( client_, SIGNAL(commandAvailable(ChatCommand)), SLOT(_processCommand(ChatCommand)) );
    connect( client_, SIGNAL(avatarAvailable(Avatar)), SLOT(_processAvatar(Avatar)) );

    connect( messageThread_, SIGNAL(finished()), SLOT(_fetchLocalMessagesFromThread()) );

    // configuration
    auto application = Singleton::get().application<ClientApplication>();
    connect( application, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    connect( application, SIGNAL(avatarChanged()), SLOT(_sendAvatar()) );
    connect( qApp, SIGNAL(aboutToQuit()), SLOT(_saveConfiguration()) );
    _updateConfiguration();

}

//_________________________________________________
void ClientInterface::readContacts( void )
{

    // input file
    const File file( XmlOptions::get().raw( "RC_FILE" ) );

    // parse the file
    XmlDocument document;
    QFile qfile( file );
    if( !document.setContent( &qfile ) )
    {
        Debug::Throw() << document.error() << endl;
        return;
    }

    // look for relevant element
    QDomNodeList topNodes = document.elementsByTagName( Xml::ContactList );
    if( topNodes.isEmpty() )
    {
        Debug::Throw() <<  "ClientInterface::readContacts - top node not found" << endl;
        return;
    }

    contacts_.clear();
    for(QDomNode node = topNodes.at(0).firstChild(); !node.isNull(); node = node.nextSibling() )
    {
        QDomElement element = node.toElement();
        if( element.isNull() ) continue;

        // special options
        if( element.tagName() == Xml::Id )
        { contacts_.append( ConnectionId( element ) ); }

    }

    // make sure 'this' connection id is in list
    if( connectionId_.isValid() )
    {

        const int i = contacts_.indexOf( connectionId_ );
        if( i >= 0 ) contacts_.replace( i, connectionId_ );
        else contacts_.append( connectionId_ );

    }

}

//_________________________________________________
void ClientInterface::writeContacts( void )
{

    // input file
    const File file( XmlOptions::get().raw( "RC_FILE" ) );

    // create document and read
    XmlDocument document;
    {
        QFile qtfile( file );
        document.setContent( &qtfile );
    }

    // create top element
    QDomElement top = document.createElement( Xml::ContactList );

    // get contacts from model and append dom element
    for( auto contact:contacts_ )
    {
        // mark contact as offline and update disconnection time
        if( contact.connectionState() != ConnectionId::Disconnected )
        {
            contact.setConnectionState( ConnectionId::Disconnected );
            contact.setDisconnectionTime( TimeStamp::now() );
        }

        // append to document
        top.appendChild( contact.domElement( document ) );
    }

    document.replaceChild( top );
    QFile qfile( file );
    if( !qfile.open( QIODevice::WriteOnly ) ) return;
    qfile.write( document.toByteArray() );

}

//______________________________________
void ClientInterface::setTryReconnect( bool value )
{
    Debug::Throw() << "ClientInterface::setTryReconnect - " << value << endl;
    tryReconnect_ = value;

    // stop reconnection timer
    if( reconnectionTimer_.isActive() ) reconnectionTimer_.stop();

    // restart
    const int timeout( 1000*60*( XmlOptions::get().get<int>( "RECONNECT_TIMEOUT" ) ) );
    if( tryReconnect_ && timeout > 0 ) reconnectionTimer_.start( timeout, this );

}

//_____________________________________________________
void ClientInterface::connect( void )
{
    Debug::Throw() << "ClientInterface::connect - " << host_.toString() << ":" << port_ << endl;
    client_->socket().abort();
    client_->socket().connectToHost( host_, port_ );
}

//_____________________________________________________
void ClientInterface::disconnect( void )
{
    Debug::Throw("ClientInterface::disconnect.\n" );
    client_->socket().disconnectFromHost();
    _setConnectionState( ConnectionId::Disconnected );
}

//____________________________________________
void ClientInterface::updateContact( ConnectionId oldConnectionId, ConnectionId connectionId )
{
    Debug::Throw( "ClientInterface::updateContact.\n" );

    // update this connection id
    if( oldConnectionId == connectionId_ )
    {
        // store old and new user state
        const auto oldUserState( connectionId_.userState() );
        const auto userState( connectionId.userState() );

        // update connection id
        connectionId_ = connectionId;

        // check user state changed
        if( userState != oldUserState )
        {
            // store in options
            if( userState != ConnectionId::Away )
            { XmlOptions::get().set<int>( QString( "USER_STATE_%1" ).arg( connectionId_.user() ), userState ); }

            // emit command
            client_->sendCommand( ChatCommand( connectionId_, ChatCommand::UserStateChanged ) );
        }

    }

    // update contact list
    const int i = contacts_.indexOf( oldConnectionId );
    if( i >= 0 ) contacts_.replace( i, connectionId );
    else contacts_.append( connectionId );

}

//_________________________________________________
void ClientInterface::removeContact( ConnectionId contact )
{
    Debug::Throw( "SystemTrayIcon::removeContact.\n" );
    contacts_.removeOne( contact );
}

//____________________________________________
void ClientInterface::pingServer( bool notify )
{
    Debug::Throw( "ClientInterface::pingServer.\n" );
    notifyPingServerSuccess_ = notify;

    if( !pingServerTimer_.isActive() && pingTimeout_ > 0 )
    {
        pingServerTimer_.start( pingTimeout_, this );
        client_->sendCommand( ChatCommand( connectionId_, ChatCommand::PingServer ) );
    }
}

//____________________________________________
void ClientInterface::pingClients( ConnectionId::List contacts, bool notify )
{
    Debug::Throw( "ClientInterface::pingClients.\n" );
    notifyPingClientSuccess_ = notify;

    if( !pingClientTimer_.isActive() && pingTimeout_ > 0 )
    {

        // generate list of ids ping
        // do nothing if non found
        pingIds_ = contacts;
        if( !pingIds_.isEmpty() )
        {
            pingClientTimer_.start( pingTimeout_, this );
            client_->sendCommand( ChatCommand( connectionId_, ChatCommand::PingClient ) );
        }

    }

}

//____________________________________________
void ClientInterface::fetchContacts( void )
{
    Debug::Throw( "ClientInterface::fetchContacts.\n" );
    client_->sendCommand( ChatCommand( connectionId_, ChatCommand::FetchContacts ) );
}

//____________________________________________
void ClientInterface::fetchMessages( QStringList recipients )
{
    Debug::Throw( "ClientInterface::fetchMessages.\n" );
    if( firstMessageTimeStamp_.isValid() )
    {

        ChatCommand command( connectionId_, ChatCommand::FetchMessages );
        command.setRecipients( recipients );
        command.setArguments( QStringList() << command.commandName() << "-t" << QString::number( firstMessageTimeStamp_ ) );
        client_->sendCommand( command );

    } else {

        fetchMessages( recipients, (HistoryMenu::Delay) XmlOptions::get().get<int>( "DEFAULT_HISTORY_SPAN" ) );

    }
}

//____________________________________________
void ClientInterface::fetchMessages( QStringList recipients, HistoryMenu::Delay delay )
{
    Debug::Throw() << "ClientInterface::fetchMessages - delay: " << delay << endl;

    // reset first message time stamp
    firstMessageTimeStamp_ = TimeStamp();
    lastMessageTimeStamp_ = TimeStamp();

    // create command
    ChatCommand command( connectionId_, ChatCommand::FetchMessages );
    command.setRecipients( recipients );

    // define relevant fetch time
    const TimeStamp timeStamp( _timeStampFromDelay( delay ) );

    // add argument, if needed
    if( timeStamp.isValid() )
    { command.setArguments( QStringList() << command.commandName() << "-t" << QString::number( timeStamp ) ); }

    // send command
    client_->sendCommand( command );
}

//____________________________________________
void ClientInterface::fetchLocalMessages( QStringList sourceRecipients )
{ fetchLocalMessages( sourceRecipients, (HistoryMenu::Delay) XmlOptions::get().get<int>( "DEFAULT_HISTORY_SPAN" ) ); }

//____________________________________________
void ClientInterface::fetchLocalMessages( QStringList sourceRecipients, HistoryMenu::Delay delay )
{
    Debug::Throw() << "ClientInterface::fetchLocalMessages - delay: " << delay << endl;

    // define relevant fetch time
    const TimeStamp timeStamp( _timeStampFromDelay( delay ) );

    // append current id to recipients
    if( sourceRecipients.empty() )
    {

        for( const auto& contact:contacts_ )
        { sourceRecipients.append( contact.user() ); }

    } else if( !sourceRecipients.contains( connectionId_.user() ) ) {

        sourceRecipients.append( connectionId_.user() );

    }

    messages_.setFilename( File( XmlOptions::get().raw("HISTORY_FILE") ) );
    if( messages_.needsReading( timeStamp )  )
    {

        if( !messageThread_->isRunning() )
        {
            messageThread_->setup( messages_, timeStamp, sourceRecipients );
            messageThread_->start();
        }

    } else {

        _sendMessageList( sourceRecipients, messages_.fetch( timeStamp, sourceRecipients ) );

    }

}

//____________________________________________
void ClientInterface::sendUserTyping( QStringList recipients )
{
    ChatCommand command( connectionId_, ChatCommand::UserTyping );
    command.setRecipients( recipients );
    client_->sendCommand( command );
}

//____________________________________________
void ClientInterface::sendMessage( ChatMessage message )
{

    Debug::Throw( "ClientInterface::sendMessage.\n" );
    if( connectionId_.connectionState() == ConnectionId::Connected )
    {

        client_->sendCommand( message.toChatCommand() );

    } else {

        // append to pending messages
        auto iter( qFind( pendingMessages_.begin(), pendingMessages_.end(), message ) );
        if( iter != pendingMessages_.end() )
        {

            message.setReceptionTime( iter->receptionTime() );
            *iter = message;

        } else {

            if( !message.receptionTime().isValid() )
            { message.setReceptionTime( TimeStamp::now() ); }

            pendingMessages_.append( message );

        }
    }

    messages_.append( message );

}

//____________________________________________
void ClientInterface::requestFileTransfer( QStringList recipients, BaseFileInfo fileInfo )
{

    Debug::Throw( "ClientInterface::requestFileTransfer.\n" );

    // creat socket and dedicated client
    FileEmissionClient* emissionClient = new FileEmissionClient( this );
    emissionClient->setConnectionId( connectionId_ );
    emissionClient->setFileInfo( fileInfo );
    fileEmissionClients_.append(emissionClient);
    connect( emissionClient, SIGNAL(finished()), SLOT(_fileEmissionClientDisconnected()) );
    emissionClient->socket().connectToHost( host_, port_ );

    // create file transfer widget
    auto application = Singleton::get().application<ClientApplication>();
    if( auto conversationWindow = application->findConversationWindow( connectionId_, recipients ) )
    {
        FileTransferManager& manager = conversationWindow->fileTransferManager();
        auto fileTransferWidget = manager.addFileTransfer( fileInfo, FileTransferWidget::Emission );
        fileTransferWidget->setRecipients( recipients );
        connect( fileTransferWidget, SIGNAL(fileTransferResent(QStringList,BaseFileInfo)), this, SLOT(requestFileTransfer(QStringList,BaseFileInfo)) );
        connect( emissionClient, SIGNAL(fileChunkAvailable(FileChunk)), fileTransferWidget, SLOT(processFileChunk(FileChunk)) );
    }

    // generate message
    const QString text = tr( "%1 sent file %2." )
        .arg( connectionId_.userString() )
        .arg( fileInfo.file().localName() );

    ChatMessage message( connectionId_, text );
    message.setRecipients( recipients );
    message.setMessageType( ChatMessage::TransferNotificationMessage );
    sendMessage( message );

    // emit signal
    emit messageRecieved( message );

    // generate command
    ChatCommand command( connectionId_, ChatCommand::FileTransferRequest );
    command.setRecipients( recipients );
    CommandLineArguments arguments;
    arguments
        << command.commandName()
        << "-f" << fileInfo.file()
        << "-s" << QString::number(fileInfo.size());
    command.setArguments( arguments );
    emissionClient->sendCommand( command );
}

//____________________________________________
void ClientInterface::acceptFileTransfer( ConnectionId sender, BaseFileInfo fileInfo, File downloadLocation )
{
    Debug::Throw( "ClientInterface::acceptFileTransfer.\n" );

    // create full path where file will be saved
    if( downloadLocation.isEmpty() ) downloadLocation = Util::home();

    // generate local file
    const File destinationFile( fileInfo.file().localName().addPath( downloadLocation ) );

    // create directory
    if( !File( destinationFile ).path().createDirectory() )
    {

        Debug::Throw(0) << "ClientInterface::acceptFileTransfer - unable to save file to " << File( destinationFile ).path() << endl;
        denyFileTransfer( sender, fileInfo );
        return;
    }

    // creat socket and dedicated client
    FileReceptionClient* receptionClient = new FileReceptionClient( this );
    receptionClient->setSender( sender );
    receptionClient->setFileInfo( fileInfo );
    receptionClient->setDestinationFile( destinationFile );
    fileReceptionClients_.append(receptionClient);
    connect( receptionClient, SIGNAL(finished()), SLOT(_fileReceptionClientDisconnected()) );

    // create file transfer widget
    auto application = Singleton::get().application<ClientApplication>();
    if( auto conversationWindow = application->findConversationWindow( connectionId_, ConnectionId::List() << sender ) )
    {
        FileTransferManager& manager = conversationWindow->fileTransferManager();
        auto fileTransferWidget = manager.addFileTransfer( fileInfo, FileTransferWidget::Reception );
        fileTransferWidget->setDestination( destinationFile );

        // connections
        connect( receptionClient, SIGNAL(fileChunkAvailable(FileChunk)), fileTransferWidget, SLOT(processFileChunk(FileChunk)) );
        connect( receptionClient, SIGNAL(transferFailed()), fileTransferWidget, SLOT(setTransferFailed()) );
    }

    // connect
    receptionClient->socket().connectToHost( host_, port_ );

    // generate fileTransferAccepted command and send through newly created socket
    // this will trigger the socket to be registered as a "file transfer" socket, by the server
    ChatCommand command( connectionId_, ChatCommand::FileTransferAccepted );
    command.addRecipient( sender.user() );
    CommandLineArguments arguments;
    arguments
        << command.commandName()
        << "-f" << fileInfo.file()
        << "-s" << QString::number(fileInfo.size());
    command.setArguments( arguments );
    receptionClient->sendCommand( command );

}

//____________________________________________
void ClientInterface::denyFileTransfer( ConnectionId sender, BaseFileInfo fileInfo )
{
    Debug::Throw( "ClientInterface::denyFileTransfer.\n" );

    // generate command
    ChatCommand command( connectionId_, ChatCommand::FileTransferDenied );
    command.addRecipient( sender.user() );
    command.setArguments( CommandLineArguments() << command.commandName() << "-f" << fileInfo.file() );
    client_->sendCommand( command );
}

//______________________________________
void ClientInterface::cancelFileTransfer( BaseFileInfo fileInfo )
{

    Debug::Throw() << "ClientInterface::cancelFileTransfer - File: " << fileInfo.file() << endl;

    // generate command
    ChatCommand command( connectionId_, ChatCommand::FileTransferCanceled );
    CommandLineArguments arguments;
    arguments
        << command.commandName()
        << "-f" << fileInfo.file()
        << "-s" << QString::number(fileInfo.size());
    command.setArguments( arguments );

    // send to client to perfom remote cancellation
    client_->sendCommand( command );

    // also process directly to perform local cancellation
    _processCommand( command );

}

//_____________________________________________________
void ClientInterface::timerEvent( QTimerEvent* event )
{
    Debug::Throw( "ClientInterface::timerEvent.\n" );
    if( event->timerId() == authorizationTimer_.timerId() )
    {

        authorizationTimer_.stop();
        if( connectionId_.connectionState() == ConnectionId::AwaitingReply )
        {
            disconnect();
            if( !tryReconnect_ )
            {
                setTryReconnect( true );
                emit error( tr( "Failed to get authorization from server" ), true );
            }

        }

    } else if( event->timerId() == pingServerTimer_.timerId() ) {

        pingServerTimer_.stop();
        notifyPingServerSuccess_ = false;

        // failed to contact server
        if( connectionId_.connectionState() == ConnectionId::Connected )
        {
            disconnect();
            if( !tryReconnect_ )
            {
                setTryReconnect( true );
                emit error( tr( "Lost connection to server"), true );
                Debug::Throw( "ClientInterface::timerEvent - timeout while waiting for server reply.\n" );
            }
        }

    } else if( event->timerId() == pingClientTimer_.timerId() ) {

        pingClientTimer_.stop();

        ConnectionId::List failedIds;
        for( const auto& connectionId:pingIds_ )
        { if( connectionId.connectionState() != ConnectionId::Disconnected ) failedIds.append( connectionId ); }

        // mark failed ids as offline
        for( auto oldConnectionId:failedIds )
        {
            int i = contacts_.indexOf( oldConnectionId );
            if( i >= 0 ) oldConnectionId = contacts_.at(i);

            ConnectionId connectionId( oldConnectionId );
            connectionId.setConnectionState( ConnectionId::Disconnected );
            connectionId.setDisconnectionTime( TimeStamp::now() );

            // replace in model
            if( i >= 0 ) contacts_.replace( i, connectionId );
            else contacts_.append( connectionId );

            // emit signal
            emit contactChanged( oldConnectionId, connectionId );
        }

        if( connectionId_.connectionState() == ConnectionId::Connected )
        {

            if( failedIds.size() == 1 ) {

                emit error( QString( tr( "Failed to contact user %1" ) ).arg( failedIds.begin()->user() ) );

            } else if( !failedIds.isEmpty() ) {

                QString buffer( tr( "The following users could not be contacted:\n" ) );
                for( const auto& id:failedIds )
                { buffer += QString( tr( " -%1\n" ) ).arg( id.user() ); }

                emit error( buffer );

            }

        }

        pingIds_.clear();
        notifyPingClientSuccess_ = false;

    } else if( event->timerId() == reconnectionTimer_.timerId() ) {

        if( connectionId_.connectionState() == ConnectionId::Disconnected && tryReconnect_ )
        {
            Debug::Throw( "ClientInterface::timerEvent - trying to reconnect.\n" );
            connect();
        }

    } else return QObject::timerEvent( event );

}

//_____________________________________________________
void ClientInterface::_error( QAbstractSocket::SocketError )
{
    Debug::Throw(0) << "ClientInterface::_error - error:" << client_->socket().errorString() << endl;
    disconnect();
    if( !tryReconnect_ )
    {
        setTryReconnect( true );
        emit error( client_->socket().errorString(), true );
    }
}

//_____________________________________________________
void ClientInterface::_connected( void )
{
    Debug::Throw( "ClientInterface::_connected.\n" );
    _setConnectionState( ConnectionId::AwaitingReply );

    // send connection request and start timer
    if( !authorizationTimer_.isActive() && authorizationTimeout_ > 0 )
    { authorizationTimer_.start( authorizationTimeout_, this ); }

    client_->sendCommand( ChatCommand( connectionId_, ChatCommand::ConnectionRequest ) );
}

//_____________________________________________________
void ClientInterface::_disconnected( void )
{
    Debug::Throw( "ClientInterface::_disconnected.\n" );
    _setConnectionState( ConnectionId::Disconnected );
}

//_____________________________________________________
void ClientInterface::_fileEmissionClientDisconnected( void )
{
    Debug::Throw( "ClientInterface::_fileEmissionClientDisconnected.\n" );
    FileEmissionClientList::iterator iter;
    while( (iter = std::find_if( fileEmissionClients_.begin(), fileEmissionClients_.end(), FileEmissionClient::SameStateFTor( QAbstractSocket::UnconnectedState ) ) ) != fileEmissionClients_.end() )
    {
        // delete socket, client, and remove from list
        (*iter)->deleteLater();
        fileEmissionClients_.erase( iter );
    }

}

//_____________________________________________________
void ClientInterface::_fileReceptionClientDisconnected( void )
{
    Debug::Throw( "ClientInterface::_fileReceptionClientDisconnected.\n" );
    FileReceptionClientList::iterator iter;
    while( (iter = std::find_if( fileReceptionClients_.begin(), fileReceptionClients_.end(), FileReceptionClient::SameStateFTor( QAbstractSocket::UnconnectedState ) ) ) != fileReceptionClients_.end() )
    {
        // delete socket, client, and remove from list
        (*iter)->deleteLater();
        fileReceptionClients_.erase( iter );
    }

}

//____________________________________________
void ClientInterface::_processCommand( ChatCommand command )
{

    Debug::Throw() << "ClientInterface::_processCommand - command: " << command.commandName() << " user: " << command.connectionId().user() << endl;
    switch( command.command() )
    {

        // connection accepted
        case ChatCommand::ConnectionAccepted:
        {

            if( command.connectionId() == connectionId_ )
            {

                setTryReconnect( false );
                authorizationTimer_.stop();

                // update connection state
                _setConnectionState( ConnectionId::Connected );

                // send avatar
                if( hasPendingAvatarChange_ )
                { _sendAvatar(); }

                // send pending commands
                for( const auto& message:pendingMessages_ )
                { client_->sendCommand( message.toChatCommand() ); }
                pendingMessages_.clear();

                // fetch contacts
                fetchContacts();

            }

            break;
        }

        // connection accepted
        case ChatCommand::ConnectionDenied:
        {

            if( command.connectionId() == connectionId_ )
            {
                disconnect();
                setTryReconnect( false );
                emit error( tr( "Lost connection to server" ), true );
            }

            break;

        }

        // pong recieved from server
        case ChatCommand::PongServer:
        {
            if( pingServerTimer_.isActive() )
            {
                pingServerTimer_.stop();

                if( notifyPingServerSuccess_ )
                { emit error( tr( "Successfully established contact with server" ) ); }

                notifyPingServerSuccess_ = false;

                break;
            }
        }

        // ping request from other client
        case ChatCommand::PingClient:
        {
            client_->sendCommand( ChatCommand( connectionId_, ChatCommand::PongClient ) );
            break;
        }

        case ChatCommand::ContactOnline:
        case ChatCommand::PongClient:
        {

            // find command contact in list
            const int i = contacts_.indexOf( command.connectionId() );
            if( i >= 0 )
            {

                // get old contact
                const ConnectionId oldConnectionId = contacts_.at(i);

                // update connection state and connection time
                ConnectionId connectionId( oldConnectionId );
                connectionId.setConnectionState( ConnectionId::Connected );
                connectionId.setConnectionTime( command.connectionId().connectionTime() );
                if( !connectionId.connectionTime().isValid() ) connectionId.setConnectionTime( TimeStamp::now() );

                // update contact list and emit signal
                contacts_.replace( i, connectionId );
                emit contactChanged( oldConnectionId, connectionId );

            } else {

                // copy contact from command
                ConnectionId connectionId( command.connectionId() );

                // make sure connection state and time are set
                connectionId.setConnectionState( ConnectionId::Connected );
                if( !connectionId.connectionTime().isValid() ) connectionId.setConnectionTime( TimeStamp::now() );

                // append contact to contact list and emit signal
                contacts_.append( connectionId );
                emit contactChanged( connectionId, connectionId );

            }

            if( command.command() == ChatCommand::PongClient )
            {

                // remove from list of pingId
                if( pingClientTimer_.isActive() )
                {

                    // remove from set
                    pingIds_.removeOne( command.connectionId() );

                    // stop timer and notify success if all clients have answered
                    if( pingIds_.isEmpty() )
                    {

                        pingClientTimer_.stop();
                        if( notifyPingClientSuccess_ )
                        { emit error( tr( "Successfully established contact with all connected users" ) ); }
                        notifyPingClientSuccess_ = false;
                    }

                }
            }

            break;

        }

        // client disconnected
        case ChatCommand::ContactOffline:
        {
            // find command contact in list
            const int i = contacts_.indexOf( command.connectionId() );
            if( i >= 0 )
            {

                // get old contact
                const ConnectionId oldConnectionId = contacts_.at(i);

                // update connection state and connection time
                ConnectionId connectionId( oldConnectionId );
                connectionId.setConnectionState( ConnectionId::Disconnected );
                connectionId.setDisconnectionTime( TimeStamp::now() );

                // update contact list and emit signal
                contacts_.replace( i, connectionId );
                emit contactChanged( oldConnectionId, connectionId );

            } else {

                // copy contact from command
                ConnectionId connectionId( command.connectionId() );

                // make sure connection state and time are set
                connectionId.setConnectionState( ConnectionId::Disconnected );
                connectionId.setDisconnectionTime( TimeStamp::now() );

                // append contact to contact list and emit signal
                contacts_.append( connectionId );
                emit contactChanged( connectionId, connectionId );

            }

            break;
        }

        // user state changed
        case ChatCommand::UserStateChanged:
        {
            // find command contact in list
            const int i = contacts_.indexOf( command.connectionId() );
            if( i >= 0 )
            {

                // get old contact
                const ConnectionId oldConnectionId = contacts_.at(i);

                // update user state
                ConnectionId connectionId( oldConnectionId );
                connectionId.setUserState( command.connectionId().userState() );

                // update contact list and emit signal
                emit contactChanged( oldConnectionId, connectionId );

                // update contacts
                contacts_.replace( i, connectionId );


            } else {

                // copy contact from command
                ConnectionId connectionId( command.connectionId() );

                // append contact to contact list and emit signal
                contacts_.append( connectionId );
                emit contactChanged( connectionId, connectionId );

            }

            break;
        }

        // clear connected clients list
        case ChatCommand::ClearContactList:
        {
            pingIds_.clear();
            emit clearContactList();
            break;
        }

        // clear messages
        case ChatCommand::ClearMessageList:
        {
            emit clearMessageList( command.recipients() );
            break;
        }

        // remote user is typing
        case ChatCommand::UserTyping:
        {
            emit userTyping( command.connectionId() );
            break;
        }

        // user name changed
        case ChatCommand::UserNameChanged:
        {

            // check arguments
            if( !chatCommandLineParser_.parse( command.arguments() ).hasOption( "-u" ) ) break;

            // get connection id from command
            ConnectionId oldConnectionId( command.connectionId() );
            const int i = contacts_.indexOf( oldConnectionId );
            if( i >= 0 ) oldConnectionId = contacts_.at( i );

            const QString newUserName = chatCommandLineParser_.option( "-u" );
            if( oldConnectionId.user() == newUserName ) break;

            ConnectionId connectionId = oldConnectionId;
            connectionId.setUser( newUserName );

            // replace in list
            if( i >= 0 ) contacts_.replace( i, connectionId );
            else contacts_.append( connectionId );

            emit contactChanged( oldConnectionId, connectionId );
            break;

        }

        // new message recieved from client
        case ChatCommand::Message:
        {
            const ChatMessage message( command );

            // update first message timeStamp
            if( message.receptionTime().isValid() && (!firstMessageTimeStamp_.isValid() ||  message.receptionTime() < firstMessageTimeStamp_ ) )
            { firstMessageTimeStamp_ = message.receptionTime(); }

            // update last message timeStamp
            if( message.receptionTime().isValid() &&  notifyMessagesFromFuture_ && (!lastMessageTimeStamp_.isValid() || message.receptionTime() > lastMessageTimeStamp_ ) )
            { lastMessageTimeStamp_ = message.receptionTime(); }

            // update local message set
            messages_.append( message );

            // emit signal
            emit messageRecieved( message );

            break;
        }

        // start recieving message logs (busy state)
        case ChatCommand::StartMessageLog:
        {
            if( chatCommandLineParser_.parse( command.arguments() ).hasOption( "-s" ) )
            {

                emit busyStateChanged( true, chatCommandLineParser_.option( "-s" ).toInt() );

            } else emit busyStateChanged( true );
            break;

        }

        // finished recieving message logs (busy state)
        case ChatCommand::StopMessageLog:
        {


            // check time of last message against now
            if( notifyMessagesFromFuture_ && lastMessageTimeStamp_ > TimeStamp::now() )
            {
                notifyMessagesFromFuture_ = false;
                _notifyMessagesFromFuture();
            }

            emit busyStateChanged( false );
            break;
        }

        // file transfer request
        case ChatCommand::FileTransferRequest:
        {

            if( chatCommandLineParser_.parse( command.arguments() ).hasOption( "-f" ) )
            {

                // build base file information from options
                const File filename = chatCommandLineParser_.option( "-f" );
                BaseFileInfo fileInfo( filename, BaseFileInfo::Document );
                if( chatCommandLineParser_.hasOption( "-s" ) )
                { fileInfo.setSize( chatCommandLineParser_.option( "-s" ).toLongLong() ); }

                emit fileTransferRequested( command.connectionId(), fileInfo );
            }

            break;
        }

        // file transfer accepted
        case ChatCommand::FileTransferAccepted:
        {
            if( chatCommandLineParser_.parse( command.arguments() ).hasOption( "-f" ) )
            {
                const File filename = chatCommandLineParser_.option( "-f" );
                BaseFileInfo fileInfo( filename, BaseFileInfo::Document );
                if( chatCommandLineParser_.hasOption( "-s" ) )
                { fileInfo.setSize( chatCommandLineParser_.option( "-s" ).toLongLong() ); }

                _sendFile( fileInfo );
            }

            break;
        }

        // file transfer accepted
        case ChatCommand::FileTransferDenied:
        {

            if( chatCommandLineParser_.parse( command.arguments() ).hasOption( "-f" ) )
            {
                const File filename = chatCommandLineParser_.option( "-f" );
                BaseFileInfo fileInfo( filename, BaseFileInfo::Document );
                if( chatCommandLineParser_.hasOption( "-s" ) )
                { fileInfo.setSize( chatCommandLineParser_.option( "-s" ).toLongLong() ); }

                // also cancel file emission client
                for( const auto& client:fileEmissionClients_ ) { if( client->fileInfo() == fileInfo ) client->cancel(); }

                auto application = Singleton::get().application<ClientApplication>();
                if( auto conversationWindow = application->findConversationWindow( connectionId_, ConnectionId::List() << command.connectionId() ) )
                { conversationWindow->fileTransferManager().setStatus( fileInfo, FileTransferWidget::Cancelled, tr( "Cancelled by recipient." ) ); }

            }

            break;

        }

        // file transfer cancelled
        case ChatCommand::FileTransferCanceled:
        {
            if( chatCommandLineParser_.parse( command.arguments() ).hasOption( "-f" ) )
            {
                const File filename = chatCommandLineParser_.option( "-f" );
                BaseFileInfo fileInfo( filename, BaseFileInfo::Document );
                if( chatCommandLineParser_.hasOption( "-s" ) )
                { fileInfo.setSize( chatCommandLineParser_.option( "-s" ).toLongLong() ); }

                // cancel both file emission clients and file reception clients
                for( const auto& client:fileEmissionClients_ ) { if( client->fileInfo() == fileInfo ) client->cancel(); }
                for( const auto& client:fileReceptionClients_ ) { if( client->fileInfo() == fileInfo ) client->cancel(); }

                // set proper status on filetransfer progress dialogs
                auto application = Singleton::get().application<ClientApplication>();
                if( auto conversationWindow = application->findConversationWindow( connectionId_, ConnectionId::List() << command.connectionId() ) )
                { conversationWindow->fileTransferManager().setStatus( fileInfo, FileTransferWidget::Cancelled ); }

                // also emit signal
                emit fileTransferCancelled( command.connectionId(), fileInfo );

            }

            break;
        }

        default: break;

    }

    return;

}

//____________________________________________
void ClientInterface::_processAvatar( Avatar avatar )
{

    Debug::Throw() << "ClientInterface::_processAvatar - recieved avatar for user " << avatar.connectionId().userString() << endl;

    // save
    const auto& image( avatar.image() );
    if( image.isNull() ) AvatarImage( avatar.connectionId().user() ).remove();
    else AvatarImage( avatar.connectionId().user(), image ).save();

    // emit signal
    emit avatarChanged( avatar.connectionId() );
}

//______________________________________
void ClientInterface::_fetchLocalMessagesFromThread( void )
{

    Debug::Throw( "ClientInterface::_fetchLocalMessagesFromThread.\n" );

    // synchronize
    ChatMessageSet messages( messageThread_->messages() );
    Debug::Throw() << "ClientInterface::_fetchLocalMessagesFromThread - found " << messages.messages().size() << " messages" << endl;

    for( const auto& message:messages_.messages() )
    { messages.append( message ); }
    messages_ = messages;

    // stop here if connected
    if( connectionId_.connectionState() != ConnectionId::Connected )
    { _sendMessageList( messageThread_->recipients(), messageThread_->messageList() ); }

}

//______________________________________
void ClientInterface::_updateConfiguration( void )
{
    Debug::Throw( "ClientInterface::_updateConfiguration.\n" );

    authorizationTimeout_ = 1000*XmlOptions::get().get<int>( "AUTHORIZATION_TIMEOUT" );
    pingTimeout_ = 1000*XmlOptions::get().get<int>( "PING_TIMEOUT" );

    if( parser_.hasOption( "--chat-host" ) )
    {

        // read from option
        QHostInfo hostInfo( QHostInfo::fromName( parser_.option( "--chat-host" ) ) );
        if( !hostInfo.addresses().isEmpty() ) host_ = hostInfo.addresses().front();

    } else {

        // read from option
        QHostInfo hostInfo( QHostInfo::fromName( XmlOptions::get().raw( "CHAT_HOST" ) ) );
        if( !hostInfo.addresses().isEmpty() ) host_ = hostInfo.addresses().front();

    }

    // port
    if( parser_.hasOption( "--chat-port" ) ) port_ = parser_.option( "--chat-port" ).toUInt();
    else  port_ = XmlOptions::get().get<unsigned int>( "CHAT_PORT" );

    // compression
    messages_.setUseCompression( XmlOptions::get().get<bool>( "COMPRESS_HISTORY" ) );

    // get userName from command line argument
    const ConnectionId oldConnectionId = connectionId_;
    if( parser_.hasOption( "--user" ) ) connectionId_.setUser( parser_.option( "--user" ).toLower() );
    else if( XmlOptions::get().contains( "USERNAME" ) ) connectionId_.setUser( XmlOptions::get().raw( "USERNAME" ) );

    // host
    for( const auto& address:QNetworkInterface::allAddresses() )
    {
        if( address != QHostAddress::LocalHost && address.toIPv4Address())
        {
            connectionId_.setHost( address.toString() );
            break;
        }
    }

    // assign default host if empty
    /* can happen when computer is not connected to the network */
    if( connectionId_.host().isEmpty() ) connectionId_.setHost( "localhost" );

    // adjust user status
    const QString userStateOptionName( QString( "USER_STATE_%1" ).arg( connectionId_.user() ) );
    if( XmlOptions::get().contains( userStateOptionName ) )
    { connectionId_.setUserState( (ConnectionId::UserState) XmlOptions::get().get<int>( userStateOptionName ) ); }

    emit contactChanged( oldConnectionId, connectionId_ );

    if( connectionId_.connectionState() == ConnectionId::Connected )
    {
        // send appropriate commands
        if( oldConnectionId.user() != connectionId_.user() )
        {

            // user name changed
            ChatCommand command( oldConnectionId, ChatCommand::UserNameChanged );
            command.setArguments( QStringList() << command.commandName() << "-u" << connectionId_.user() );
            client_->sendCommand( command );

        }

        if( oldConnectionId.userState() != connectionId_.userState() )
        {

            // user state changed
            client_->sendCommand( ChatCommand( connectionId_, ChatCommand::UserStateChanged ) );

        }

    }

    // reset reconnection timer
    setTryReconnect( tryReconnect_ );

    return;
}

//____________________________________________
void ClientInterface::_sendAvatar( void )
{

    Debug::Throw( "ClientInterface::sendAvatar.\n" );
    if( connectionId_.connectionState() == ConnectionId::Connected )
    {

        client_->sendAvatar( Avatar( connectionId_, AvatarImage( connectionId_.user() ) ) );
        hasPendingAvatarChange_ = false;

    } else {

        hasPendingAvatarChange_ = true;

    }

}

//______________________________________
void ClientInterface::_saveConfiguration( void )
{
    Debug::Throw( "ClientInterface::_saveConfiguration.\n" );
    messages_.write();
}

//______________________________________
void ClientInterface::_setConnectionState( ConnectionId::ConnectionState state )
{
    if( state == connectionId_.connectionState() ) return;

    // copy
    const ConnectionId oldConnectionId = connectionId_;

    // update state
    connectionId_.setConnectionState( state );

    // update connection/disconnection time
    if( state == ConnectionId::Connected ) connectionId_.setConnectionTime( TimeStamp::now() );
    else if( state == ConnectionId::Disconnected ) connectionId_.setDisconnectionTime( TimeStamp::now() );

    // emit signal
    emit contactChanged( oldConnectionId, connectionId_ );

    // also update contact list
    const int i = contacts_.indexOf( oldConnectionId );
    if( i >= 0 ) contacts_.replace( i, connectionId_ );
    else contacts_.append( connectionId_ );

}

//__________________________________________________
TimeStamp ClientInterface::_timeStampFromDelay( HistoryMenu::Delay delay ) const
{
    switch( delay )
    {
        default:
        case HistoryMenu::AllMessages:
        break;

        case HistoryMenu::Today: return QDateTime( QDate::currentDate() ).toTime_t(); break;
        case HistoryMenu::OneDay: return QDateTime( QDate::currentDate() ).addDays(-1).toTime_t(); break;
        case HistoryMenu::OneWeek: return QDateTime( QDate::currentDate() ).addDays(-7).toTime_t(); break;
        case HistoryMenu::TwoWeeks: return QDateTime( QDate::currentDate() ).addDays(-14).toTime_t(); break;
        case HistoryMenu::OneMonth: return QDateTime( QDate::currentDate() ).addMonths(-1).toTime_t(); break;
        case HistoryMenu::SixMonths: return QDateTime( QDate::currentDate() ).addMonths(-6).toTime_t(); break;
        case HistoryMenu::OneYear: return QDateTime( QDate::currentDate() ).addYears(-1).toTime_t(); break;
    }

    return TimeStamp();

}

//__________________________________________________
void ClientInterface::_sendMessageList( const QStringList& recipients, const ChatMessage::List& messages )
{
    if( messages.empty() ) return;

    Debug::Throw() << "ClientInterface::_sendMessageList - found " << messages.size() << " messages" << endl;

    emit busyStateChanged( true, messages.size() );
    emit clearMessageList( recipients );

    for( const auto& message:messages )
    {

        // update first message timeStamp
        if( message.receptionTime().isValid() && (!firstMessageTimeStamp_.isValid() ||  message.receptionTime() < firstMessageTimeStamp_ ) )
        { firstMessageTimeStamp_ = message.receptionTime(); }

        // update last message timeStamp
        if( message.receptionTime().isValid() &&  notifyMessagesFromFuture_ && (!lastMessageTimeStamp_.isValid() || message.receptionTime() > lastMessageTimeStamp_ ) )
        { lastMessageTimeStamp_ = message.receptionTime(); }

        emit messageRecieved( message );

    }

    // check time of last message against now
    if( notifyMessagesFromFuture_ && lastMessageTimeStamp_ > TimeStamp::now() )
    {
        notifyMessagesFromFuture_ = false;
        _notifyMessagesFromFuture();
    }

    emit busyStateChanged( false );

}

//__________________________________________________
void ClientInterface::_notifyMessagesFromFuture( void )
{ emit error( tr( "Your backlogs contains messages sent from the future. You should check your date and time settings.\n" ), false ); }

//____________________________________________
void ClientInterface::_sendFile( BaseFileInfo fileInfo )
{
    Debug::Throw( "ClientInterface::_sendFile.\n" );

    // find relevant file emission client
    for( const auto& client:fileEmissionClients_ )
    { if( client->fileInfo() == fileInfo ) client->send(); }

}
