#ifndef ClientInterface_h
#define ClientInterface_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Avatar.h"
#include "BaseFileInfo.h"
#include "ChatCommandLineParser.h"
#include "ConnectionId.h"
#include "ChatMessage.h"
#include "ChatMessageSet.h"
#include "File.h"
#include "FileChunk.h"
#include "HistoryMenu.h"

#include <QBasicTimer>
#include <QFile>
#include <QTcpServer>
#include <QHostAddress>

class ChatClient;
class ChatMessageThread;
class FileEmissionClient;
class FileReceptionClient;

class ClientInterface: public QObject, public Counter
{

    //* Qt meta object declaration
    Q_OBJECT

    public:

    //* constructor
    ClientInterface( QObject*, const CommandLineParser& );

    //*@name accessors
    //@{

    //* id
    const ConnectionId& connectionId( void ) const
    { return connectionId_; }

    //* contacts
    const ConnectionId::List& contacts( void ) const
    { return contacts_; }

    //@}

    //*@name modifiers
    //@{

    //* read contact list
    void readContacts( void );

    //* write contact list
    void writeContacts( void );

    //* set try reconnect flag
    /** adjust timeouts accordingly */
    void setTryReconnect( bool );

    //@}

    using QObject::connect;

    Q_SIGNALS:

    //* new error recieved
    void error( QString, bool tryReconnect = false );

    //* contact changed (old, new)
    void contactChanged( ConnectionId, ConnectionId );

    //* avatar changed
    void avatarChanged( ConnectionId );

    //* busy state changed
    void busyStateChanged( bool, int = 0 );

    //* user typing
    void userTyping( ConnectionId );

    //* clear list of connected clients
    void clearContactList( void );

    //* clear message list
    void clearMessageList( QStringList );

    //* new message recieved
    void messageRecieved( ChatMessage );

    //* emitted when file transfer is requested from remote client
    void fileTransferRequested( ConnectionId, BaseFileInfo );

    //* emitted when file transfer is cancelled by remote client
    void fileTransferCancelled( ConnectionId, BaseFileInfo );

    public Q_SLOTS:

    //* connect
    void connect( void );

    //* disconnect
    void disconnect( void );

    //* update user state
    void updateContact( ConnectionId, ConnectionId );

    //* remove a contact from contact list
    void removeContact( ConnectionId );

    //* ping server
    void pingServer( bool notify );

    //* ping clients
    void pingClients( ConnectionId::List, bool notify );

    //* fetch connected clients from server
    void fetchContacts( void );

    //* fetch messages from server
    void fetchMessages( QStringList );

    //* fetch messages from server
    void fetchMessages( QStringList, HistoryMenu::Delay );

    //* fetch messages from server
    void fetchLocalMessages( QStringList = QStringList() );

    //* fetch messages from local cache
    void fetchLocalMessages( QStringList, HistoryMenu::Delay );

    //* notify that user is typing
    void sendUserTyping( QStringList );

    //* send message to remote client
    void sendMessage( ChatMessage );

    //* request file transfer from remote client
    void requestFileTransfer( QStringList, BaseFileInfo );

    //* tell remote client that file transfer is accepted
    void acceptFileTransfer( ConnectionId, BaseFileInfo, File );

    //* tell remote client that file transfer is denied
    void denyFileTransfer( ConnectionId, BaseFileInfo );

    //* cancel file transfer
    void cancelFileTransfer( BaseFileInfo );

    protected:

    void timerEvent( QTimerEvent* );

    protected Q_SLOTS:

    //* client recieves errors
    void _error( QAbstractSocket::SocketError );

    //* connection granted
    void _connected( void );

    //* connection closed
    void _disconnected( void );

    //* file transfer client disconnected
    void _fileEmissionClientDisconnected( void );

    //* file transfer client disconnected
    void _fileReceptionClientDisconnected( void );

    //* process command
    void _processCommand( ChatCommand );

    //* process avatar
    void _processAvatar( Avatar );

    //* fetch local messages from thread
    void _fetchLocalMessagesFromThread( void );

    private Q_SLOTS:

    //* update configuration
    void _updateConfiguration( void );

    //* send avatar
    void _sendAvatar( void );

    //* save configuration
    void _saveConfiguration( void );

    private:

    //* changes application state, emit signal if changed
    void _setConnectionState( ConnectionId::ConnectionState );

    //* get time stamp from HistoryMenu delay
    TimeStamp _timeStampFromDelay( HistoryMenu::Delay ) const;

    //* send message list
    void _sendMessageList( const QStringList&, const ChatMessage::List& );

    //* notify when messages from the future has been recieved (due to wrong time settings)
    void _notifyMessagesFromFuture( void );

    //* send file to remote client
    void _sendFile( BaseFileInfo );

    // command line parser (passed at initialization)
    CommandLineParser parser_;

    //* debug level
    int debugLevel_ = 1;

    // host
    QHostAddress host_ = QHostAddress::Null;

    // port
    int port_ = 0;

    //* Client
    ChatClient* client_ = nullptr;

    //* list of file emission clients
    using FileEmissionClientList = QList<FileEmissionClient*>;
    FileEmissionClientList fileEmissionClients_;

    //* list of file reception clients
    using FileReceptionClientList = QList<FileReceptionClient*>;
    FileReceptionClientList fileReceptionClients_;

    //* used to retrieve clients matching a given file information
    class SameFileInfoFTor
    {

        public:

        //* constructor
        SameFileInfoFTor( const BaseFileInfo& fileInfo ):
            fileInfo_( fileInfo )
            {}

        //* predicate
        bool operator() ( FileEmissionClient* ) const;
        bool operator() ( FileReceptionClient* ) const;

        private:

        //* prediction
        BaseFileInfo fileInfo_;
    };

    //* chat command line parser
    ChatCommandLineParser chatCommandLineParser_;

    //* connection id
    ConnectionId connectionId_;

    //* set of connected clients
    ConnectionId::List contacts_;

    //* list of ids being pinged
    ConnectionId::List pingIds_;

    //* chat message list
    ChatMessageSet messages_;

    //* pending messages, sent while connection is down
    ChatMessage::List pendingMessages_;

    //* true if avatar was changed (and needs server update) while not yet connected
    bool hasPendingAvatarChange_ = false;

    //* thread, to read messages from file
    ChatMessageThread* messageThread_ = nullptr;

    //* reception timestamp of the oldest recieved message
    TimeStamp firstMessageTimeStamp_;
    TimeStamp lastMessageTimeStamp_;

    //* timeout
    QBasicTimer authorizationTimer_;
    QBasicTimer pingServerTimer_;
    QBasicTimer pingClientTimer_;
    QBasicTimer reconnectionTimer_;

    int authorizationTimeout_ = 0;
    int pingTimeout_ = 0;
    bool notifyPingServerSuccess_ = false;
    bool notifyPingClientSuccess_ = false;
    bool notifyMessagesFromFuture_ = true;

    //* try automatic reconnection at next ping
    /* no notification will be emitted if failed */
    bool tryReconnect_ = false;

};

#endif
