/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ConnectionId.h"
#include "Debug.h"
#include "XmlDef.h"
#include "XmlTimeStamp.h"
#include "Util.h"

#include <QHostAddress>

//____________________________________________________
ConnectionId::ConnectionId( const QString& user, const QString& host ):
    Counter( "ConnectionId" ),
    user_( user.toLower() ),
    host_( host )
{ Debug::Throw( "ConnectionId::ConnectionId.\n" ); }

//____________________________________________________
ConnectionId::ConnectionId( const QDomElement& element ):
    Counter( "ConnectionId" )
{
    Debug::Throw( "ConnectionId::ConnectionId (dom).\n" );

    // parse attributes
    QDomNamedNodeMap attributes( element.attributes() );
    for( int i=0; i<attributes.count(); i++ )
    {
        QDomAttr attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;
        if( attribute.name() == Xml::UserName ) setUser( attribute.value().toLower() );
        else if( attribute.name() == Xml::Host ) setHost( attribute.value() );
        else if( attribute.name() == Xml::UserState ) setUserState( (UserState) attribute.value().toInt() );
        else Debug::Throw() << "ConnectionId::ConnectionId - unrecognized attribute " << attribute.name() << endl;
    }

    // loop over children
    // parse children elements
    for(QDomNode chidNode = element.firstChild(); !chidNode.isNull(); chidNode = chidNode.nextSibling() )
    {
        QDomElement childElement = chidNode.toElement();
        if( childElement.isNull() ) continue;
        QString tagName( childElement.tagName() );
        if( tagName == Xml::ConnectionTime ) setConnectionTime( XmlTimeStamp( childElement ) );
        else if( tagName == Xml::DisconnectionTime ) setDisconnectionTime( XmlTimeStamp( childElement ) );
        else Debug::Throw() << "ConnectionId::ConnectionId - unrecognized tag " << tagName << endl;
    }
}

//____________________________________________________
QDomElement ConnectionId::domElement( QDomDocument& document ) const
{

    Debug::Throw( "ConnectionId::domElement" );
    QDomElement out( document.createElement( Xml::Id ) );
    if( !user_.isEmpty() ) out.setAttribute( Xml::UserName, user_.toLower() );
    if( !host_.isEmpty() ) out.setAttribute( Xml::Host, host_ );

    // user state Away state (set by Idletime), should not be written. One write 'online' instead
    if( userState_ == Away ) out.setAttribute( Xml::UserState, QString::number( Online ) );
    else  out.setAttribute( Xml::UserState, QString::number( userState_ ) );

    // time stamp
    if( connectionTime_.isValid() )
    { out.appendChild( XmlTimeStamp( connectionTime_ ).domElement( Xml::ConnectionTime, document ) ); }

    // time stamp
    if( disconnectionTime_.isValid() )
    { out.appendChild( XmlTimeStamp( disconnectionTime_ ).domElement( Xml::DisconnectionTime, document ) ); }

    return out;

}

//____________________________________________________
QString ConnectionId::userString( void ) const
{
    QString user( user_ );
    if(!user.isEmpty() ) user[0] = user[0].toUpper();
    return user;
}

//____________________________________________________
QString ConnectionId::userStateString( bool capitalize ) const
{
    QString out;
    switch( connectionState_ )
    {
        case Disconnected: out = QObject::tr( "Offline" ); break;
        case AwaitingReply: out = QObject::tr( "Waiting for server authorization" ); break;
        case Connected: out = userStateString( userState_, true ); break;
    }

    return capitalize ? out:out.toLower();
}

//____________________________________________________
QString ConnectionId::userStateString( UserState userState, bool capitalize )
{
    QString out;
    switch( userState )
    {
        default:
        case Online: out = QObject::tr( "Online" ); break;
        case AwayForced:
        case Away: out = QObject::tr( "Away" ); break;
        case Unavailable: out = QObject::tr( "Unavailable" ); break;
    }

    return capitalize ? out:out.toLower();

}

//____________________________________________________
QTextStream& operator << (QTextStream& out, const ConnectionId& id )
{
    out
        << "user: " << id.user()
        << " connection state: " << id.connectionState_
        << " user state: " << id.userState_
        << " connection: " << id.connectionTime_.toString()
        << " disconnection: " << id.disconnectionTime_.toString();
    return out;


}

//______________________________________________________
QDataStream& operator << (QDataStream& stream, const ConnectionId& id )
{
    static const quint32 version = 0;
    stream
        << version
        << id.user_
        << id.host_
        << quint32(id.userState_ )
        << id.connectionTime_
        << id.disconnectionTime_;
    return stream;
}

//______________________________________________________
QDataStream& operator >> (QDataStream& stream, ConnectionId& id )
{
    quint32 version;
    stream >> version;

    if( version == 0 )
    {

        quint32 userState;
        stream
            >> id.user_
            >> id.host_
            >> userState
            >> id.connectionTime_
            >> id.disconnectionTime_;
        id.userState_ = (ConnectionId::UserState) userState;

    } else Debug::Throw(0) << "Unrecognized ConnectionId version: " << version << endl;
    return stream;
}
