#ifndef ConnectionId_h
#define ConnectionId_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "TimeStamp.h"

#include <QDataStream>
#include <QDomDocument>
#include <QDomElement>
#include <QString>

//* Stores application name and user to uniquely identify an application.
class ConnectionId:public Counter
{
    public:

    //* list of connection ids
    using List = QList<ConnectionId>;

    //* constructor
    ConnectionId( const QString& user = QString(), const QString& host = QString() );

    //* constructor
    ConnectionId( const QDomElement& );

    //* equal to operator
    bool operator == (const ConnectionId& other ) const
    { return user_ == other.user_; }

    //* equal to operator
    bool operator != (const ConnectionId& other ) const
    { return !( other == *this ); }

    //* lower than to operator
    bool operator < (const ConnectionId& other ) const
    {
        if( user_ != other.user_ ) return user_ < other.user_;
        else if( host_ != other.host_ ) return host_ < other.host_;
        else return false;
    }

    //*@name accessors
    //@{

    //* dom element
    QDomElement domElement( QDomDocument& ) const;

    //* user
    const QString& user( void ) const
    { return user_; }

    //* capitalized user name
    QString userString( void ) const;

    //* user state enumeration
    enum UserState
    {
        Online,
        Away,
        AwayForced,
        Unavailable
    };

    //* user state
    UserState userState( void ) const
    { return userState_; }

    //* user state string
    QString userStateString( bool capitalize = true ) const;

    //* user state string
    static QString userStateString( UserState, bool = true );

    //* connection state
    enum ConnectionState
    {
        AwaitingReply,
        Connected,
        Disconnected
    };

    //* connection state
    ConnectionState connectionState( void ) const
    { return connectionState_; }

    //* host
    const QString& host( void ) const
    { return host_; }

    //* returns true if user and name makes sense
    bool isValid( void ) const
    { return !(user_.isEmpty() || host_.isEmpty() ); }

    //* connection time
    const TimeStamp& connectionTime( void ) const
    { return connectionTime_; }

    //* disconnection time
    const TimeStamp& disconnectionTime( void ) const
    { return disconnectionTime_; }

    //@}

    //*@name modifiers
    //@{

    //* user
    void setUser( QString value )
    { if( !value.isEmpty() ) user_ = value.toLower(); }

    //* user state
    void setUserState( UserState value )
    { userState_ = value; }

    //* connection state
    void setConnectionState( ConnectionState value )
    { connectionState_ = value; }

    //* host
    void setHost( const QString& value )
    { if( !value.isEmpty() ) host_ = value; }

    //* connection time
    void setConnectionTime( const TimeStamp& time )
    { connectionTime_ = time; }

    //* connection time
    void setDisconnectionTime( const TimeStamp& time )
    { disconnectionTime_ = time; }

    //@}

    private:

    //* user
    QString user_;

    //* host
    QString host_;

    //* connection state
    ConnectionState connectionState_ = Disconnected;

    //* user state
    UserState userState_ = Online;

    //* connection time
    TimeStamp connectionTime_;

    //* disconnection time
    TimeStamp disconnectionTime_;

    //* streamer
    friend QTextStream& operator << ( QTextStream&, const ConnectionId& );

    //*@name serializer
    //@{
    friend QDataStream& operator << (QDataStream&, const ConnectionId& );
    friend QDataStream& operator >> (QDataStream&, ConnectionId& );
    //@}

};

//* hash
inline uint qHash( const ConnectionId& contact )
{ return qHash(contact.user()); }

#endif
