/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ContactListDelegate.h"

#include "ConnectionId.h"
#include "ContactListModel.h"
#include "Debug.h"

#include <QAbstractTextDocumentLayout>
#include <QApplication>
#include <QStyleOptionViewItem>
#include <QPainter>

//_________________________________________________________
ContactListDelegate::ContactListDelegate( QObject *parent ):
    QAbstractItemDelegate( parent ),
    Counter( "Delegate" )
{ Debug::Throw( "ContactListDelegate::ContactListDelegate.\n" ); }

//_________________________________________________________
void ContactListDelegate::paint( QPainter *painter, const QStyleOptionViewItem &option, const QModelIndex &index ) const
{

    // check index
    if ( !index.isValid() ) return;

    // selection
    QStyleOptionViewItemV4 viewOption(option);
    viewOption.showDecorationSelected = true;
    QStyle *style = viewOption.widget ? viewOption.widget->style() : QApplication::style();
    style->drawPrimitive(QStyle::PE_PanelItemViewItem, &viewOption, painter, viewOption.widget);

    // create layout
    QTextDocument document;
    document.setDefaultFont( option.font );
    QTextCursor cursor( &document );

    // modify color
    QTextCharFormat charFormat( cursor.charFormat() );
    QPalette::ColorGroup colorGroup = option.state & QStyle::State_Enabled ? QPalette::Normal : QPalette::Disabled;

    const auto model( static_cast<const ContactListModel*>( index.model() ) );
    if( colorGroup == QPalette::Normal && model->contains( index ) )
    {
        // get contact and check connection state
        const auto& contact( model->get( index ) );
        if( contact.connectionState() == ConnectionId::Disconnected ) colorGroup = QPalette::Disabled;
    }

    if( colorGroup == QPalette::Normal && !(option.state & QStyle::State_Active) ) colorGroup = QPalette::Inactive;

    QColor foreground;
    if( option.state & QStyle::State_Selected ) foreground = option.palette.color( colorGroup, QPalette::HighlightedText );
    else foreground = option.palette.color( colorGroup, QPalette::Text );

    charFormat.setForeground( foreground );
    cursor.setCharFormat( charFormat );
    _layoutText( cursor, index );

    // draw icon
    const QIcon icon = model->data( index, Qt::DecorationRole ).value<QIcon>();
    const QPixmap pixmap( icon.pixmap( QSize( iconSize_, iconSize_ ) ) );
    painter->drawPixmap( option.rect.topLeft()+QPoint( margin_, margin_ ), pixmap );

    // draw text
    const QRect textRect( option.rect.adjusted( iconSize_ + 2*margin_, 0, 0, 0 ) );
    const auto textSize( document.size() );

    painter->save();
    painter->translate( textRect.left(), textRect.top() + 0.5*( textRect.height()-textSize.height() ) );
    painter->setClipRect( QRect( QPoint( 0, 0 ), textRect.size() ) );
    document.drawContents( painter );
    painter->restore();

    // draw focus
    _drawFocus( painter, option, option.rect );

}

//_________________________________________________________
QSize ContactListDelegate::sizeHint( const QStyleOptionViewItem &option, const QModelIndex &index ) const
{

    if ( !index.isValid() ) return QSize();

    // icon rect
    const QSize iconRect( iconSize_+2*margin_, iconSize_+2*margin_ );

    // create document and layout text
    QTextDocument document;
    document.setDefaultFont( option.font );
    QTextCursor cursor( &document );
    _layoutText( cursor, index );

    const QSize textRect( document.idealWidth(), document.size().height() );

    // size
    return QSize( iconRect.width() + textRect.width(), qMax( iconRect.height(), textRect.height() ) );

}

//_________________________________________________________
void ContactListDelegate::_layoutText( QTextCursor& cursor, const QModelIndex& index ) const
{
    // cast model
    const auto model( static_cast<const ContactListModel*>( index.model() ) );
    if( !model->contains( index ) ) return;

    const auto& contact( model->get( index ) );
    QTextCharFormat defaultFormat( cursor.charFormat() );
    QTextCharFormat boldFormat( defaultFormat );
    boldFormat.setFontWeight( QFont::Bold );
    cursor.insertText( contact.userString(), boldFormat );

    if( contact.connectionState() == ConnectionId::Connected )
    {
        switch( contact.userState() )
        {
            default:
            case ConnectionId::Online:
            break;

            case ConnectionId::Away:
            case ConnectionId::AwayForced:
            cursor.insertText( tr( " (away)" ), defaultFormat );
            break;

            case ConnectionId::Unavailable:
            cursor.insertText( tr( " (unavailable)" ), defaultFormat );
            break;

        }
    }

    // connection/disconnection time
    if( contact.connectionState() == ConnectionId::Disconnected && contact.disconnectionTime().isValid() )
    {

        cursor.insertText( "\n", boldFormat );
        cursor.insertText( tr( "Last seen: %1" ).arg( _timeString( contact.disconnectionTime() ) ), defaultFormat );

    } else if( contact.connectionTime().isValid() ) {

        cursor.insertText( "\n", boldFormat );
        cursor.insertText( tr( "Online since %1" ).arg( _timeString( contact.connectionTime() ) ), defaultFormat );

    }

    // pending messages
    const int pendingMessages = model->pendingMessages( contact );
    if( pendingMessages == 1 ) cursor.insertText( tr( "\nOne pending message" ) );
    else if( pendingMessages > 1 ) cursor.insertText( tr("\n%1 pending messages").arg( pendingMessages ) );

    // pending transfer requests
    const int pendingTransferRequests = model->pendingTransferRequests( contact );
    if( pendingTransferRequests == 1 ) cursor.insertText( tr( "\nOne pending file transfer request" ) );
    else if( pendingTransferRequests > 1 ) cursor.insertText( tr("\n%1 pending file transfer requests").arg( pendingTransferRequests ) );

}

//_________________________________________________________
void ContactListDelegate::_drawFocus( QPainter *painter, const QStyleOptionViewItem &option, const QRect &rect ) const
{
    if (option.state & QStyle::State_HasFocus)
    {

        QStyleOptionFocusRect o;
        o.QStyleOption::operator=(option);
        o.rect = rect;
        o.state |= QStyle::State_KeyboardFocusChange;
        QPalette::ColorGroup colorGroup = (option.state & QStyle::State_Enabled) ? QPalette::Normal : QPalette::Disabled;
        o.backgroundColor = option.palette.color( colorGroup, (option.state & QStyle::State_Selected) ? QPalette::Highlight : QPalette::Background );

        QApplication::style()->drawPrimitive( QStyle::PE_FrameFocusRect, &o, painter );

    }

}

//_____________________________________________
QString ContactListDelegate::_timeString( const TimeStamp& timeStamp ) const
{
    if( timeStamp.isSameDay( TimeStamp::now() ) )
    {

        return timeStamp.toString( TimeStamp::Time );

    } else return timeStamp.toString( TimeStamp::Short );

}
