/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ContactListModel.h"
#include "CustomPixmap.h"
#include "Debug.h"
#include "IconEngine.h"
#include "IconNames.h"

#include <QPalette>

//_______________________________________________
const QString ContactListModel::columnTitles_[ ContactListModel::nColumns ] =
{
    tr( "People" ),
    tr( "Host" ),
    tr( "Connected since" ),
    tr( "Last seen" )
};

//______________________________________________________________________
ContactListModel::ContactListModel( QObject* parent ):
    ListModel<ConnectionId>( parent ),
    Counter( "ContactListModel" )
{}

//______________________________________________________________________
QVariant ContactListModel::data( const QModelIndex& index, int role ) const
{

    // check index
    if( !contains( index ) ) return QVariant();

    // retrieve associated file info
    const ConnectionId& contact( get()[index.row()] );

    // return text associated to file and column
    switch( role )
    {

        case Qt::DisplayRole:
        {
            switch( index.column() )
            {
                case Name: return contact.userString();
                case Host: return contact.host();
                case ConnectionTime: return contact.connectionTime().toString();
                case DisconnectionTime: return contact.disconnectionTime().toString();
                default: return QVariant();
            }
            break;
        }

        case Qt::DecorationRole:
        {
            switch( index.column() )
            {
                case Name: return icon( contact );
                default: return QVariant();
            }
            break;
        }

        case Qt::ForegroundRole:
        {

            if( contact.connectionState() != ConnectionId::Connected )
            {

                return QPalette().color( QPalette::Disabled, QPalette::Text );

            } else return QVariant();

        }

        default: return QVariant();

    }

    return QVariant();

}

//______________________________________________________________________
QVariant ContactListModel::headerData(int section, Qt::Orientation orientation, int role ) const
{

    if(
        orientation == Qt::Horizontal &&
        role == Qt::DisplayRole &&
        section >= 0 &&
        section < nColumns )
    { return columnTitles_[section]; }

    // return empty
    return QVariant();

}

//______________________________________________________________________
QIcon ContactListModel::icon( const ConnectionId& contact ) const
{

    auto iter = avatars_.find( contact );
    if( contact.connectionState() == ConnectionId::Connected )
    {

        switch( contact.userState() ) {

            default:
            case ConnectionId::Online: return ( iter == avatars_.end() ) ? IconEngine::get( IconNames::UserOnline ):iter->normal;
            case ConnectionId::Away:
            case ConnectionId::AwayForced: return ( iter == avatars_.end() ) ? IconEngine::get( IconNames::UserAway ):iter->away;
            case ConnectionId::Unavailable: return ( iter == avatars_.end() ) ? IconEngine::get( IconNames::UserBusy ):iter->disabled;

        }

    } else {

        return ( iter == avatars_.end() ) ? IconEngine::get( IconNames::UserOffline ):iter->disabled;

    }

}

//______________________________________________________________________
bool ContactListModel::hasPendingMessages( void ) const
{ return !( pendingMessages_.isEmpty() && fileTransferRequests_.isEmpty() ); }

//______________________________________________________________________
int ContactListModel::pendingMessages( ConnectionId contact ) const
{
    auto iterator = pendingMessages_.find( contact );
    return ( iterator == pendingMessages_.end() ) ? 0:iterator.value();
}

//______________________________________________________________________
void ContactListModel::setAvatar( ConnectionId contact, QPixmap avatar )
{
    emit layoutAboutToBeChanged();
    if( avatar.isNull() ) avatars_.remove( contact );
    else {

        const IconList icons =
        {
            QIcon( avatar ),
            QIcon( CustomPixmap( avatar ).transparent( 0.6 ) ),
            QIcon( CustomPixmap( avatar ).desaturated().transparent( 0.6 ) )
        };

        avatars_.insert( contact, icons );

    }
    emit layoutChanged();
}

//______________________________________________________________________
void ContactListModel::clearPendingMessages( ConnectionId contact )
{
    auto messageIterator = pendingMessages_.find( contact );
    auto transferIterator = fileTransferRequests_.find( contact );
    const bool modified( !(messageIterator == pendingMessages_.end() && transferIterator == fileTransferRequests_.end() ) );
    if( modified ) emit layoutAboutToBeChanged();
    if( messageIterator != pendingMessages_.end() ) pendingMessages_.erase( messageIterator );
    if( transferIterator != fileTransferRequests_.end() ) fileTransferRequests_.erase( transferIterator );
    if( modified ) emit layoutChanged();
}

//______________________________________________________________________
void ContactListModel::incrementPendingMessages( ConnectionId contact )
{
    auto iterator = pendingMessages_.find( contact );
    emit layoutAboutToBeChanged();
    if( iterator == pendingMessages_.end() ) pendingMessages_.insert( contact, 1 );
    else iterator.value()++;
    emit layoutChanged();
}

//______________________________________________________________________
int ContactListModel::pendingTransferRequests( ConnectionId contact ) const
{
    auto iterator = fileTransferRequests_.find( contact );
    return ( iterator == fileTransferRequests_.end() ) ? 0:iterator.value();
}

//______________________________________________________________________
void ContactListModel::incrementFileTransferRequests( ConnectionId contact )
{
    auto iterator = fileTransferRequests_.find( contact );
    emit layoutAboutToBeChanged();
    if( iterator == fileTransferRequests_.end() ) fileTransferRequests_.insert( contact, 1 );
    else iterator.value()++;
    emit layoutChanged();
}

//______________________________________________________________________
void ContactListModel::decrementFileTransferRequests( ConnectionId contact )
{
    auto iterator = fileTransferRequests_.find( contact );
    if( iterator == fileTransferRequests_.end() ) return;
    else
    {
        emit layoutAboutToBeChanged();
        iterator.value()--;
        emit layoutChanged();
    }
}
