/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ContactListWindow.h"

#include "AvatarPixmap.h"
#include "BaseContextMenu.h"
#include "BaseStatusBar.h"
#include "ClientApplication.h"
#include "ContactListDelegate.h"
#include "ContactListMenu.h"
#include "CustomToolBar.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "IdleTime.h"
#include "ConversationWindow.h"
#include "QuestionDialog.h"
#include "ReconnectionDialog.h"
#include "SingleTimeDialog.h"
#include "Singleton.h"
#include "TreeView.h"
#include "XmlDef.h"
#include "XmlOptions.h"

#include <QHeaderView>
#include <QLayout>
#include <QShortcut>

//______________________________________________________________________
ContactListWindow::ContactListWindow( QWidget* parent ):
    BaseMainWindow( parent ),
    Counter( "ContactListWindow" )
{

    Debug::Throw( "ContactListWindow::ContactListWindow.\n" );
    setOptionName( "CONTACT_LIST_WINDOW" );

    setWindowTitle( tr( "Contact List - Homechat" ) );

    // actions
    _installActions();
    auto application( Singleton::get().application<ClientApplication>() );
    addAction( &application->closeAction() );

    // main widget and layout
    QWidget* main( new QWidget( this ) );
    setCentralWidget( main );

    // setup list
    QVBoxLayout* vLayout = new QVBoxLayout();
    vLayout->setSpacing(5);
    vLayout->setMargin(0);
    main->setLayout( vLayout );

    vLayout->addWidget( list_ = new TreeView( main ) );
    list_->setModel( &contactListModel_ );
    list_->setSortingEnabled( false );
    list_->setSelectionMode( QAbstractItemView::ContiguousSelection );
    list_->setItemDelegate( new ContactListDelegate( this ) );

    list_->setMask( 1 );
    list_->forceAlternatingRowColors();
    list_->header()->hide();
    list_->setContextMenuPolicy( Qt::CustomContextMenu );

    connect( list_, SIGNAL(customContextMenuRequested(QPoint)), SLOT(_raiseContextMenu(QPoint)) );
    connect( list_, SIGNAL(activated(QModelIndex)), SLOT(_startConversation(QModelIndex)) );
    connect( list_->selectionModel(), SIGNAL(currentChanged(QModelIndex,QModelIndex)), SLOT(_updateActions()) );

    // statusBar
    setStatusBar( statusBar_ = new BaseStatusBar( this ) );
    statusBar_->addLabel(1);
    statusBar_->addClock();

    // menu bar
    ContactListMenu* menu( new ContactListMenu( this ) );
    setMenuBar( menu );

    // toolbar
    CustomToolBar* toolbar  = new CustomToolBar( tr( "Main Toolbar" ), this, "MAIN_TOOLBAR" );
    toolbar->addAction( toggleConnectionAction_ );
    toolbar->addWidget( userStateButton_ );

    // configuration
    connect( application, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    connect( application, SIGNAL(avatarChanged()), SLOT(_updateAvatar()) );
    _updateConfiguration();
    _updateAvatar();

}

//_________________________________________________
void ContactListWindow::setConnectionId( ConnectionId connectionId )
{
    Debug::Throw() << "ContactListWindow::setConnectionId - connectionId: " << connectionId << endl;

    // store
    const bool userChanged( connectionId_.user() != connectionId.user() );
    connectionId_ = connectionId;

    if( userChanged ) _updateAvatar();

    // replace in model
    auto index( contactListModel_.index( connectionId_ ) );
    if( index.isValid() ) contactListModel_.replace( index, connectionId_ );
    else contactListModel_.add( connectionId_ );

    // update window title
    if( !connectionId_.user().isEmpty() ) setWindowTitle( QString( "%1 - Homechat" ).arg( connectionId_.userString() ) );
    else setWindowTitle( "Homechat" );

    // update user state button icon
    _updateUserStateButton();
}

//______________________________________________________________________
void ContactListWindow::setContacts( const ConnectionId::List& contacts )
{
    contactListModel_.clear();
    if( connectionId_.isValid() ) contactListModel_.add( connectionId_ );
    for( const auto& contact:contacts )
    {
        contactListModel_.add( contact );
        updateAvatar( contact );
    }
}

//_________________________________________________
bool ContactListWindow::eventFilter( QObject* object, QEvent* event )
{
    if( event->type() == QEvent::ActivationChange )
    {

        ConversationWindow* conversationWindow = qobject_cast<ConversationWindow*>( object );
        if( conversationWindow && conversationWindow->isActiveWindow() )
        {

            // get contacts
            // clear corresponding number of pending messages
            auto contacts( conversationWindow->contacts() );
            for( const auto& contact:contacts )
            { contactListModel_.clearPendingMessages( contact ); }

            // emit notification
            _updatePendingMessages();

            // send pending file transfer requests
            _sendFileTransferRequests( conversationWindow );

        }

    }

    return BaseMainWindow::eventFilter( object, event );

}

//_________________________________________________
void ContactListWindow::updateContact( ConnectionId oldConnectionId, ConnectionId connectionId )
{

    Debug::Throw() << "ContactListWindow::updateContact -" << " old: " << oldConnectionId << endl;
    Debug::Throw() << "ContactListWindow::updateContact -" << " new: " << connectionId << endl;

    // first update old contact from model
    auto index( contactListModel_.index( oldConnectionId ) );
    if( index.isValid() ) oldConnectionId = contactListModel_.get( index );

    // store whether userState or connection state has Changed
    const bool connectionStateChanged( oldConnectionId.connectionState() != connectionId.connectionState() );
    const bool userStateChanged( oldConnectionId.userState() != connectionId.userState() );

    // update model
    if( index.isValid() ) contactListModel_.replace( index, connectionId );
    else contactListModel_.add( connectionId );

    // update this connection id
    if( oldConnectionId == connectionId_ )
    {

        // update this connection id
        connectionId_ = connectionId;

        if( connectionStateChanged )
        {
            switch( connectionId_.connectionState() )
            {
                case ConnectionId::AwaitingReply:
                {
                    // update widget enable state
                    userStateActions_->setEnabled( false );
                    pingServerAction_->setEnabled( false );

                    // update toggle connection action
                    toggleConnectionAction_->setEnabled( true );
                    toggleConnectionAction_->setIcon( IconEngine::get( IconNames::Connect ) );
                    toggleConnectionAction_->setText( tr( "Sign Out" ) );
                    break;
                }

                case ConnectionId::Connected:
                {

                    // close all Reconnection dialogs
                    for( const auto& dialog:findChildren<ReconnectionDialog*>() )
                    { dialog->close(); }

                    // update widget enable state
                    userStateActions_->setEnabled( true );
                    pingServerAction_->setEnabled( true );

                    // update toggle +connection action
                    toggleConnectionAction_->setEnabled( true );
                    toggleConnectionAction_->setIcon( IconEngine::get( IconNames::Connect ) );
                    toggleConnectionAction_->setText( tr( "Sign Out" ) );

                    // start idle time monitoring
                    if( idleTime_->isSupported() && !idleTime_->isRunning() ) idleTime_->start();

                    // re-start check connection timer
                    checkConnectionsTimer_.stop();
                    const int timeout( 1000*60*( XmlOptions::get().get<int>( "CHECK_CONNECTIONS_TIMEOUT" ) ) );
                    if( timeout > 0 ) checkConnectionsTimer_.start( timeout, this );
                    break;

                }

                case ConnectionId::Disconnected:
                {

                    // update widget enable state
                    userStateActions_->setEnabled( false );
                    pingServerAction_->setEnabled( false );

                    // update toggle connection action
                    toggleConnectionAction_->setEnabled( true );
                    toggleConnectionAction_->setIcon( IconEngine::get( IconNames::Disconnect ) );
                    toggleConnectionAction_->setText( tr( "Sign In" ) );

                    // stop idle time monitoring
                    idleTime_->stop();

                    // stop connection check timer
                    checkConnectionsTimer_.stop();

                    // mark all contacts in list as offline and update disconnection time
                    for( auto contact:contactListModel_.get() )
                    {
                        if( contact.connectionState() != ConnectionId::Disconnected )
                        {
                            contact.setConnectionState( ConnectionId::Disconnected );
                            contact.setDisconnectionTime( TimeStamp::now() );
                            contactListModel_.add( contact );
                        }
                    }

                    break;
                }
            }

        }

        _updateUserStateButton();
        _updateStatusBar();

    } else {

        // check if notification is needed
        bool needsNotification = !isActiveWindow();
        if( needsNotification )
        {
            // check if there is an active matching main window
            auto application( Singleton::get().application<ClientApplication>() );
            auto conversationWindow( application->findConversationWindow( connectionId_, ConnectionId::List() << connectionId ) );
            needsNotification = !( conversationWindow && conversationWindow->isActiveWindow() );

        }

        // update ping clients action
        if( connectionStateChanged )
        { pingClientsAction_->setEnabled( _numOnlineContacts() > 0 ); }

        // emit notifications
        if( connectionStateChanged && needsNotification )
        {
            switch( connectionId.connectionState() )
            {
                case ConnectionId::Connected:
                emit newTextNotification( QString( tr( "%1 has joined" ) ).arg( connectionId.userString() ) );
                emit newSoundNotification( SoundNotifications::ContactOnline );
                break;

                case ConnectionId::Disconnected:
                emit newTextNotification( QString( tr( "%1 has quit" ) ).arg( connectionId.userString() ) );
                emit newSoundNotification( SoundNotifications::ContactOffline );
                break;

                default: break;
            }

        } else if( userStateChanged && connectionId.connectionState() == ConnectionId::Connected && needsNotification ) {

            emit newTextNotification( tr( "%1 is %2" ).arg( connectionId.userString() ).arg( connectionId.userStateString( false ) ) );

        }

    }

}

//_________________________________________________
void ContactListWindow::clearContactList( void )
{
    Debug::Throw( "ContactListWindow::clearContactList.\n" );

    // mark all contacts as offline
    for( auto contact:contactListModel_.get() )
    {
        contact.setConnectionState( ConnectionId::Disconnected );
        contactListModel_.add( contact );
    }

    // re-add current, to have the proper state
    contactListModel_.add( connectionId_ );
}

//_________________________________________________
void ContactListWindow::updateAvatar( ConnectionId connectionId )
{
    Debug::Throw() << "ContactListWindow::updateAvatar - user: " << connectionId.userString() << endl;
    contactListModel_.setAvatar( connectionId, AvatarPixmap( connectionId.user() ) );
    list_->update();
}

//_________________________________________________
void ContactListWindow::updateBusyState( bool busy, int )
{ busy_ = busy; }

//_________________________________________________
void ContactListWindow::notifyError( QString error, bool tryReconnect )
{

    Debug::Throw( "ContactListWindow::notifyError.\n" );
    if( isHidden() ) emit newTextNotification( error );
    else if( tryReconnect ) {

        // dialog must be modal to main window only
        // it should not prevent systray to work, for instance
        error += tr( "\nTry reconnect ?" );
        ReconnectionDialog* dialog = new ReconnectionDialog( this, error );
        dialog->setAttribute( Qt::WA_DeleteOnClose, true );
        dialog->setWindowModality( Qt::WindowModal );
        connect( &dialog->okButton(), SIGNAL(clicked()), this, SIGNAL(requestConnection()) );
        dialog->show();

    } else {

        // dialog must be modal to main window only
        // it should not prevent systray to work, for instance
        InformationDialog* dialog = new InformationDialog( this, error );
        dialog->setAttribute( Qt::WA_DeleteOnClose, true );
        dialog->setWindowModality( Qt::WindowModal );
        dialog->show();

    }

}

//________________________________________________________________
void ContactListWindow::notifyNewMessage( ChatMessage message )
{

    // do nothing if connectionId is not in contact list
    if( !contactListModel_.index( message.connectionId() ).isValid() ) return;

    // do nothing if connection id is oneself
    if( message.connectionId() == connectionId_ ) return;

    // check if notification must be sent
    const bool needSoundNotification = !busy_;
    bool needTextNotification = !busy_;
    if( needTextNotification )
    {

        // find matching main window
        auto application( Singleton::get().application<ClientApplication>() );
        auto conversationWindow( application->findConversationWindow( connectionId_, ConnectionId::List() << message.connectionId() ) );
        needTextNotification &= !(conversationWindow && conversationWindow->isActiveWindow());

    }

    // emit notification
    if( message.text().isEmpty() )
    {

        if( needTextNotification )
        { emit newTextNotification( QString( tr( "%1 has removed a message" ) ).arg( message.connectionId().userString() ) ); }

    } else if( message.isTransferNotificationMessage() ) {

        if( needTextNotification ) emit newTextNotification( QString( message.text() ) );
        if( needSoundNotification ) emit newSoundNotification( SoundNotifications::FileTransferRequest );

    }  else if( message.isNormalMessage() ) {

        if( needTextNotification )
        {
            emit newTextNotification( QString( tr( "%1 says: '%2'" ) )
                .arg( message.connectionId().userString() )
                .arg( message.text() ) );
            contactListModel_.incrementPendingMessages( message.connectionId() );
        }

        if( needSoundNotification ) emit newSoundNotification( SoundNotifications::NewMessage );

    }

    // update model
    if( !needTextNotification ) contactListModel_.clearPendingMessages( message.connectionId() );

    // update keeping track of pending messages
    _updatePendingMessages();

}

//_________________________________________________
void ContactListWindow::notifyFileTransferRequest( ConnectionId sender, BaseFileInfo fileInfo )
{
    Debug::Throw( "ConversationWindow::notifyFileTransferRequest.\n" );

    // do nothing if sender is not in list of contacts
    if( sender == connectionId_ ) return;
    if( !contactListModel_.index( sender ).isValid() ) return;

    // find matching main window
    auto application( Singleton::get().application<ClientApplication>() );
    auto conversationWindow( application->findConversationWindow( connectionId_, ConnectionId::List() << sender ) );

    if( conversationWindow ) conversationWindow->notifyFileTransferRequest( sender, fileInfo );
    else {

        fileTransferRequests_[sender].append( fileInfo );
        contactListModel_.incrementFileTransferRequests( sender );

    }
    return;

}

//_________________________________________________
void ContactListWindow::notifyFileTransferCancelled( ConnectionId sender, BaseFileInfo fileInfo )
{
    Debug::Throw( "ConversationWindow::notifyFileTransferCancelled.\n" );

    // do nothing if sender is not in list of contacts
    if( sender == connectionId_ ) return;
    if( !contactListModel_.index( sender ).isValid() ) return;

    // find matching main window
    auto application( Singleton::get().application<ClientApplication>() );
    auto conversationWindow( application->findConversationWindow( connectionId_, ConnectionId::List() << sender ) );

    if( conversationWindow ) conversationWindow->notifyFileTransferCancelled( sender, fileInfo );
    else if( fileTransferRequests_[sender].removeOne( fileInfo ) ) contactListModel_.decrementFileTransferRequests( sender );

    return;

}

//________________________________________________________________
void ContactListWindow::closeEvent( QCloseEvent *event )
{
    Debug::Throw( "ContactListWindow::closeEvent.\n" );
    event->accept();

    // trigger close action, unless systemtray icon is used
    if( !( QSystemTrayIcon::isSystemTrayAvailable() && XmlOptions::get().get<bool>( "USE_SYSTEMTRAY" ) ) )
    {

        Singleton::get().application<ClientApplication>()->closeAction().trigger();

    } else if( !XmlOptions::get().contains( "SHOW_CLOSE_DIALOG" ) || XmlOptions::get().get<bool>( "SHOW_CLOSE_DIALOG" ) ) {

        SingleTimeDialog dialog( this, tr(
            "Closing the main window will keep Homechat running in the System Tray. "
            "Use <b>Exit</b> from the menu, or the Homechat tray icon to exit the application." ) );

        dialog.setWindowTitle( tr( "System Tray - Homechat" ) );
        dialog.exec();
        XmlOptions::get().set<bool>( "SHOW_CLOSE_DIALOG", !dialog.dontShowAgain() );

    }

}

//_______________________________________________________
void ContactListWindow::timerEvent( QTimerEvent* event )
{

    if( event->timerId() == checkConnectionsTimer_.timerId() ) {

        if( connectionId_.connectionState() == ConnectionId::Connected )
        {

            // ping server
            requestPingServer( false );

            // ping clients
            auto contacts( _onlineContacts() );
            if( !contacts.isEmpty() )
            { emit requestPingClients( contacts, false ); }

        }

    } else return BaseMainWindow::timerEvent( event );

}

//______________________________________________________________________
void ContactListWindow::_updateActions( void )
{

    auto index( list_->selectionModel()->currentIndex() );
    if( index.isValid() )
    {

        auto contact( contactListModel_.get( index ) );
        removeContactAction_->setEnabled( contact != connectionId_ && contact.connectionState() == ConnectionId::Disconnected );
        startConversationAction_->setEnabled( contact != connectionId_ );

    } else {

        removeContactAction_->setEnabled( false );
        startConversationAction_->setEnabled( false );

    }

}

//_________________________________________________
void ContactListWindow::_toggleIdleTime( bool enabled )
{
    Debug::Throw() << "ContactListWindow::_toggleIdleTime" << endl;
    XmlOptions::get().set<bool>( "DETECT_IDLE_TIME", enabled );
    if( !enabled &&
        connectionId_.connectionState() == ConnectionId::Connected &&
        connectionId_.userState() == ConnectionId::Away )
    { _updateUserState( ConnectionId::Online ); }
}

//_________________________________________________
void ContactListWindow::_updateIdleState( bool isIdle )
{
    Debug::Throw() << "ContactListWindow::_updateIdleState - idle: " << isIdle << endl;
    if( connectionId_.connectionState() != ConnectionId::Connected ) return;
    if( toggleIdleTimeAction_->isEnabled() && !toggleIdleTimeAction_->isChecked() ) return;
    if( isIdle )
    {

        if( connectionId_.userState() == ConnectionId::Online )
        { _updateUserState( ConnectionId::Away ); }

    } else if( connectionId_.userState() == ConnectionId::Away ) {

        _updateUserState( ConnectionId::Online );

    }
}

//_________________________________________________
void ContactListWindow::_updateUserState( ConnectionId::UserState userState )
{

    // update connection id
    ConnectionId oldConnectionId( connectionId_ );
    connectionId_.setUserState( userState );

    // also update contacts
    auto index( contactListModel_.index( oldConnectionId ) );
    if( index.isValid() ) contactListModel_.replace( index, connectionId_ );
    else contactListModel_.add( connectionId_ );

    // emit signal
    emit contactChanged( oldConnectionId, connectionId_ );

    // update status bar
    _updateStatusBar();

}

//_________________________________________________
void ContactListWindow::_toggleConnection( void )
{
    Debug::Throw( "ContactListWindow::_toggleConnection.\n" );

    // emit connection/disconnection request
    switch( connectionId_.connectionState() )
    {
        default:
        case ConnectionId::Disconnected:
        emit requestConnection();
        break;

        case ConnectionId::AwaitingReply:
        case ConnectionId::Connected:
        emit requestDisconnection();
        break;
    }

    return;

}

//______________________________________________________________________
void ContactListWindow::_raiseContextMenu( QPoint position )
{

    Debug::Throw( "ContactListWindow::_raiseContextMenu.\n" );
    BaseContextMenu menu( list_ );
    menu.addAction( startConversationAction_ );
    menu.addAction( removeContactAction_ );
    menu.exec( list_->viewport()->mapToGlobal( position ) );

}

//______________________________________________________________________
void ContactListWindow::_removeContact( void )
{
    Debug::Throw( "ContactListWindow::_removeContact.\n" );

    // get index and check
    auto index( list_->selectionModel()->currentIndex() );
    if( !index.isValid() ) return;

    // get contact and check
    auto contact( contactListModel_.get( index ) );
    if( !(contact.connectionState() == ConnectionId::Disconnected ) ) return;

    // ask confirmation
    QuestionDialog dialog( this, tr( "Remove contact '%1' from this list ?" ).arg( contact.userString() ) );
    dialog.okButton().setText( tr( "Remove" ) );
    if( !dialog.centerOnParent().exec() ) return;

    // remove
    contactListModel_.remove( contact );

    // emit signal
    emit contactRemoved( contact );
}

//______________________________________________________________________
void ContactListWindow::_startConversation( void )
{ return _startConversation( list_->selectionModel()->currentIndex() ); }

//______________________________________________________________________
void ContactListWindow::_startConversation( QModelIndex index )
{

    Debug::Throw( "ContactListWindow::_startConversation\n" );
    if( !index.isValid() ) return;

    // get contact and check
    auto contact( contactListModel_.get( index ) );
    if( contact == connectionId_ ) return;

    auto application( Singleton::get().application<ClientApplication>() );
    auto conversationWindow( application->findConversationWindow( connectionId_, ConnectionId::List() << contact ) );
    if( conversationWindow ) conversationWindow->uniconify();
    else {

        conversationWindow = application->createConversationWindow( connectionId_, ConnectionId::List() << contact );
        conversationWindow->centerOnDesktop();
        conversationWindow->uniconify();

    }

}

//________________________________________________________
void ContactListWindow::_userStateActionTriggered( QAction* action )
{
    Debug::Throw( "ContactListWindow::_userStateActionTriggered.\n" );

    ActionMap::const_iterator iter( userStateActionMap_.find( action ) );
    if( iter != userStateActionMap_.constEnd() )
    { _updateUserState( iter.value() ); }

}

//________________________________________________________
void ContactListWindow::_pingClients( void )
{
    auto contacts( _onlineContacts() );
    if( !contacts.isEmpty() )
    { emit requestPingClients( contacts, true ); }
}

//______________________________________
void ContactListWindow::_updateConfiguration( void )
{
    Debug::Throw( "ContactListWindow::_updateConfiguration.\n" );

    // idle time
    idleTime_->setTimeout( 1000 * 60 * XmlOptions::get().get<int>( "IDLE_TIMEOUT" ) );

    toggleIdleTimeAction_->setEnabled( idleTime_->isSupported() && idleTime_->isEnabled() );
    toggleIdleTimeAction_->setChecked( XmlOptions::get().get<bool>( "DETECT_IDLE_TIME" ) );

    // connections timeout
    if( checkConnectionsTimer_.isActive() ) checkConnectionsTimer_.stop();
    const int timeout( 1000*60*( XmlOptions::get().get<int>( "CHECK_CONNECTIONS_TIMEOUT" ) ) );
    if( timeout > 0 ) checkConnectionsTimer_.start( timeout, this );

    return;
}

//_________________________________________________
void ContactListWindow::_updateAvatar( void )
{
    Debug::Throw( "ContactListWindow::_updateAvatar.\n" );
    contactListModel_.setAvatar( connectionId_, AvatarPixmap( connectionId_.user() ) );
}

//______________________________________________________________________
void ContactListWindow::_installActions( void )
{
    Debug::Throw( "ContactListWindow::_installActions" );

    // visibility action
    addAction( visibilityAction_ = new QAction( tr( "Show Contact List" ), this ) );
    connect( visibilityAction_, SIGNAL(triggered()), this, SLOT(uniconify()) );

    // toggle connection
    addAction( toggleConnectionAction_ = new QAction( IconEngine::get( IconNames::Disconnect ), tr( "Sign In" ), this ) );
    connect( toggleConnectionAction_, SIGNAL(triggered()), this, SLOT(_toggleConnection()) );

    // connected clients
    addAction( fetchContactsAction_ = new QAction( IconEngine::get( IconNames::Find ), tr( "Fetch Contacts" ), this ) );

    // server ping
    addAction( pingServerAction_ = new QAction( tr( "Check Connection to Server" ), this ) );
    connect( pingServerAction_, SIGNAL(triggered()), this, SIGNAL(requestPingServer()) );
    pingServerAction_->setEnabled( false );

    //  client ping
    addAction( pingClientsAction_ = new QAction( tr( "Check Users Connection" ), this ) );
    connect( pingClientsAction_, SIGNAL(triggered()), SLOT(_pingClients()) );
    pingClientsAction_->setEnabled( false );

    // remove contact
    addAction( removeContactAction_ = new QAction( IconEngine::get( IconNames::Remove ), tr( "Remove Contact" ), this ) );
    connect( removeContactAction_, SIGNAL(triggered()), this, SLOT(_removeContact()) );
    removeContactAction_->setEnabled( false );

    // remove contact
    addAction( startConversationAction_ = new QAction( tr( "Start Conversation" ), this ) );
    connect( startConversationAction_, SIGNAL(triggered()), this, SLOT(_startConversation()) );
    startConversationAction_->setEnabled( false );

    // toggle idle time
    addAction( toggleIdleTimeAction_ = new QAction( tr( "Mark Away When Idle" ), this ) );
    toggleIdleTimeAction_->setCheckable( true );
    toggleIdleTimeAction_->setChecked( true );
    toggleIdleTimeAction_->setEnabled( false );
    connect( toggleIdleTimeAction_, SIGNAL(toggled(bool)), this, SLOT(_toggleIdleTime(bool)) );

    // idle time
    idleTime_ = new IdleTime( this );
    connect( idleTime_, SIGNAL(idleStateChanged(bool)), this, SLOT(_updateIdleState(bool)) );

    // user state actions
    using ActionPair = QPair<QAction*, ConnectionId::UserState>;
    QVector<ActionPair> actions;
    actions
        << ActionPair( new QAction( IconEngine::get( IconNames::UserOnline ), ConnectionId::userStateString( ConnectionId::Online ), this ), ConnectionId::Online )
        << ActionPair( new QAction( IconEngine::get( IconNames::UserAway ), ConnectionId::userStateString( ConnectionId::Away ), this ), ConnectionId::AwayForced )
        << ActionPair( new QAction( IconEngine::get( IconNames::UserBusy ), ConnectionId::userStateString( ConnectionId::Unavailable ), this ), ConnectionId::Unavailable );

    userStateActions_ = new QActionGroup( this );
    for( const auto& actionPair:actions )
    {
        userStateActionMap_.insert( actionPair.first, actionPair.second );
        userStateActions_->addAction( actionPair.first );
    }

    connect( userStateActions_, SIGNAL(triggered(QAction*)), SLOT(_userStateActionTriggered(QAction*)) );
    userStateActions_->setEnabled( false );

    // user state button
    userStateButton_ = new CustomToolButton();
    userStateButton_->setIcon( IconEngine::get( IconNames::UserOnline ) );
    userStateButton_->setText( tr( "Status" ) );
    userStateButton_->setToolTip( tr( "Change status" ) );
    userStateButton_->setMenu( new QMenu() );
    userStateButton_->setPopupMode( QToolButton::InstantPopup );
    userStateButton_->menu()->addActions( userStateActions_->actions() );
    userStateButton_->setEnabled( false );
}

//_________________________________________________
void ContactListWindow::_updateUserStateButton( void ) const
{ userStateButton_->setEnabled( connectionId_.connectionState() == ConnectionId::Connected ); }

//_________________________________________________
void ContactListWindow::_updateStatusBar( void ) const
{ statusBar_->label().setTextAndUpdate( connectionId_.userStateString() ); }

//______________________________________________________________________
ConnectionId::List ContactListWindow::_onlineContacts( void ) const
{
    ConnectionId::List out;
    for( const auto& connectionId:contactListModel_.get() )
    {
        if( connectionId != connectionId_ && connectionId.connectionState() != ConnectionId::Disconnected )
        { out.append( connectionId ); }
    }
    return out;
}

//______________________________________________________________________
void ContactListWindow::_updatePendingMessages( void )
{
    const bool hasPendingMessages( contactListModel_.hasPendingMessages() );
    if( hasPendingMessages_ != hasPendingMessages )
    {
        hasPendingMessages_ = hasPendingMessages;
        emit this->hasPendingMessages( hasPendingMessages_ );
    }

}

//______________________________________________________________________
void ContactListWindow::_sendFileTransferRequests( ConversationWindow* conversationWindow )
{
    for( const auto& contact:conversationWindow->contacts() )
    {
        auto iterator = fileTransferRequests_.find( contact );
        if( iterator == fileTransferRequests_.end() ) continue;

        for( const auto& fileInfo:iterator.value() )
        { conversationWindow->notifyFileTransferRequest( contact, fileInfo ); }

        fileTransferRequests_.erase( iterator );

    }
}

//______________________________________________________________________
int ContactListWindow::_numOnlineContacts( void ) const
{

    int out( 0 );
    for( const auto& connectionId:contactListModel_.get() )
    { if( connectionId != connectionId_ && connectionId.connectionState() != ConnectionId::Disconnected ) ++out; }

    return out;
}
