#ifndef ContactListWindow_h
#define ContactListWindow_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "BaseMainWindow.h"

#include "ClientInterface.h"
#include "ContactListModel.h"
#include "Counter.h"
#include "SoundNotifications.h"

#include <QAction>
#include <QHash>

class TreeView;
class IdleTime;
class BaseStatusBar;
class ConversationWindow;

//* handles list of connection attributes
class ContactListWindow: public BaseMainWindow, public Counter
{

    Q_OBJECT

    public:

    //* constructor
    ContactListWindow( QWidget* = nullptr );

    //*@name accessors
    //@{

    //* contacts
    ConnectionId::List contacts( void ) const
    { return contactListModel_.get(); }

    //* status bar
    BaseStatusBar& statusBar( void ) const
    { return *statusBar_; }

    //@}

    //*@name modifiers
    //@{

    //* id
    void setConnectionId( ConnectionId );

    //* contacts
    void setContacts( const ConnectionId::List& );

    //@}

    //*@name actions
    //@{

    //* show/hide window
    QAction& visibilityAction( void ) const
    { return *visibilityAction_; }

    //* connection
    QAction& toggleConnectionAction( void ) const
    { return *toggleConnectionAction_; }

    //* connection
    QAction& pingServerAction( void ) const
    { return *pingServerAction_; }

    //* connection
    QAction& pingClientsAction( void ) const
    { return *pingClientsAction_; }

    //* fetch connected clients
    QAction& fetchContactsAction( void ) const
    { return *fetchContactsAction_; }

    //* remove contact
    QAction& removeContactAction( void ) const
    { return *removeContactAction_; }

    //* start conversation
    QAction& startConversationAction( void ) const
    { return *startConversationAction_; }

    //* toggle idle time detection
    QAction& toggleIdleTimeAction( void ) const
    { return *toggleIdleTimeAction_; }

    //* idle state
    IdleTime& idleTime( void ) const
    { return *idleTime_; }

    //* user state action group
    QActionGroup& userStateActions( void ) const
    { return *userStateActions_; }

    //@}

    //* event filters
    virtual bool eventFilter( QObject*, QEvent* );

    Q_SIGNALS:

    //* emitted when a contact is manually removed from the list
    void contactRemoved( ConnectionId );

    //* request connection
    void requestConnection( void );

    //* request disconnection
    void requestDisconnection( void );

    //* request ping server
    void requestPingServer( bool notify = true );

    //* request ping clients
    void requestPingClients( ConnectionId::List, bool notify );

    //* a contact has changed
    void contactChanged( ConnectionId, ConnectionId );

    //* new notification
    void newTextNotification( QString );

    //* new notification
    void newSoundNotification( SoundNotifications::Type );

    //* emitted when there are pending messages
    void hasPendingMessages( bool );

    public Q_SLOTS:

    //* update connectionId
    void updateContact( ConnectionId, ConnectionId);

    //* update avatar
    void updateAvatar( ConnectionId );

    //* clear list of contacts
    /**
    contacts are not removed from the list.
    they are marked offline instead. Will be updated by
    information provided by the client interface
    */
    void clearContactList( void );

    //* update busy state
    void updateBusyState( bool, int );

    //* new error recieved
    void notifyError( QString, bool );

    //* process new message
    /*
    it is used to send proper notifications,
    disregarding whether there is a conversation open or not
    will also be used to keep track of the number of pending messages
    */
    void notifyNewMessage( ChatMessage );

    //* notify file transfer request
    void notifyFileTransferRequest( ConnectionId, BaseFileInfo );

    //* notify file transfer cancelled
    void notifyFileTransferCancelled( ConnectionId, BaseFileInfo );

    protected:

    //* close event
    virtual void closeEvent( QCloseEvent* );

    //* timer event
    virtual void timerEvent( QTimerEvent* );

    protected Q_SLOTS:

    //* update actions
    void _updateActions( void );

    //* toggle idle time
    void _toggleIdleTime( bool );

    //* update idle state
    void _updateIdleState( bool );

    //* update user state
    void _updateUserState( ConnectionId::UserState );

    //* toggle connection
    void _toggleConnection( void );

    //* raise contect menu
    void _raiseContextMenu( QPoint );

    //* remove current contact from list
    void _removeContact( void );

    //* start conversation
    void _startConversation( void );

    //* start conversation
    void _startConversation( QModelIndex );

    //* user state action triggered
    void _userStateActionTriggered( QAction* );

    private Q_SLOTS:

    //* ping clients
    void _pingClients( void );

    //* configuration
    void _updateConfiguration( void );

    //* configuration
    void _updateAvatar( void );

    private:

    //* install actions
    void _installActions( void );

    //* update user state button
    void _updateUserStateButton( void ) const;

    //* update status bar
    void _updateStatusBar( void ) const;

    //* update pending messages (and emit signal)
    void _updatePendingMessages( void );

    //* send file transfer requests for a given main window
    void _sendFileTransferRequests( ConversationWindow* );

    //* number of online contacts (excluding this)
    int _numOnlineContacts( void ) const;

    //* get list of online contacts (excluding this)
    ConnectionId::List _onlineContacts( void ) const;

    //* current connection id
    ConnectionId connectionId_;

    //* idle time
    IdleTime* idleTime_ = nullptr;

    //* user treeview
    TreeView* list_ = nullptr;

    //* model
    ContactListModel contactListModel_;

    //* status bar
    BaseStatusBar* statusBar_;

    //* check connections timer
    QBasicTimer checkConnectionsTimer_;

    //* true if window is busy recieving logs
    /** no notifications will be sent */
    bool busy_ = false;

    //* true if there are pending messages
    bool hasPendingMessages_ = false;

    //* pending file transfer requests
    using FileTransferHash=QHash<ConnectionId,BaseFileInfo::List>;
    FileTransferHash fileTransferRequests_;

    //*@name actions
    //@{

    //* visibility
    QAction* visibilityAction_;

    //* toggle connection
    QAction* toggleConnectionAction_ = nullptr;

    //* ping server
    QAction* pingServerAction_ = nullptr;

    //* ping clients
    QAction* pingClientsAction_ = nullptr;

    //* fetch clients
    QAction* fetchContactsAction_ = nullptr;

    //* remove contact
    QAction* removeContactAction_;

    //* start conversation
    QAction* startConversationAction_;

    //* toggle idle time action
    QAction* toggleIdleTimeAction_;

    //* user state button
    QToolButton* userStateButton_ = nullptr;

    //* user state actions
    QActionGroup* userStateActions_ = nullptr;

    //* map actions to delay
    using ActionMap = QMap<QAction*, ConnectionId::UserState>;
    ActionMap userStateActionMap_;

    //@}

};

#endif
