/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ContactToolTipWidget.h"
#include "Debug.h"
#include "GridLayout.h"
#include "GridLayoutItem.h"
#include "QtUtil.h"
#include "Singleton.h"
#include "TimeStamp.h"
#include "XmlOptions.h"

#include <QLayout>
#include <QPainter>
#include <QPaintEvent>

//_______________________________________________________
ContactToolTipWidget::ContactToolTipWidget( QWidget* parent ):
    BaseToolTipWidget( parent )
{

    Debug::Throw( "ContactToolTipWidget::ContactToolTipWidget.\n" );

    // layout
    QHBoxLayout* hLayout = new QHBoxLayout();
    hLayout->setMargin( 10 );
    hLayout->setSpacing( 10 );
    setLayout( hLayout );

    hLayout->addWidget( iconLabel_ = new QLabel( this ) );
    iconLabel_->setMargin( 2 );
    iconLabel_->setAlignment( Qt::AlignHCenter|Qt::AlignTop );


    QVBoxLayout* vLayout = new QVBoxLayout();
    vLayout->setMargin( 0 );
    vLayout->setSpacing( 5 );
    hLayout->addLayout( vLayout );

    // file
    vLayout->addWidget( userNameLabel_ = new QLabel( this ) );
    userNameLabel_->setAlignment( Qt::AlignCenter );
    userNameLabel_->setFont( QtUtil::titleFont( userNameLabel_->font() ) );
    userNameLabel_->setMargin( 1 );

    // separator
    vLayout->addWidget( separator_ = new QFrame( this ) );
    separator_->setFrameStyle( QFrame::HLine );

    // grid layout
    GridLayout* gridLayout = new GridLayout();
    gridLayout->setMaxCount( 2 );
    gridLayout->setColumnAlignment( 0, Qt::AlignVCenter|Qt::AlignRight );
    gridLayout->setColumnAlignment( 1, Qt::AlignVCenter|Qt::AlignLeft );
    gridLayout->setMargin( 0 );
    gridLayout->setSpacing( 5 );
    vLayout->addLayout( gridLayout );

    // items
    ( hostItem_ = new GridLayoutItem( this, gridLayout ) )->setKey( tr( "IP address:" ) );
    ( connectionTimeItem_ = new GridLayoutItem( this, gridLayout ) )->setKey( tr( "Online since:" ) );
    ( disconnectionTimeItem_ = new GridLayoutItem( this, gridLayout ) )->setKey( tr( "Last seen:" ) );

    // hide host
    hostItem_->hide();

    // configuration
    connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();

}

//_______________________________________________________
void ContactToolTipWidget::setConnectionId( const ConnectionId& connectionId, QIcon icon )
{
    Debug::Throw()
        << "ContactToolTipWidget::setConnectionId - user: "
        << connectionId.userString()
        << endl;

    // local storage
    // icon_ = icon;
    connectionId_ = connectionId;

    // update icon
    if( !icon.isNull() )
    {

        icon_ = icon;

        iconLabel_->setPixmap( icon_.pixmap( QSize( pixmapSize_, pixmapSize_ ) ) );
        iconLabel_->show();

    } else iconLabel_->hide();

    bool needSeparator( false );
    if( !connectionId.user().isEmpty() )
    {

        userNameLabel_->setText( connectionId.userString() );
        userNameLabel_->show();

        if( connectionId.connectionState() == ConnectionId::Disconnected )
        {

            if( connectionId.disconnectionTime().isValid() )
            {

                needSeparator = true;
                disconnectionTimeItem_->setText( _timeString( connectionId.disconnectionTime() ) );

            } else {

                disconnectionTimeItem_->hide();

            }

            connectionTimeItem_->hide();

        } else {

            if( connectionId.connectionTime().isValid() )
            {

                needSeparator = true;
                connectionTimeItem_->setText( _timeString( connectionId.connectionTime() ) );

            } else {

                connectionTimeItem_->hide();

            }

            disconnectionTimeItem_->hide();

        }

    } else {

        // file and separator
        userNameLabel_->hide();
        connectionTimeItem_->hide();
        disconnectionTimeItem_->hide();

    }

    separator_->setVisible( needSeparator );

    adjustSize();

}

//_____________________________________________
void ContactToolTipWidget::_updateConfiguration( void )
{
    Debug::Throw( "ContactToolTipWidget::_updateConfiguration.\n" );
    if( XmlOptions::get().contains( "TOOLTIPS_PIXMAP_SIZE" ) ) setPixmapSize( XmlOptions::get().get<unsigned int>( "TOOLTIPS_PIXMAP_SIZE" ) );
}

//_____________________________________________
QString ContactToolTipWidget::_timeString( const TimeStamp& timeStamp ) const
{
    if( timeStamp.isSameDay( TimeStamp::now() ) )
    {

        return timeStamp.toString( TimeStamp::Time );

    } else return timeStamp.toString( TimeStamp::Short );

}
