/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ConversationWindow.h"

#include "AvatarPixmap.h"
#include "ClientApplication.h"
#include "Command.h"
#include "ContactList.h"
#include "ContactListWindow.h"
#include "CustomToolBar.h"
#include "CustomToolButton.h"
#include "EmoticonButton.h"
#include "EmoticonProvider.h"
#include "EmoticonSelectionMenu.h"
#include "FileDialog.h"
#include "FileTransferRequestDialog.h"
#include "FileTransferManager.h"
#include "HistoryMenu.h"
#include "HtmlHighlighter.h"
#include "IconNames.h"
#include "IconEngine.h"
#include "InformationDialog.h"
#include "LineEditor.h"
#include "LogWidget.h"
#include "ConversationWindowMenu.h"
#include "MessageBoardWidget.h"
#include "MessageEditorWidget.h"
#include "OpenWithDialog.h"
#include "ProgressStatusBar.h"
#include "QuestionDialog.h"
#include "Singleton.h"
#include "TimeStamp.h"
#include "TreeView.h"
#include "XmlDocument.h"

#include <QDesktopServices>
#include <QLayout>
#include <QMimeData>
#include <QSystemTrayIcon>
#include <QUrl>

//_________________________________________________
ConversationWindow::ConversationWindow( QWidget* parent ):
    BaseMainWindow( parent ),
    Counter( "ConversationWindow" ),
    contactListModel_( this )
{

    Debug::Throw( "ConversationWindow::ConversationWindow.\n" );
    setOptionName( "MAIN_WINDOW" );

    // actions
    _installActions();
    auto application( Singleton::get().application<ClientApplication>() );

    // main widget and layout
    QWidget* main( new QWidget( this ) );
    setCentralWidget( main );

    mainLayout_ = new QBoxLayout( QBoxLayout::LeftToRight );
    mainLayout_->setSpacing(2);
    mainLayout_->setMargin(0);
    main->setLayout(mainLayout_);

    // connected clients list
    contactList_ = new ContactList( main );
    contactList_->setModel( &contactListModel_ );
    contactList_->setOrientation( Qt::Vertical );
    mainLayout_->addWidget( contactList_ );

    // message board
    QVBoxLayout* vLayout = new QVBoxLayout();
    vLayout->setMargin(0);
    vLayout->setSpacing(2);
    mainLayout_->addLayout( vLayout );

    splitter_ = new QSplitter( main );
    splitter_->setOrientation( Qt::Vertical );
    vLayout->addWidget( splitter_ );

    messageBoardWidget_ = new MessageBoardWidget();
    messageBoardWidget_->viewport()->installEventFilter( this );
    TextEditor::Container* container = new TextEditor::Container( splitter_, messageBoardWidget_ );
    splitter_->addWidget( container );

    splitter_->addWidget( fileTransferManager_ = new FileTransferManager( splitter_ ) );
    fileTransferManager_->hide();

    // horizontal layout
    QHBoxLayout* hLayout = new QHBoxLayout();
    hLayout->setMargin(0);
    vLayout->addLayout( hLayout );

    // message editor
    messageEditorWidget_ = new MessageEditorWidget( main );
    messageEditorWidget_->installEventFilter( this );
    connect( messageEditorWidget_, SIGNAL(textEdited(QString)), this, SLOT(_editorModified()) );
    hLayout->addWidget( messageEditorWidget_ );

    // emoticons
    EmoticonButton* emoticonButton = new EmoticonButton( main );
    emoticonButton->setUpdateFromOptions( false );
    emoticonButton->setColorize( true );
    {
        emoticonButton->setIcon( QIcon( ":/smile.svgz" ) );
        emoticonButton->setAutoRaise( true );
        EmoticonSelectionMenu* menu = new EmoticonSelectionMenu();
        emoticonButton->setMenu( menu );
        emoticonButton->setPopupMode( QToolButton::InstantPopup );
        emoticonButton->setIconSize( IconSize( IconSize::Small ) );
        connect( menu, SIGNAL(emoticonSelected(Emoticon)), this, SLOT(_insertEmoticon(Emoticon)) );
    }

    hLayout->addWidget( emoticonButton );

    // statusBar
    setStatusBar( statusBar_ = new ProgressStatusBar( this ) );
    statusBar_->addClock();

    // menu bar
    ConversationWindowMenu* menu( new ConversationWindowMenu( this ) );
    setMenuBar( menu );

    // toolbar
    CustomToolBar* toolbar  = new CustomToolBar( tr( "Main Toolbar" ), this, "MAIN_TOOLBAR" );
    toolbar->addAction( &application->contactListWindow().toggleConnectionAction() );
    toolbar->addAction( sendFileAction_ );
    toolbar->addWidget( historyButton_ );

    // reset replace iterator
    _clearReplaceIterator();

    // connections
    messageEditorWidget_->setFocus();
    connect( messageEditorWidget_, SIGNAL(returnPressed()), this, SLOT(_sendMessage()) );
    connect( &messageEditorWidget_->cancelMessageEditionAction(), SIGNAL(triggered()), this, SLOT( _cancelMessageEdition()) );
    connect( &messageEditorWidget_->removeMessageAction(), SIGNAL(triggered()), this, SLOT(_removeReplaceMessage()) );
    connect( &messageBoardWidget_->editMessageAction(), SIGNAL(triggered()), this, SLOT(_editContextMenuMessage()) );
    connect( &messageBoardWidget_->removeMessageAction(), SIGNAL(triggered()), this, SLOT(_removeContextMenuMessage()) );
    connect( &messageBoardWidget_->openLinkAction(), SIGNAL(triggered()), this, SLOT(_openLink()) );
    connect( messageBoardWidget_, SIGNAL(linkActivated(QString)), this, SLOT(_openLink(QString)) );

    connect( fileTransferManager_, SIGNAL(fileTransferAdded(BaseFileInfo)), SLOT(_showFileTransfers()) );
    connect( &fileTransferManager_->visibilityAction(), SIGNAL(triggered()), SLOT(_showFileTransfers()) );

    // configuration
    connect( application, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    connect( application, SIGNAL(avatarChanged()), SLOT(_updateAvatar()) );
    _updateConfiguration();
    _updateAvatar();

}

//______________________________________________________________________
void ConversationWindow::setContacts( const ConnectionId::List& contacts )
{
    contactListModel_.clear();
    if( connectionId_.isValid() ) contactListModel_.add( connectionId_ );
    for( const auto& contact:contacts )
    {
        contactListModel_.add( contact );
        updateAvatar( contact );
    }
}

//_________________________________________________
void ConversationWindow::setConnectionId( ConnectionId connectionId )
{

    Debug::Throw() << "ConversationWindow::setConnectionId - connectionId: " << connectionId << endl;

    // store
    const bool userChanged( connectionId_.user() != connectionId.user() );
    connectionId_ = connectionId;

    if( userChanged ) _updateAvatar();

    // replace in model
    auto index( contactListModel_.index( connectionId_ ) );
    if( index.isValid() ) contactListModel_.replace( index, connectionId_ );
    else contactListModel_.add( connectionId_ );

    // update window title
    if( !connectionId_.user().isEmpty() ) setWindowTitle( QString( "%1 - Homechat" ).arg( connectionId_.userString() ) );
    else setWindowTitle( "Homechat" );

    // update actions
    sendFileAction_->setEnabled( connectionId_.connectionState() == ConnectionId::Connected );


}

//_________________________________________________
bool ConversationWindow::eventFilter( QObject* object, QEvent* event )
{
    if( object == messageEditorWidget_ ) return messageEditorEventFilter( object, event );
    else if( object == messageBoardWidget_->viewport() ) return messageBoardEventFilter( object, event );
    else return BaseMainWindow::eventFilter( object, event );
}

//_________________________________________________
bool ConversationWindow::messageEditorEventFilter( QObject* object, QEvent* event )
{

    switch( event->type() )
    {

        case QEvent::KeyPress:
        {
            QKeyEvent *keyEvent = static_cast<QKeyEvent*>(event);
            switch( keyEvent->key() )
            {
                default: break;
                case Qt::Key_Up:
                {

                    // check message list
                    if( messageList_.isEmpty() ) break;

                    // find previous message in stack that matches same user
                    ChatMessage::List::iterator iterator( replaceIterator_ );
                    while( iterator != messageList_.begin() )
                    {
                        --iterator;
                        if( ( iterator->connectionId() == connectionId_ && !iterator->text().isEmpty() && iterator->isNormalMessage() ) || iterator == messageList_.begin() ) break;
                    }

                    // if found, update editor text
                    if( iterator->connectionId() == connectionId_ )
                    {
                        replaceIterator_ = iterator;
                        replaceMessage_ = *iterator;
                        messageEditorWidget_->setText( replaceMessage_.text() );
                    }

                    break;
                }

                case Qt::Key_Down:
                {

                    // check message list and iterator
                    if( messageList_.isEmpty() ) break;
                    if( replaceIterator_ == messageList_.end() ) break;

                    // find previous message in stack that matches same user
                    ChatMessage::List::iterator iterator( replaceIterator_ );
                    while( iterator != messageList_.end() )
                    {
                        ++iterator;
                        if( iterator == messageList_.end() || ( iterator->connectionId() == connectionId_ && !iterator->text().isEmpty() && iterator->isNormalMessage() ) ) break;
                    }

                    // if found, update editor text
                    if( iterator == messageList_.end() )
                    {

                        messageEditorWidget_->clear();
                        _clearReplaceIterator();

                    } else {

                        replaceIterator_ = iterator;
                        replaceMessage_ = *iterator;
                        messageEditorWidget_->setText( replaceMessage_.text() );

                    }

                    break;
                }

            }

        }

        case QEvent::ContextMenu:
        {
            messageEditorWidget_->cancelMessageEditionAction().setEnabled( replaceIterator_ != messageList_.end() );
            messageEditorWidget_->removeMessageAction().setEnabled( replaceIterator_ != messageList_.end() );
            break;
        }

        default: break;

    }

    // base class
    return BaseMainWindow::eventFilter( object, event );
}


//_________________________________________________
bool ConversationWindow::messageBoardEventFilter( QObject* object, QEvent* event )
{

    switch( event->type() )
    {
        case QEvent::ContextMenu:
        {

            // find text cursor and block number
            const QTextCursor textCursor( messageBoardWidget_->cursorForPosition( static_cast<QContextMenuEvent*>( event )->pos() ) );
            const int blockNumber( textCursor.blockNumber() );

            // find matching message
            contextMenuMessage_ = ChatMessage();
            for( const auto& message:messageSet_ )
            {
                const auto& blockRange( message.blockRange() );
                if( blockNumber >= blockRange.first && blockNumber <= blockRange.second )
                {
                    contextMenuMessage_ = message;
                    break;
                }
            }

            messageBoardWidget_->editMessageAction().setEnabled( contextMenuMessage_.isValid() && contextMenuMessage_.connectionId() == connectionId_ && contextMenuMessage_.isNormalMessage() );
            messageBoardWidget_->removeMessageAction().setEnabled( contextMenuMessage_.isValid() && contextMenuMessage_.connectionId() == connectionId_ && contextMenuMessage_.isNormalMessage() );
        }
        break;

        case QEvent::DragEnter:
        {

            QDragEnterEvent* dragEnterEvent = static_cast<QDragEnterEvent*>(event);

            // check mimedata
            auto mimeData( dragEnterEvent->mimeData() );
            if( !mimeData->hasUrls() ) return true;

            // check urls
            auto urls( mimeData->urls() );
            if( !( urls.size() == 1 && urls.front().isLocalFile() ) ) return true;

            dragEnterEvent->acceptProposedAction();
            dragEnterEvent->accept();

            return true;
        }

        case QEvent::DragMove:
        { return true; }

        case QEvent::Drop:
        {

            QDropEvent* dropEvent = static_cast<QDropEvent*>(event);

            // check mimedata
            auto mimeData( dropEvent->mimeData() );
            if( !mimeData->hasUrls() ) return true;

            // check urls
            auto urls( mimeData->urls() );
            if( urls.size() != 1 ) return true;

            auto url( urls.front() );
            if( !url.isLocalFile() ) return true;

            const File filename( url.toLocalFile() );

            // accept event
            dropEvent->acceptProposedAction();
            dropEvent->accept();

            if( QuestionDialog( this, tr( "Send '%1' ?" ).arg( filename ) ).centerOnParent().exec() )
            {
                // ask for file transfer
                BaseFileInfo fileInfo( filename );
                fileInfo.setType( BaseFileInfo::Document );
                fileInfo.setSize( filename.fileSize() );
                emit requestFileTransfer( _contactNames(), fileInfo );
            }

            return true;
        }

        default: break;

    }

    // base class
    return BaseMainWindow::eventFilter( object, event );

}

//_________________________________________________
void ConversationWindow::updateContact( ConnectionId oldConnectionId, ConnectionId connectionId )
{

    Debug::Throw() << "ConversationWindow::updateContact -" << " old: " << oldConnectionId << endl;
    Debug::Throw() << "ConversationWindow::updateContact -" << " new: " << connectionId << endl;

    // do nothing if old contact is not in model
    auto index( contactListModel_.index( oldConnectionId ) );
    if( !index.isValid() ) return;

    // update old connection id based on model
    oldConnectionId = contactListModel_.get( index );

    // store whether userState or connection state has Changed
    const bool connectionStateChanged( oldConnectionId.connectionState() != connectionId.connectionState() );

    // update model
    contactListModel_.replace( index, connectionId );

    if( oldConnectionId == connectionId_ )
    {

        // update this connection id
        connectionId_ = connectionId;

        // update actions
        sendFileAction_->setEnabled( connectionId_.connectionState() == ConnectionId::Connected );

        if( connectionStateChanged )
        {
            switch( connectionId_.connectionState() )
            {

                case ConnectionId::Connected:
                {
                    emit requestMessages( _contactNames() );
                    break;
                }

                case ConnectionId::Disconnected:
                {

                    // mark all contacts in list as offline and update disconnection time
                    for( auto contact:contactListModel_.get() )
                    {
                        if( contact.connectionState() != ConnectionId::Disconnected )
                        {
                            contact.setConnectionState( ConnectionId::Disconnected );
                            contact.setDisconnectionTime( TimeStamp::now() );
                            contactListModel_.add( contact );
                        }
                    }

                    break;
                }

                default: break;

            }

        }

        _updateStatusBar();

    }

}

//_________________________________________________
void ConversationWindow::updateAvatar( ConnectionId connectionId )
{
    Debug::Throw() << "ConversationWindow::updateAvatar - user: " << connectionId.userString() << endl;
    contactListModel_.setAvatar( connectionId, AvatarPixmap( connectionId.user() ) );
}

//_________________________________________________
void ConversationWindow::clearContactList( void )
{
    Debug::Throw( "ConversationWindow::clearContactList.\n" );

    // mark all contacts as offline
    for( auto contact:contactListModel_.get() )
    {
        contact.setConnectionState( ConnectionId::Disconnected );
        contactListModel_.add( contact );
    }

    // re-add current, to have the proper state
    contactListModel_.add( connectionId_ );
}

//_________________________________________________
void ConversationWindow::updateBusyState( bool busy, int size )
{
    Debug::Throw() << "ConversationWindow::updateBusyState - busy: " << busy << endl;
    busy_ = busy;
    if( busy_ )
    {
        // need at lest 10 messages to trigger progressbar
        if( size > 10 )
        {
            messageBoardWidget_->setUpdatesEnabled( false );
            statusBar_->showProgressBar();
            statusBar_->progressBar().setMaximum( size );
            statusBar_->progressBar().reset();
        }

    } else {

        messageBoardWidget_->commit();
        messageBoardWidget_->setUpdatesEnabled( true );
        statusBar_->showLabel();

    }

}

//_________________________________________________
void ConversationWindow::printMessage( ChatMessage message )
{
    Debug::Throw() << "ConversationWindow::printMessage - user: " << message.connectionId().user() << " text: " << message.text() << endl;

    // check validity
    if( !message.isValid() ) return;

    // check if contacts match
    // to accept the message both the current id and at least one of
    // my contacts must be in the message recipients list
    auto messageRecipients( message.recipients() );
    messageRecipients.append( message.connectionId().user() );
    if( messageRecipients.contains( connectionId_.user() ) )
    {
        bool accepted = false;
        for( const auto& contact:_contactNames() )
        {
            if( messageRecipients.contains( contact ) )
            { accepted = true; break; }

        }

        if( !accepted ) return;

    } else return;

    // check reception time
    if( !message.receptionTime().isValid() ) message.setReceptionTime( TimeStamp::now() );

    // increment progressbar if visible
    if( statusBar_->progressBar().isVisible() )
    { statusBar_->progressBar().setValue( statusBar_->progressBar().value()+1 ); }

    // append message
    if( _appendMessage( message ) )
    {

        // clear message board and last printed message
        messageBoardWidget_->clear();

        // print full list
        _printMessageList();
        if( !busy_ ) messageBoardWidget_->commit();

    } else {

        _printMessage( message );
        if( !busy_ ) messageBoardWidget_->commit();

    }

}

//_______________________________________________________
void ConversationWindow::timerEvent( QTimerEvent* event )
{

    if( event->timerId() == userTypingTimer_.timerId() )
    {

        // stop timer and save time
        userTypingTimer_.stop();
        _updateStatusBar();

    } else if( event->timerId() == thisUserTypingTimer_.timerId() ) {

        thisUserTypingTimer_.stop();

    } else return BaseMainWindow::timerEvent( event );

}

//_________________________________________________
void ConversationWindow::clearMessageList( QStringList recipients )
{
    Debug::Throw( "ConversationWindow::clearMessageList.\n" );

    // contact names agains recipients
    for( const auto& contact:_contactNames() )
    { if( !recipients.contains( contact ) ) return; }

    // clear messages
    messageSet_.clear();
    messageList_.clear();
    _clearReplaceIterator( false );

    // clear board
    messageBoardWidget_->clear();
}

//_________________________________________________
void ConversationWindow::notifyUserTyping( ConnectionId connectionId )
{
    Debug::Throw( "ConversationWindow::notifyUserTyping.\n" );
    if( notifyUserTyping_ && contactListModel_.index( connectionId ).isValid() )
    {
        statusBar_->label().setTextAndUpdate( QString( tr( "%1 is writing a message ..." ) ).arg( connectionId.userString() ) );
        if( userTypingTimer_.isActive() ) userTypingTimer_.stop();
        userTypingTimer_.start( userTypingTimeout_, this );
    }
}

//_________________________________________________
void ConversationWindow::notifyFileTransferRequest( ConnectionId sender, BaseFileInfo fileInfo )
{
    Debug::Throw( "ConversationWindow::notifyFileTransferRequest.\n" );

    // do nothing if sender is not in list of contacts
    if( !contactListModel_.index( sender ).isValid() ) return;

    // create dialog to ask for acceptation
    FileTransferRequestDialog* dialog = new FileTransferRequestDialog( this, sender, fileInfo );
    dialog->setAttribute( Qt::WA_DeleteOnClose, true );

    // forward signals
    connect( dialog, SIGNAL(denyFileTransfer(ConnectionId,BaseFileInfo)), this, SIGNAL(denyFileTransfer(ConnectionId, BaseFileInfo)));
    connect( dialog, SIGNAL(acceptFileTransfer(ConnectionId,BaseFileInfo,File)), this, SIGNAL(acceptFileTransfer(ConnectionId, BaseFileInfo, File)));

    // show dialog
    dialog->show();

    return;
}

//_________________________________________________
void ConversationWindow::notifyFileTransferCancelled( ConnectionId sender, BaseFileInfo fileInfo )
{
    Debug::Throw( "ConversationWindow::notifyFileTransferCancelled.\n" );

    // do nothing if sender is not in list of contacts
    if( !contactListModel_.index( sender ).isValid() ) return;

    // create dialog to ask for acceptation
    for( const auto& dialog:findChildren<FileTransferRequestDialog*>() )
    {
        if( dialog->sender() == sender && dialog->fileInfo() == fileInfo )
        {
            dialog->deleteLater();
            break;
        }
    }

    return;
}

//_________________________________________________
void ConversationWindow::_sendMessage( void )
{
    const QString text( messageEditorWidget_->text() );
    if( text.isEmpty() )
    {
        // reset replace iterator and do nothing
        _clearReplaceIterator();
        return;
    }

    // create message and print
    ChatMessage message( connectionId_, text );
    message.setRecipients( _contactNames() );

    // adjust creation and reception time from replaceIterator if valid
    if( replaceMessage_.isValid() )
    {
        message.setCreationTime( replaceMessage_.creationTime() );
        message.setReceptionTime( replaceMessage_.receptionTime() );
        _clearReplaceIterator();
    }

    // print
    printMessage( message );

    // send to server
    emit newMessage( message );
    messageEditorWidget_->clear();

}

//________________________________________________________
void ConversationWindow::_processDelayChanged(HistoryMenu::Delay delay )
{
    emit( connectionId_.connectionState() == ConnectionId::Connected ) ?
        requestMessages( _contactNames(), delay ) :
        requestLocalMessages( _contactNames(), delay );
}

//________________________________________________________
void ConversationWindow::_reload( void )
{
    emit( connectionId_.connectionState() == ConnectionId::Connected ) ?
        requestMessages( _contactNames() ) :
        requestLocalMessages( _contactNames() );
}

//________________________________________________________
void ConversationWindow::_insertEmoticon( Emoticon emoticon )
{
    messageEditorWidget_->insert( emoticon.strings().front() );
    messageEditorWidget_->setFocus();
}

//________________________________________________________
void ConversationWindow::_editorModified( void )
{
    // check timer
    if( thisUserTypingTimer_.isActive() ) return;

    /*
    emit signal and start timer
    this prevents too many signals to be sent
    */
    thisUserTypingTimer_.start( userTypingTimeout_/2, this );

    // emit signal
    emit userTyping( _contactNames() );
}

//______________________________________
void ConversationWindow::_editContextMenuMessage( void )
{
    Debug::Throw( "ConversationWindow::_editContextMenuMessage.\n" );
    if( !( contextMenuMessage_.isValid() && contextMenuMessage_.connectionId() == connectionId_ && contextMenuMessage_.isNormalMessage() ) ) return;

    // copy context menu message into replaceMessage_, and update iterator
    replaceMessage_ = contextMenuMessage_;
    _updateReplaceIterator();

    // update editor
    messageEditorWidget_->setText( contextMenuMessage_.text() );

}

//______________________________________
void ConversationWindow::_removeContextMenuMessage( void )
{

    if( !( contextMenuMessage_.isValid() && contextMenuMessage_.connectionId() == connectionId_ && contextMenuMessage_.isNormalMessage() ) ) return;

    Debug::Throw()
        << "ConversationWindow::_editContextMenuMessage - message: "
        << contextMenuMessage_
        << endl;

    contextMenuMessage_.setText( QString() );

    // print
    printMessage( contextMenuMessage_ );

    // send to server
    emit newMessage( contextMenuMessage_ );

}

//______________________________________
void ConversationWindow::_removeReplaceMessage( void )
{

    if( replaceIterator_ == messageList_.end() ) return;

    Debug::Throw()
        << "ConversationWindow::_removeReplaceMessage - message: "
        << *replaceIterator_
        << endl;

    replaceIterator_->setText( QString() );
    replaceMessage_ = *replaceIterator_;

    // print
    printMessage( replaceMessage_ );

    // send to server
    emit newMessage( replaceMessage_ );

    // reset replace iterator
    _clearReplaceIterator();

    // clear editor widget
    messageEditorWidget_->clear();

}

//______________________________________
void ConversationWindow::_cancelMessageEdition( void )
{
    Debug::Throw( "ConversationWindow::_cancelMessageEdition.\n" );
    if( replaceIterator_ == messageList_.end() ) return;

    // clear replace iterator
    _clearReplaceIterator();

    // clear message editor
    messageEditorWidget_->clear();
}

//______________________________________
void ConversationWindow::_selectFile( void )
{
    Debug::Throw( "ConversationWindow::_selectFile.\n" );
    File filename( FileDialog(this).getFile() );
    if( !filename.isNull() )
    {
        BaseFileInfo fileInfo( filename );
        fileInfo.setType( BaseFileInfo::Document );
        fileInfo.setSize( filename.fileSize() );
        emit requestFileTransfer( _onlineContactNames(), fileInfo );
    }

}

//_____________________________________________
void ConversationWindow::_openLink( void )
{
    Debug::Throw( "ConversationWindow::_openLink.\n" );
    const QString anchor( messageBoardWidget_->anchor() );
    if( anchor.isEmpty() ) return;

    OpenWithDialog dialog( this );
    dialog.setWindowTitle( tr( "Open Link - Homechat" ) );
    dialog.setLink( anchor );
    dialog.setOptionName( "OPEN_LINK_APPLICATIONS" );
    dialog.realizeWidget();
    dialog.exec();

}

//_____________________________________________
void ConversationWindow::_openLink( QString anchor )
{

    Debug::Throw( "ConversationWindow::_openLink.\n" );
    if( !anchor.isEmpty() ) QDesktopServices::openUrl( QUrl::fromEncoded( anchor.toLatin1() ) );

}

//______________________________________
void ConversationWindow::_showFileTransfers( void )
{

    Debug::Throw( "ConversationWindow::_showFileTransfers.\n" );

    // first show
    if( !fileTransferManager_->isVisible() )
    { fileTransferManager_->show(); }

    const QSize size( fileTransferManager_->sizeHint() );

    // second, check splitter
    auto sizes = splitter_->sizes();
    if( sizes[1] < size.height() )
    {
        sizes[0] += sizes[1] - size.height();
        sizes[1] = size.height();
        splitter_->setSizes( sizes );
    }

}

//______________________________________
void ConversationWindow::_updateConfiguration( void )
{
    Debug::Throw( "ConversationWindow::_updateConfiguration.\n" );
    notifyUserTyping_ = XmlOptions::get().get<bool>( "NOTIFY_USER_TYPING" );
    if( !notifyUserTyping_ ) _updateStatusBar();

    bool changed( false );
    if( messageBoardWidget_->htmlHighlighter().setEnabled( XmlOptions::get().get<bool>( "SHOW_LINKS" ) ) ) changed = true;
    if( messageBoardWidget_->emoticonProvider().setEnabled( XmlOptions::get().get<bool>( "USE_EMOTICONS" ) ) ) changed = true;
    if( messageBoardWidget_->emoticonProvider().setIconSize( XmlOptions::get().get<int>( "EMOTICON_SIZE" ) ) ) changed = true;
    if( messageBoardWidget_->emoticonProvider().setColorize( XmlOptions::get().get<bool>( "COLORIZE_EMOTICONS" ) ) ) changed = true;
    if( changed  )
    {

        // clear message board and last printed message
        messageBoardWidget_->clear();

        // print full list
        _printMessageList();
        messageBoardWidget_->commit();

    }

    // contact list
    if( XmlOptions::get().get<bool>( "FLAT_CONTACT_LIST" ) )
    {

        contactList_->setFrameStyle( QFrame::NoFrame );
        contactList_->setBackgroundRole( QPalette::Window );
        contactList_->viewport()->setBackgroundRole( QPalette::Window );

    } else {

        contactList_->setFrameStyle( QFrame::StyledPanel | QFrame::Sunken );
        contactList_->viewport()->setBackgroundRole( QPalette::Base );

    }

    if( XmlOptions::get().get<bool>( "HORIZONTAL_CONTACT_LIST" ) )
    {

        mainLayout_->setDirection( QBoxLayout::TopToBottom );
        contactList_->setOrientation( Qt::Horizontal );

    } else {

        mainLayout_->setDirection( QBoxLayout::LeftToRight );
        contactList_->setOrientation( Qt::Vertical );

    }

    return;
}

//_________________________________________________
void ConversationWindow::_updateAvatar( void )
{
    Debug::Throw( "ConversationWindow::_updateAvatar.\n" );
    contactListModel_.setAvatar( connectionId_, AvatarPixmap( connectionId_.user() ) );
}

//_________________________________________________
void ConversationWindow::_installActions( void )
{
    Debug::Throw( "ConversationWindow::_installActions.\n" );

    // close window
    addAction( closeWindowAction_ = new QAction( IconEngine::get( IconNames::Close ), tr( "Close Window" ), this ) );
    closeWindowAction_->setShortcut( Qt::CTRL+Qt::Key_W );
    closeWindowAction_->setToolTip( tr( "Close window" ) );
    connect( closeWindowAction_, SIGNAL(triggered()), this, SLOT(close()) );

    // reload
    addAction( reloadAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Reload Messages" ), this ) );
    reloadAction_->setShortcut( QKeySequence::Refresh );
    connect( reloadAction_, SIGNAL(triggered()), SLOT(_reload()) );

    // send file action
    addAction( sendFileAction_ = new QAction( IconEngine::get( IconNames::Attach ), tr( "Send File" ), this ) );
    connect( sendFileAction_, SIGNAL(triggered()), this, SLOT(_selectFile()) );
    sendFileAction_->setEnabled( false );

    // history
    historyButton_ = new CustomToolButton();
    historyButton_->setIcon( IconEngine::get( IconNames::History ) );
    historyButton_->setText( tr( "Messages History" ) );
    historyButton_->setToolTip( tr( "Show messages history" ) );
    historyButton_->setMenu( historyMenu_ = new HistoryMenu( this ) );
    historyButton_->setPopupMode( QToolButton::InstantPopup );

    connect(historyMenu_, SIGNAL(delayChanged(HistoryMenu::Delay)), SLOT(_processDelayChanged(HistoryMenu::Delay)) );

}

//_________________________________________________
bool ConversationWindow::_appendMessage( const ChatMessage& message )
{
    Debug::Throw( "ConversationWindow::_appendMessage.\n" );

    // first check if message is in set
    ChatMessage::Set::iterator iter( messageSet_.find( message ) );
    if( iter != messageSet_.end() )
    {

        Debug::Throw( "ConversationWindow::_appendMessage - found existing message.\n" );

        // insert new message if text is not empty
        if( message.text().isEmpty() )
        {

            // remove from set
            messageSet_.erase( iter );

            // remove from list
            messageList_.removeOne( message );

        } else {

            // copy, adjust reception time and insert in set
            ChatMessage copy( message );
            copy.setReceptionTime( iter->receptionTime() );

            // remove old from set
            messageSet_.erase( iter );

            // insert in set
            messageSet_.insert( copy );

            // replace in list (should be faster than re-generating)
            const int index( messageList_.lastIndexOf( copy ) );
            if( index >= 0 ) messageList_[index] = copy;
            else {

                // should not happen
                messageList_ = messageSet_.toList();
                qSort( messageList_.begin(), messageList_.end(), ChatMessage::SortReceptionTimeFTor() );

            }

        }

        // restore replace iterator
        _updateReplaceIterator();

        return true;

    } else if( message.text().isEmpty() ) {

        // do nothing and return false if message is empty
        return false;

    } else {

        // insert in set
        messageSet_.insert( message );

        // try append at the back of the list
        if( messageList_.isEmpty() || messageList_.last().receptionTime() <= message.receptionTime() )
        {

            messageList_.append( message );
            _updateReplaceIterator();
            return false;

        } else {

            // regenerate sorted list from set
            messageList_ = messageSet_.toList();
            qSort( messageList_.begin(), messageList_.end(), ChatMessage::SortReceptionTimeFTor() );
            _updateReplaceIterator();
            return true;

        }

    }

}

//_________________________________________________
void ConversationWindow::_clearReplaceIterator( bool clearMessage )
{
    replaceIterator_ = messageList_.end();
    if( clearMessage ) replaceMessage_ = ChatMessage();
}

//_________________________________________________
void ConversationWindow::_updateStatusBar( void ) const
{ statusBar_->label().setTextAndUpdate( connectionId_.userStateString() ); }

//_________________________________________________
void ConversationWindow::_updateReplaceIterator( void )
{
    if( !replaceMessage_.isValid() ) replaceIterator_ = messageList_.end();
    else replaceIterator_ = qFind( messageList_.begin(), messageList_.end(), replaceMessage_ );
}

//_________________________________________________
void ConversationWindow::_printMessageList( void )
{ for( const auto& message:messageList_ ) _printMessage( message ); }

//_________________________________________________
void ConversationWindow::_printMessage( ChatMessage message )
{

    // do not print empty messages
    if( message.text().isEmpty() )
    {
        message.setBlockRange( 0, 0 );
        ChatMessage::Set::iterator iter( messageSet_.find( message ) );
        messageSet_.erase( iter );
        messageSet_.insert( message );
        return;
    }

    // append message to message board and store corresponding block range
    const auto blockRange = messageBoardWidget_->appendMessage( message );

    // update message
    ChatMessage::Set::iterator iter( messageSet_.find( message ) );
    if( iter != messageSet_.end() )
    { const_cast<ChatMessage&>( *iter ).setBlockRange( blockRange ); }

}

//______________________________________________________________________
int ConversationWindow::_numContacts( void ) const
{

    int out( 0 );
    for( const auto& connectionId:contactListModel_.get() )
    { if( connectionId != connectionId_ ) ++out; }

    return out;
}

//______________________________________________________________________
ConnectionId::List ConversationWindow::_contacts( void ) const
{
    ConnectionId::List out;
    for( const auto& connectionId:contactListModel_.get() )
    { if( connectionId != connectionId_ ) out.append( connectionId ); }
    return out;
}

//______________________________________________________________________
QStringList ConversationWindow::_contactNames( void ) const
{
    QStringList out;
    for( const auto& connectionId:contactListModel_.get() )
    { if( connectionId != connectionId_ ) out.append( connectionId.user() ); }
    return out;
}

//______________________________________________________________________
QStringList ConversationWindow::_onlineContactNames( void ) const
{
    QStringList out;
    for( const auto& connectionId:contactListModel_.get() )
    {
        if( connectionId != connectionId_ && connectionId.userState() == ConnectionId::Online )
        { out.append( connectionId.user() ); }
    }
    return out;
}
