#ifndef ConversationWindow_h
#define ConversationWindow_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "BaseFileInfo.h"
#include "BaseMainWindow.h"
#include "ChatMessage.h"
#include "Counter.h"
#include "ConnectionId.h"
#include "ContactListModel.h"
#include "ClientInterface.h"
#include "Emoticon.h"
#include "File.h"
#include "Key.h"
#include "MessageBoardWidget.h"
#include "SoundNotifications.h"

#include <QActionGroup>
#include <QBasicTimer>
#include <QLayout>
#include <QSplitter>
#include <QToolButton>

class ContactList;
class HistoryMenu;
class MessageEditorWidget;
class ProgressStatusBar;
class FileTransferManager;

class ConversationWindow: public BaseMainWindow, public Counter, public Base::Key
{

    //* Qt meta object declaration
    Q_OBJECT

    public:

    //* constructor
    ConversationWindow( QWidget* = nullptr );

    //*@name accesors
    //@{

    //* message editor
    MessageEditorWidget& messageEditorWidget( void ) const
    { return *messageEditorWidget_; }

    //* file transfer manager
    FileTransferManager& fileTransferManager( void ) const
    { return *fileTransferManager_; }

    //* status bar
    ProgressStatusBar& statusBar( void ) const
    { return *statusBar_; }

    //* connection ids
    ConnectionId connectionId( void ) const
    { return connectionId_; }

    //* contacts
    ConnectionId::List contacts( void ) const
    { return contactListModel_.get(); }

    //* return true if given contact is registered to this window
    bool hasContact( ConnectionId contact ) const
    { return contactListModel_.index( contact ).isValid(); }

    //* return true if given contact is registered to this window
    bool hasContact( QString contactName ) const
    { return _contactNames().contains( contactName ); }

    //@}

    //*@name modifiers
    //@{

    //* set contacts
    void setContacts( const ConnectionId::List& );

    //* id
    void setConnectionId( ConnectionId );

    //@}

    //*@name actions
    //@{

    //* close window action
    QAction& closeWindowAction( void ) const
    { return *closeWindowAction_; }

    //* reload action
    QAction& reloadAction( void ) const
    { return *reloadAction_; }

    //* send file action
    QAction& sendFileAction( void ) const
    { return *sendFileAction_; }

    //* history menu
    HistoryMenu& historyMenu( void ) const
    { return *historyMenu_; }

    //@}

    //*@name event filters
    //@{
    virtual bool eventFilter( QObject*, QEvent* );
    virtual bool messageEditorEventFilter( QObject*, QEvent* );
    virtual bool messageBoardEventFilter( QObject*, QEvent* );
    //@}

    Q_SIGNALS:

    //* request messages
    void requestMessages( QStringList );

    //* request messages
    void requestMessages( QStringList, HistoryMenu::Delay );

    //* request messages
    void requestLocalMessages( QStringList );

    //* request messages
    void requestLocalMessages( QStringList, HistoryMenu::Delay );

    //* send a new message
    void newMessage( ChatMessage );

    //* send file request
    void requestFileTransfer( QStringList, BaseFileInfo );

    //* accept file transfer
    void acceptFileTransfer( ConnectionId, BaseFileInfo, File );

    //* den file transfer
    void denyFileTransfer( ConnectionId, BaseFileInfo );

    //* send user typing
    void userTyping( QStringList );

    //* new notification
    void newTextNotification( QString );

    //* new notification
    void newSoundNotification( SoundNotifications::Type );

    public Q_SLOTS:

    //* update connectionId
    void updateContact( ConnectionId, ConnectionId);

    //* update avatar
    void updateAvatar( ConnectionId );

    //* clear list of contacts
    /**
    contacts are not removed from the list.
    they are marked offline instead. Will be updated by
    information provided by the client interface
    */
    void clearContactList( void );

    //* update busy state
    void updateBusyState( bool, int );

    //* new message recieved
    void printMessage( ChatMessage );

    //* clear local messages
    void clearMessageList( QStringList );

    //* notify that a user is typing
    void notifyUserTyping( ConnectionId );

    //* notify file transfer request
    void notifyFileTransferRequest( ConnectionId, BaseFileInfo );

    //* notify file transfer cancelled
    void notifyFileTransferCancelled( ConnectionId, BaseFileInfo );

    protected:

    //* timer event
    virtual void timerEvent( QTimerEvent* );

    protected Q_SLOTS:

    //* send message from editor
    void _sendMessage( void );

    //* process delay changed
    void _processDelayChanged( HistoryMenu::Delay );

    //* process reload
    void _reload( void );

    //* insert emoticon
    void _insertEmoticon( Emoticon );

    //* editor modified
    void _editorModified( void );

    //* edit context menu message
    void _editContextMenuMessage( void );

    //* remove context menu message
    void _removeContextMenuMessage( void );

    //* remove replace message
    void _removeReplaceMessage( void );

    //* cancel message edition
    void _cancelMessageEdition( void );

    //* select file to send
    void _selectFile( void );

    //* view link
    void _openLink( void );

    //* view link
    void _openLink( QString );

    //* show file transfer
    void _showFileTransfers( void );

    private Q_SLOTS:

    //* configuration
    void _updateConfiguration( void );

    //* avatar
    void _updateAvatar( void );

    private:

    //* install actions
    void _installActions( void );

    //* add message to internal storing
    /** returns true if message list order was changed */
    bool _appendMessage( const ChatMessage& );

    //* clear replace iterator
    void _clearReplaceIterator( bool clearMessage = true );

    //* update status bar
    void _updateStatusBar( void ) const;

    //* update replace iterator so that it matches message
    void _updateReplaceIterator( void );

    //* print entire message list
    void _printMessageList( void );

    //* print message
    void _printMessage( ChatMessage );

    //* number of contacts (excluding this)
    int _numContacts( void ) const;

    //* get list of contacts excluding this
    ConnectionId::List _contacts( void ) const;

    //* get list of contact names (excluding this)
    QStringList _contactNames( void ) const;

    //* get list of online contact names (excluding this)
    QStringList _onlineContactNames( void ) const;

    //* id
    ConnectionId connectionId_;

    //* splitter
    QSplitter* splitter_ = nullptr;

    //* message board
    MessageBoardWidget* messageBoardWidget_ = nullptr;

    //* line editor
    MessageEditorWidget* messageEditorWidget_ = nullptr;

    //* file transfer manager
    FileTransferManager* fileTransferManager_ = nullptr;

    //* status bar
    ProgressStatusBar* statusBar_ = nullptr;

    //* main layout
    QBoxLayout* mainLayout_ = nullptr;

    //* contact list
    ContactList* contactList_ = nullptr;

    //* contact model
    ContactListModel contactListModel_;

    //* chat messages list
    ChatMessage::Set messageSet_;
    ChatMessage::List messageList_;

    //* message replacement
    ChatMessage::List::iterator replaceIterator_;
    ChatMessage replaceMessage_;

    //* message that matches context menu event
    ChatMessage contextMenuMessage_;

    //*@name actions
    //@{

    //* close window action
    QAction* closeWindowAction_ = nullptr;

    //* history action
    QToolButton* historyButton_ = nullptr;

    //* history menu
    HistoryMenu* historyMenu_ = nullptr;

    //* reload action
    QAction* reloadAction_ = nullptr;

    //* send file action
    QAction* sendFileAction_ = nullptr;

    //@}

    //* true if mainwindow is busy (recieving logs)
    /** no notifications will be sent */
    bool busy_ = false;

    //*@name user typing
    //@{

    //* enable user typing
    bool notifyUserTyping_ = true;

    //* user typing timeout
    int userTypingTimeout_ = 500;

    //* this user typing
    QBasicTimer thisUserTypingTimer_;

    //* user typing timer
    QBasicTimer userTypingTimer_;

    //@}

};

#endif
