/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "FileTransferManager.h"
#include "MimeTypeIconProvider.h"
#include "IconEngine.h"
#include "BaseIconNames.h"

#include <QLabel>
#include <QToolButton>

//________________________________________________________________________
FileTransferManager::FileTransferManager( QWidget* parent ):
    QWidget( parent )
{
    Debug::Throw( "FileTransferManager::FileTransferManager.\n" );

    _installActions();

    iconProvider_ = new MimeTypeIconProvider( this );

    setLayout( new QVBoxLayout() );
    layout()->setMargin(0);
    layout()->setSpacing(0);

    titleLabel_ = new QLabel( tr( "File transfers" ), this );
    titleLabel_->setAlignment( Qt::AlignCenter );
    titleLabel_->setMargin( 4 );
    layout()->addWidget( titleLabel_ );

    layout()->addWidget( scrollArea_ = new QScrollArea( this ) );
    scrollArea_->setWidgetResizable ( true );
    scrollArea_->viewport()->setAutoFillBackground( true );
    scrollArea_->viewport()->setBackgroundRole( QPalette::Base );
    scrollArea_->viewport()->setForegroundRole( QPalette::Text );

    mainWidget_ = new QWidget( scrollArea_ );

    QVBoxLayout* vLayout = new QVBoxLayout();
    vLayout->setSpacing(0);
    vLayout->setMargin(0);
    mainWidget_->setLayout( vLayout );

    vLayout->addLayout( mainLayout_ = new QVBoxLayout() );
    vLayout->addStretch();

    mainLayout_->setSpacing(0);
    mainLayout_->setMargin(0);

    scrollArea_->setWidget( mainWidget_ );
    mainWidget_->show();

}

//________________________________________________________________________
FileTransferWidget* FileTransferManager::fileTransfer( const BaseFileInfo& fileInfo, FileTransferWidget::Mode mode ) const
{

    // loop over existing file transfer widgets and find matching
    for( const auto& widget:mainWidget_->findChildren<FileTransferWidget*>() )
    {
        if(
            widget->fileInfo() == fileInfo
            && widget->mode() == mode
            && widget->status() == FileTransferWidget::Starting )
        { return widget; }
    }

    return nullptr;

}

//________________________________________________________________________
FileTransferWidget* FileTransferManager::addFileTransfer( const BaseFileInfo& fileInfo, FileTransferWidget::Mode mode )
{

    const bool isFirst( mainWidget_->findChildren<FileTransferWidget*>().isEmpty() );

    // add new widget
    FileTransferWidget* widget = new FileTransferWidget( mainWidget_, mode );
    widget->setIcon( iconProvider_->icon( fileInfo.file().extension() ) );
    widget->setFileInfo( fileInfo );
    widget->separator().setVisible( !isFirst );
    connect( widget, SIGNAL(fileTransferCancelled(BaseFileInfo)), this, SIGNAL(fileTransferCancelled(BaseFileInfo)) );

    mainLayout_->addWidget( widget );
    if( !isFirst )
    {

        // make sure new widget is visible
        widget->show();
        mainWidget_->adjustSize();
        const QPoint position( widget->geometry().bottomLeft() );
        scrollArea_->ensureVisible( position.x(), position.y() );

    } else {

        // setup minimum size hint
        const QSize size( widget->minimumSizeHint() );
        sizeHint_.setWidth( size.width() + 2*scrollArea_->frameWidth() );
        sizeHint_.setHeight( size.height() + titleLabel_->sizeHint().height() + 2*scrollArea_->frameWidth() );

    }

    connect( widget, SIGNAL(destroyed()), this, SLOT(_checkEmptyList()) );
    connect( widget, SIGNAL(destroyed()), this, SLOT(_updateSeparators()) );

    visibilityAction_->setEnabled( true );
    emit fileTransferAdded(fileInfo);

    return widget;

}

//________________________________________________________________________
void FileTransferManager::setStatus( const BaseFileInfo& fileInfo, FileTransferWidget::Status status, const QString& string )
{
    for( const auto& widget:mainWidget_->findChildren<FileTransferWidget*>() )
    {
        if( widget->fileInfo() == fileInfo
            && widget->status() != FileTransferWidget::Completed
            && widget->status() != FileTransferWidget::Cancelled )
        { widget->setStatus( status, string ); }
    }
}

//________________________________________________________________________
void FileTransferManager::_checkEmptyList( void )
{
    Debug::Throw( "FileTransferManager::_checkEmptyList.\n" );
    if( mainWidget_->findChildren<FileTransferWidget*>().isEmpty() )
    {
        visibilityAction_->setEnabled( false );
        hide();
    }
}

//________________________________________________________________________
void FileTransferManager::_updateSeparators( void )
{
    Debug::Throw( "FileTransferManager::_updateSeparators.\n" );
    bool isFirst = true;
    for( const auto& widget:mainWidget_->findChildren<FileTransferWidget*>() )
    {
        widget->separator().setVisible( !isFirst );
        isFirst = false;
    }

}

//________________________________________________________________________
void FileTransferManager::_installActions( void )
{
    Debug::Throw( "FileTransferManager::_installActions.\n" );
    addAction( visibilityAction_ = new QAction( tr( "Show File Transfers" ), this ) );
    visibilityAction_->setEnabled( false );
}
