/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "FileTransferWidget.h"
#include "Command.h"
#include "BaseIconNames.h"
#include "IconEngine.h"
#include "IconSize.h"
#include "InformationDialog.h"
#include "OpenWithDialog.h"
#include "XmlOptions.h"

#include <QLayout>
#include <QDesktopServices>
#include <QUrl>

//________________________________________________________________________
FileTransferWidget::FileTransferWidget( QWidget* parent, Mode mode ):
    QWidget( parent ),
    mode_( mode )
{
    Debug::Throw( "FileTransferWidget::FileTransferWidget.\n" );

    QGridLayout* layout = new QGridLayout();
    setLayout( layout );
    layout->setSpacing( 5 );
    layout->setMargin( 5 );

    layout->addWidget( separator_ = new QFrame( this ), 0, 0, 1, 4 );
    separator_->setFrameStyle( QFrame::HLine );

    layout->addWidget( iconLabel_ = new QLabel( this ), 1, 0, 3, 1 );
    iconLabel_->setAlignment( Qt::AlignTop );
    iconLabel_->hide();

    layout->addWidget( fileLabel_ = new QLabel( this ), 1, 1, 1, 3 );
    layout->addWidget( progressBar_ = new QProgressBar( this ), 2, 1, 1, 1 );

    _installButtons();
    _updateButtons();

    if( mode_ == Reception ) layout->addWidget( openWithButton_, 2, 2, 1, 1 );
    if( mode_ == Emission ) layout->addWidget( resendButton_, 2, 2, 1, 1 );
    layout->addWidget( cancelButton_, 2, 3, 1, 1 );

    layout->addWidget( statusLabel_ = new QLabel( this ), 3, 1, 1, 3 );

}

//________________________________________________________________________
void FileTransferWidget::setIcon( const QIcon& icon )
{
    if( !icon.isNull() )
    {
        iconLabel_->setPixmap( icon.pixmap( QSize( 32, 32 ) ) );
        iconLabel_->show();
    } else iconLabel_->hide();
}

//________________________________________________________________________
void FileTransferWidget::setFileInfo( const BaseFileInfo& fileInfo )
{

    // store
    fileInfo_ = fileInfo;

    // adjust progress bar maximum
    progressBar_->setMaximum( fileInfo.size() );

    // update file label
    const QString localName = fileInfo.file().localName();
    const QString sizeString = File::sizeString( fileInfo.size() );
    const QString message = tr( "%1 (size: %2)" ).arg( localName ).arg( sizeString );
    fileLabel_->setText( message );

    // update status
    setStatus( Starting );

}

//________________________________________________________________________
void FileTransferWidget::setDestination( const File& destination )
{
    destination_ = destination;
    openWithButton_->setVisible( mode_ == Reception && !destination_.isEmpty() );
}

//________________________________________________________________________
void FileTransferWidget::setStatus( FileTransferWidget::Status status, const QString& string )
{

    Debug::Throw() << "FileTransferWidget::setStatus - status: " << status << endl;

    if( !string.isEmpty() ) _setStatusString( string );
    if( status_ == status ) return;
    status_ = status;

    switch( status )
    {
        case Starting:
        if( string.isEmpty() ) _setStatusString( mode_ == Emission ? tr( "Waiting for recipient..."):tr( "Wainting for sender..." ) );
        progressBar_->setValue(0);
        break;

        case Transferring:
        if( string.isEmpty() ) _setStatusString( mode_ == Emission ? tr( "Sending..."):tr( "Receiving..." ) );
        break;

        case Completed:
        if( string.isEmpty() ) _setStatusString( tr( "Done." ) );
        if( mode_ == Emission && XmlOptions::get().get<bool>( "AUTOCLOSE_EMISSION_TRANSFERS" ) )
        { deleteLater(); }

        break;

        case Cancelled:
        if( string.isEmpty() ) _setStatusString( tr( "Cancelled." ) );
        emit fileTransferCancelled(fileInfo_);
        break;

        case Failed:
        if( string.isEmpty() ) _setStatusString( tr( "Transfer failed." ) );
        break;

        default:
        break;

    }

    _updateButtons();

    return;
}

//________________________________________________________________________
void FileTransferWidget::processFileChunk( FileChunk fileChunk )
{
    if( fileChunk.filename() != fileInfo_.file() ) return;

    // initialize
    if( fileChunk.flags() & FileChunk::First )
    {
        progressBar_->setValue(0);
        setStatus( Transferring );
    }

    // append file chunk
    if( !( fileChunk.flags() & FileChunk::Hash ) )
    { progressBar_->setValue( progressBar_->value() + fileChunk.size() ); }

    // finish
    if( fileChunk.flags() & FileChunk::Last )
    {
        progressBar_->setValue( progressBar_->maximum() );
        if( status_ == Transferring ) setStatus( Completed );
    }

}

//________________________________________________________________________
void FileTransferWidget::setTransferFailed( void )
{ setStatus( Failed ); }

//________________________________________________________________________
void FileTransferWidget::_updateButtons( void )
{
    if( mode_ == Reception ) openWithButton_->setEnabled( status_ == Completed );
    if( mode_ == Emission )
    {
        resendButton_->setEnabled( status_ == Cancelled );
        resendButton_->setVisible( status_ == Cancelled );
    }

    if( status_ == Starting || status_ == Transferring ) cancelButton_->setToolTip( tr( "Cancel file transfer" ) );
    else cancelButton_->setToolTip( tr( "Remove from the list" ) );

    // update cancel button icon when transfer is completed
    if( status_ == Completed ) cancelButton_->setIcon(  IconEngine::get( IconNames::DialogOk  ) );

}

//________________________________________________________________________
void FileTransferWidget::_cancel( void )
{
    if( status_ == Starting || status_ == Transferring ) setStatus( Cancelled );
    else deleteLater();
}

//________________________________________________________________________
void FileTransferWidget::_resend( void )
{
    if( status_ == Cancelled && mode_ == Emission )
    {
        emit fileTransferResent(recipients_, fileInfo_);
        deleteLater();
    }
}

//________________________________________________________________________
void FileTransferWidget::_open( void )
{
    if( mode_ != Reception || destination_.isEmpty() ) return;

    OpenWithDialog dialog( this );
    dialog.setWindowTitle( tr( "Open File - Homechat" ) );
    dialog.setFile( destination_ );
    dialog.setOptionName( "OPEN_FILE_APPLICATIONS" );
    dialog.realizeWidget();
    dialog.exec();

}

//________________________________________________________________________
void FileTransferWidget::_setStatusString( const QString& status )
{ statusLabel_->setText( status ); }

//________________________________________________________________________
void FileTransferWidget::_installButtons( void )
{

    Debug::Throw( "FileTransferWidget::_installButtons.\n" );

    // open
    if( mode_ == Reception )
    {
        openWithButton_ = new QToolButton( this );
        openWithButton_->setIcon( IconEngine::get( IconNames::Open ) );
        openWithButton_->setToolTip( tr( "Open file" ) );
        openWithButton_->setAutoRaise( true );
        openWithButton_->setToolButtonStyle(Qt::ToolButtonIconOnly );
        connect( openWithButton_, SIGNAL(clicked()), SLOT(_open()) );
    }

    // resend
    if( mode_ == Emission )
    {
        resendButton_ = new QToolButton( this );
        resendButton_->setIcon( IconEngine::get( IconNames::Reload ) );
        resendButton_->setToolTip( tr( "Resend file" ) );
        resendButton_->setAutoRaise( true );
        resendButton_->setToolButtonStyle(Qt::ToolButtonIconOnly );
        connect( resendButton_, SIGNAL(clicked()), SLOT(_resend()) );
    }

    // cancel
    cancelButton_ = new QToolButton( this );
    cancelButton_->setIcon( IconEngine::get( IconNames::DialogClose ) );
    cancelButton_->setToolTip( tr( "Cancel file transfer" ) );
    cancelButton_->setAutoRaise( true );
    cancelButton_->setToolButtonStyle(Qt::ToolButtonIconOnly );
    connect( cancelButton_, SIGNAL(clicked()), SLOT(_cancel()) );

}
