#ifndef MessageBoardWidget_h
#define MessageBoardWidget_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ChatMessage.h"
#include "TextEditor.h"
#include "TimeStamp.h"

#include <QTextBlockFormat>
#include <QTextCharFormat>


class EmoticonProvider;
class HtmlHighlighter;
class SyntaxHighlighter;

class MessageBoardWidget: public TextEditor
{

    Q_OBJECT

    public:

    //* constructor
    MessageBoardWidget( QWidget* = nullptr );

    //*@name accessors
    //@{

    //* html
    HtmlHighlighter& htmlHighlighter( void ) const
    { return *htmlHighlighter_; }

    //* emoticons
    EmoticonProvider& emoticonProvider( void ) const
    { return *emoticonProvider_; }

    //@}

    //*@name modifiers
    //@{

    //* appends message, return block range
    ChatMessage::BlockRange appendMessage( const ChatMessage& );

    //* commit pending messages
    void commit( void );

    //* install actions in context menu
    virtual void installContextMenuActions( BaseContextMenu*, bool = true );

    //@}

    //*@name actions
    //@{

    //* edit message action
    QAction& editMessageAction( void ) const
    { return *editMessageAction_; }

    //* remove message action
    QAction& removeMessageAction( void ) const
    { return *removeMessageAction_; }

    //* open link
    QAction& openLinkAction( void ) const
    { return *openLinkAction_; }

    //@}

    public Q_SLOTS:

    //* clear
    /**
    need to bypass TextEditor::clear,
    which is disabled in read-only mode
    */
    virtual void clear( void );

    protected Q_SLOTS:

    //* slider
    void _verticalScrollBarMoved( int );

    protected:

    //* keypress event
    virtual void keyPressEvent( QKeyEvent* );

    //* wheel events
    virtual void wheelEvent( QWheelEvent* );

    //* change event
    virtual void changeEvent( QEvent* );

    //* create mime data from selection
    virtual QMimeData* createMimeDataFromSelection( void ) const;

    //* lock
    bool _locked( void ) const
    { return locked_; }

    private:

    //* install actions
    void _installActions( void );

    //* block count
    int _blockCount( void ) const
    { return document()->blockCount() + buffer_.size(); }

    //* append date
    void _appendDate( const TimeStamp& );

    //* append message
    void _appendMessage( const QString& );

    //* append message
    void _appendTransferNotification( const QString& );

    //* true if locked
    bool locked_ = false;

    //* syntax highlighter
    SyntaxHighlighter* syntaxHighlighter_ = nullptr;

    //* html highlighter
    HtmlHighlighter* htmlHighlighter_ = nullptr;

    //* emoticon provider
    EmoticonProvider* emoticonProvider_ = nullptr;

    //* default block format
    QTextBlockFormat defaultBlockFormat_;

    //* default char format
    QTextCharFormat defaultCharFormat_;

    //* last appended message
    ChatMessage lastMessage_;

    //* buffer
    QStringList buffer_;

    //*@name actions
    //@{

    //* edit message action
    QAction* editMessageAction_ = nullptr;

    //* remove message action
    QAction* removeMessageAction_ = nullptr;

    //* open link
    QAction* openLinkAction_ = nullptr;

    //@}

};

#endif
