/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "MessageEditorWidget.h"

#include "BaseContextMenu.h"
#include "Color.h"
#include "Debug.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "TextFormat.h"
#include "Singleton.h"
#include "XmlOptions.h"

#if USE_ASPELL
#include "SuggestionMenu.h"
#endif

#include <QTextCharFormat>
#include <QInputMethodEvent>

//______________________________________________________
MessageEditorWidget::MessageEditorWidget( QWidget* parent ):
    LineEditor( parent )
{
    Debug::Throw( "MessageEditorWidget::MessageEditorWidget.\n" );

    _installActions();

    connect( this, SIGNAL(textChanged(QString)), this, SLOT(_spellcheck()) );

    connect( Singleton::get().application(), SIGNAL(spellCheckConfigurationChanged()), SLOT(_updateSpellCheckConfiguration()) );
    _updateSpellCheckConfiguration();

}

//______________________________________________________________________________
void MessageEditorWidget::installContextMenuActions( BaseContextMenu* menu )
{

    Debug::Throw( "MessageEditorWidget::installContextMenuActions.\n" );

    // retrieve default context menu
    // second argument is to remove un-necessary actions
    LineEditor::installContextMenuActions( menu );

    // add specific actions
    menu->addSeparator();
    menu->addAction( removeMessageAction_ );
    menu->addAction( cancelMessageEditionAction_ );

    return;
}

//_______________________________________________________
void MessageEditorWidget::_toggleAutoSpell( bool state )
{
    #if USE_ASPELL
    Debug::Throw( "MessageEditorWidget::_toggleAutoSpell.\n" );

    // propagate to textHighlight
    spellParser_.setEnabled( state );
    XmlOptions::get().set<bool>( "AUTOSPELL_ENABLED", state );

    if( !state )
    {

        misspelledWords_.clear();
        QList<QInputMethodEvent::Attribute> attributes;
        QInputMethodEvent event(QString(), attributes);
        inputMethodEvent( &event );

    } else {

        _spellcheck();

    }

    return;
    #else
    Q_UNUSED( state )
    #endif
}

//_________________________________________________
void MessageEditorWidget::_spellcheck( void )
{
    Debug::Throw() << "MessageEditorWidget::_spellcheck." << endl;

    #if USE_ASPELL

    // check if parser is enabled
    if( !spellParser_.isEnabled() ) return;

    // get list of misspelled words and compare
    const SpellCheck::Word::Set& words( spellParser_.parse( text() ) );
    if( words == misspelledWords_ ) return;

    // update highlighting
    misspelledWords_ = words;
    QList<QInputMethodEvent::Attribute> attributes;
    for( const auto& word:misspelledWords_ )
    {
        QInputMethodEvent::Attribute attribute(
            QInputMethodEvent::TextFormat,
            word.position() - cursorPosition(), word.length(),
            spellcheckFormat_ );
        attributes.append( attribute );
    }

    QInputMethodEvent event(QString(), attributes);
    inputMethodEvent( &event );

    #endif

}

//________________________________________________
void MessageEditorWidget::contextMenuEvent( QContextMenuEvent* event )
{

    Debug::Throw( "MessageEditorWidget::contextMenuEvent.\n" );

    if( _autoSpellContextEvent( event ) ) return;
    else {

        LineEditor::contextMenuEvent( event );
        return;

    }

}

//________________________________________________
bool MessageEditorWidget::_autoSpellContextEvent( QContextMenuEvent* event )
{
    Debug::Throw( "MessageEditorWidget::_autoSpellContextEvent.\n" );

    #if USE_ASPELL

    // check autospell enability
    if( !spellParser_.isEnabled() ) return false;

    // cursor position
    const int position( cursorPositionAt( event->pos() ) );
    auto iter = std::find_if(
        misspelledWords_.begin(),
        misspelledWords_.end(),
        SpellCheck::Word::AtPositionFTor( position ) );

    if( iter == misspelledWords_.end() )
    { return false; }

    const SpellCheck::Word& word( *iter );
    if( spellParser_.interface().isWordIgnored( word ) ) return false;

    setSelection( word.position(), word.length() );

    // create suggestion menu
    SpellCheck::SuggestionMenu menu( this, word, isReadOnly() );
    menu.interface().setFilter( spellParser_.interface().filter() );
    menu.interface().setDictionary( spellParser_.interface().dictionary() );

    // set connections
    connect( &menu, SIGNAL(ignoreWord(QString)), SLOT(_ignoreMisspelledWord(QString)) );
    connect( &menu, SIGNAL(suggestionSelected(QString)), SLOT(_replaceMisspelledSelection(QString)) );

    // insert dictionary action and separator
    menu.addAction( dictionaryMenuAction_ );
    menu.addSeparator();

    // execute
    menu.exec( event->globalPos() );
    return true;

    #else
    Q_UNUSED( event )
    return false;
    #endif

}

//__________________________________________________
void MessageEditorWidget::_ignoreMisspelledWord( QString word )
{
    Debug::Throw() << "MessageEditorWidget::_ignoreMisspelledWord - word: " << word << endl;
    #if USE_ASPELL
    spellParser_.interface().ignoreWord( word );
    _spellcheck();
    #endif
    return;

}

//__________________________________________________
void MessageEditorWidget::_replaceMisspelledSelection( QString word )
{

    #if USE_ASPELL
    Debug::Throw() << "MessageEditorWidget::_replaceMisspelledSelection - word: " << word << endl;
    insert( word );
    #else
    Q_UNUSED( word )
    #endif
    return;

}

//_______________________________________
void MessageEditorWidget::_selectDictionary( const QString& dictionary )
{
    Debug::Throw( "MessageEditorWidget::_selectDictionary.\n" );

    #if USE_ASPELL

    // local reference to interface
    SpellCheck::SpellInterface& interface( spellParser_.interface() );

    if( dictionary == interface.dictionary() || !interface.hasDictionary( dictionary ) ) return;

    // update interface
    interface.setDictionary( dictionary );
    dictionaryMenu_->select( dictionary );

    // rehighlight if needed
    _spellcheck();

    #else
    Q_UNUSED( dictionary )
    #endif

    return;

}

//_______________________________________________
void MessageEditorWidget::_updateSpellCheckConfiguration( void )
{

    #if USE_ASPELL

    // store local reference to spell interface
    SpellCheck::SpellInterface& interface( spellParser_.interface() );

    // load default filter and dictionaries
    interface.setFilter( XmlOptions::get().raw("DICTIONARY_FILTER") );
    interface.setDictionary( XmlOptions::get().raw( "DICTIONARY" ) );

    // read options
    const QColor color( XmlOptions::get().get<Base::Color>("AUTOSPELL_COLOR") );
    const Format::TextFormatFlags format = (Format::TextFormatFlags)XmlOptions::get().get<int>("AUTOSPELL_FONT_FORMAT");

    // reset format
    spellcheckFormat_ = QTextCharFormat();
    spellcheckFormat_.setFontWeight( (format&Format::Bold) ? QFont::Bold : QFont::Normal );
    spellcheckFormat_.setFontItalic( format&Format::Italic );
    spellcheckFormat_.setFontUnderline( format&Format::Underline );
    spellcheckFormat_.setFontOverline( format&Format::Overline );
    spellcheckFormat_.setFontStrikeOut( format&Format::Strike );
    if( color.isValid() ) spellcheckFormat_.setForeground( color );

    // update dictionary
    dictionaryMenu_->reset();
    const QString dictionary( XmlOptions::get().raw( "DICTIONARY" ) );
    dictionaryMenu_->select( dictionary );

    // update autospell action
    autoSpellAction_->setEnabled( color.isValid() );
    autoSpellAction_->setChecked( XmlOptions::get().get<bool>( "AUTOSPELL_ENABLED" ) && color.isValid() );
    _spellcheck();
    #endif

}

//_______________________________________________
void MessageEditorWidget::_installActions( void )
{

    Debug::Throw( "MessageEditorWidget::_installActions.\n" );

    #if USE_ASPELL

    dictionaryMenu_ = new SpellCheck::DictionaryMenu( this );
    dictionaryMenuAction_ = dictionaryMenu_->menuAction();
    connect( dictionaryMenu_, SIGNAL(selectionChanged(QString)), SLOT(_selectDictionary(QString)) );

    // autospell
    addAction( autoSpellAction_ = new QAction( IconEngine::get( IconNames::SpellCheck ), "Automatic Spell Checking", this ) );
    autoSpellAction_->setShortcut( Qt::Key_F6 );
    autoSpellAction_->setShortcutContext( Qt::WidgetShortcut );
    autoSpellAction_->setCheckable( true );
    connect( autoSpellAction_, SIGNAL(toggled(bool)), SLOT(_toggleAutoSpell(bool)) );
    #endif

    // remove message actions
    addAction( removeMessageAction_ = new QAction( IconEngine::get( IconNames::Cut ), tr( "Remove Message" ), this ) );
    removeMessageAction_->setShortcut( QKeySequence::Cut );

    // edit messsage actions
    addAction( cancelMessageEditionAction_ = new QAction( IconEngine::get( IconNames::DialogCancel ), tr( "Cancel Message Edition" ), this ) );

}
