/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ServerInterface.h"

#include "AvatarImage.h"
#include "ChatMessageMigration.h"
#include "File.h"
#include "Singleton.h"
#include "XmlOptions.h"

#include <QCoreApplication>
#include <QDateTime>
#include <QDir>
#include <QHostInfo>

//____________________________________________
ServerInterface::ServerInterface( QObject* parent, const CommandLineParser& parser ):
    QObject( parent ),
    Counter( "ServerInterface" ),
    parser_( parser )
{

    // debug level
    if( parser_.hasFlag( "--verbose" ) ) debugLevel_ = 0;
    else debugLevel_ = 1;

    messages_.setDebugLevel( debugLevel_ );

    // configuration
    connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    connect( qApp, SIGNAL(aboutToQuit()), SLOT(_saveConfiguration()) );
    _updateConfiguration();
}

//____________________________________________
void ServerInterface::initialize( void )
{
    Debug::Throw( "ServerInterface::initialize.\n" );

    if( !server_ )
    {
        // create server
        server_ = new QTcpServer( this );
        connect( server_, SIGNAL(newConnection()), SLOT(_newConnection()) );
    }

    // host and port from arguments
    if( parser_.hasOption( "--chat-host" ) ) host_ = QHostAddress( parser_.option( "--chat-host" ) );
    if( parser_.hasOption( "--chat-port" ) ) port_ = parser_.option( "--chat-port" ).toUInt();

    // initialize
    Debug::Throw() << "ServerInterface::initialize - Listening to " << host_.toString() << ":" << port_ << endl;

    if( !server_->listen( host_, port_ ) )
    {  Debug::Throw(0) << "ServerInterface::initialize - Failed to listen to " << host_.toString() << ":" << port_ << endl; }

}

//______________________________________
bool ServerInterface::writeMessages( void )
{

    Debug::Throw( "ServerInterface::write.\n" );
    if( !messages_.write() )
    {

        // send a message to all clients
        ChatCommand command( ConnectionId( "server" ), ChatCommand::Message );
        command.setCreationTime( TimeStamp::now() );
        command.setReceptionTime( TimeStamp::now() );
        command.setText( tr( "Unable to write messages to disc." ) );
        _broadcast( command );

        command.setCreationTime( TimeStamp::now()+1 );
        command.setReceptionTime( TimeStamp::now()+1 );
        command.setText( tr( "Please check write permissions on server host." ) );
        _broadcast( command );

        return false;

    } else return true;

}

//_____________________________________________________
void ServerInterface::timerEvent( QTimerEvent* event )
{
    if( event->timerId() == autoSaveTimer_.timerId() )
    {

        autoSaveTimer_.stop();
        writeMessages();

    } else if( event->timerId() == fullSaveTimer_.timerId() ) {

        Debug::Throw( debugLevel_ )
            << "ServerInterface::_timerEvent -"
            << " time: " << TimeStamp::now().toString()
            << " performing full merge" << endl;

        // write messages
        if( writeMessages() )
        {
            // merge
            messages_.mergeMessages();

            // clear new messages
            messages_.clearNewMessages();

        }

    } else return QObject::timerEvent( event );
}

//_____________________________________________________
void ServerInterface::_newConnection()
{

    // check pending connection
    while( server_->hasPendingConnections() )
    {

        // create client from pending connection
        QTcpSocket* socket( server_->nextPendingConnection() );
        Debug::Throw( debugLevel_ ) << "ServerInterface::_newConnection - " << socket << endl;

        ChatClient *client( new ChatClient( this, socket ) );
        connect( &client->socket(), SIGNAL(disconnected()), SLOT(_clientDisconnected()) );

        connect( client, SIGNAL(avatarAvailable(Avatar)), SLOT(_processAvatar(Avatar)) );
        connect( client, SIGNAL(commandAvailable(ChatCommand)), SLOT(_processCommand(ChatCommand)) );
        connect( client, SIGNAL(fileChunkAvailable(FileChunk)), SLOT(_broadcast(FileChunk)) );

        connectedClients_.append( client );

    }

}

//_____________________________________________________
void ServerInterface::_clientDisconnected( void )
{

    // remove from accepted client map
    {
        ClientList::iterator iter;
        while( ( iter = std::find_if(  acceptedClients_.begin(), acceptedClients_.end(), SameStateFTor( QAbstractSocket::UnconnectedState ) )  ) != acceptedClients_.end() )
        {
            Debug::Throw( debugLevel_ ) << "ServerInterface::_clientDisconnected - user:" << (*iter)->connectionId().user() << endl;

            // update id status
            ConnectionId connectionId( (*iter)->connectionId() );
            connectionId.setConnectionState( ConnectionId::Disconnected );
            connectionId.setDisconnectionTime( TimeStamp::now() );

            // broadcast client has disconnected and status changed
            _broadcast( ChatCommand( connectionId, ChatCommand::ContactOffline ), (*iter) );

            // erase from map of accepted clients
            acceptedClients_.erase( iter );

        }
    }

    // remove from file transfer client map
    {
        ClientMap::iterator iter;
        while( ( iter = std::find_if( fileTransferClients_.begin(), fileTransferClients_.end(), SameStateFTor( QAbstractSocket::UnconnectedState ) ) ) != fileTransferClients_.end() )
        { fileTransferClients_.erase( iter ); }
    }

    // remove from list of connected clients
    {
        ClientList::iterator iter;
        while( ( iter = std::find_if( connectedClients_.begin(), connectedClients_.end(), SameStateFTor( QAbstractSocket::UnconnectedState ) ) ) != connectedClients_.end() )
        {
            (*iter)->deleteLater();
            connectedClients_.erase( iter );
        }
    }

}

//_____________________________________________________
void ServerInterface::_processCommand( ChatCommand command )
{

    Debug::Throw( debugLevel_ )
        << "ServerInterface::_processCommand -"
        << " time: " << command.creationTime().toString()
        << " user: " << command.connectionId().user()
        << " command: " << command.commandName()
        << endl;

    // find matching client
    auto iter( std::find_if(  connectedClients_.begin(), connectedClients_.end(), ChatClient::SameIdFTor( command.clientId() ) ) );
    Q_ASSERT( iter != connectedClients_.end() );
    _processCommand( command, *iter );

    return;
}

//_____________________________________________________
void ServerInterface::_processCommand( ChatCommand command, ChatClient* sender )
{

    switch( command.command() )
    {

        // connection requested by client
        case ChatCommand::ConnectionRequest:
        {

            // copy connection id and assign connection time
            ConnectionId connectionId( command.connectionId() );
            connectionId.setConnectionTime( TimeStamp::now() );

            // assign connection id to sender
            sender->setConnectionId( connectionId );

            // see if there is already a client with this connectionId
            auto iter = std::find_if( acceptedClients_.begin(), acceptedClients_.end(), SameConnectionIdFTor( connectionId ) );
            if( iter != acceptedClients_.end() )
            {
                if( (*iter) != sender )
                {
                    // send disconnection signal
                    (*iter)->sendCommand( ChatCommand( connectionId, ChatCommand::ConnectionDenied ) );
                    acceptedClients_.erase( iter );

                    // store new cliend in list of accepted clients
                    acceptedClients_.append( sender );
                }

            } else acceptedClients_.append( sender );

            // return accept to calling client
            sender->sendCommand( ChatCommand( connectionId, ChatCommand::ConnectionAccepted ) );

            // broadcast client connected, UserStateChanged and avatar
            _broadcast( ChatCommand( connectionId, ChatCommand::ContactOnline ), sender );
            _broadcast( ChatCommand( connectionId, ChatCommand::UserStateChanged ), sender );
            _broadcast( Avatar( connectionId, AvatarImage( connectionId.user() ) ), sender );

            break;
        }

        // ping request from client
        case ChatCommand::PingServer:
        {
            sender->sendCommand( ChatCommand( command.connectionId(), ChatCommand::PongServer ) );
            break;
        }

        // message sent
        case ChatCommand::Message:
        {

            // append to local message list
            ChatMessage message( command );

            // add to local
            messages_.append( message );

            // trigger autosave
            if( autoSaveTimeout_ > 0 )
            {

                if( !autoSaveTimer_.isActive() ) autoSaveTimer_.start( autoSaveTimeout_, this );

            } else writeMessages();

            // broadcast to all connected clients but sender
            _broadcast( command, sender );
            break;

        }

        // message list request
        case ChatCommand::FetchMessages:
        {

            // find start time
            TimeStamp timeStamp;
            if( chatCommandLineParser_.parse( command.arguments() ).hasOption( "-t" ) )
            {
                timeStamp.setTime( chatCommandLineParser_.option( "-t" ).toInt() );
                Debug::Throw( debugLevel_ ) << "ServerInterface::_processCommand - timeStamp: " << timeStamp.toString() << endl;
            }

            // read
            messages_.read( timeStamp );

            // find all matching messages
            QStringList sourceRecipients( command.recipients() );
            sourceRecipients.append( command.connectionId().user() );

            const ChatMessage::List messages( messages_.fetch( timeStamp, sourceRecipients ) );
            if( messages.empty() ) break;

            // notify start message log
            {
                ChatCommand startLogCommand( command.connectionId(), ChatCommand::StartMessageLog );
                startLogCommand.setArguments( QStringList() << startLogCommand.commandName() << "-s" << QString::number( messages.size() ) );
                startLogCommand.setRecipients( command.recipients() );
                sender->sendCommand( startLogCommand );
            }

            // clear message list
            ChatCommand clearMessageListcommand( command.connectionId(), ChatCommand::ClearMessageList );
            clearMessageListcommand.setRecipients( command.recipients() );
            sender->sendCommand( clearMessageListcommand );
            Debug::Throw( debugLevel_ ) << "ServerInterface::_processCommand - found " << messages.size() << " messages" << endl;

            // send messages
            for( const auto& message:messages )
            { sender->sendCommand( message.toChatCommand() ); }

            // notify end message log
            {
                ChatCommand stopLogCommand( command.connectionId(), ChatCommand::StopMessageLog );
                stopLogCommand.setRecipients( command.recipients() );
                sender->sendCommand( stopLogCommand );
            }


            // finally also send pending commands (e.g. chat transfer requests)
            _sendPendingCommands( sender );

            break;
        }

        // connected client list request
        case ChatCommand::FetchContacts:
        {

            // clear list of connected clients
            sender->sendCommand( ChatCommand( command.connectionId(), ChatCommand::ClearContactList ) );

            // loop over connected clients
            for( const auto& client:acceptedClients_ )
            {

                if( client->connectionId() != command.connectionId() )
                {
                    // send online, state, and avatar
                    sender->sendCommand( ChatCommand( client->connectionId(), ChatCommand::ContactOnline ) );
                    sender->sendCommand( ChatCommand( client->connectionId(), ChatCommand::UserStateChanged ) );
                }

                // send avatars
                sender->sendAvatar( Avatar( client->connectionId(), AvatarImage( client->connectionId().user() ) ) );

            }

            break;

        }

        // user state changed
        case ChatCommand::UserStateChanged:
        {

            // change status stored in accepted clients
            auto iter = std::find_if( acceptedClients_.begin(), acceptedClients_.end(), SameConnectionIdFTor( command.connectionId() ) );
            if( iter != acceptedClients_.end() ) (*iter)->setConnectionId( command.connectionId() );

            // broadcast
            _broadcast( command, sender );
            break;

        }

        // user state changed
        case ChatCommand::UserNameChanged:
        {

            // find matching iterator in accepted clients list
            auto iter = std::find_if( acceptedClients_.begin(), acceptedClients_.end(), SameClientFTor(sender) );
            if( iter == acceptedClients_.end() ) break;

            // find new user
            QString newUserName;
            if( chatCommandLineParser_.parse( command.arguments() ).hasOption( "-u" ) )
            {

                newUserName = chatCommandLineParser_.option( "-u" );

            } else newUserName = command.connectionId().user();

            // check if user name indeed changed
            ConnectionId connectionId( (*iter)->connectionId() );
            if( connectionId.user() == newUserName ) break;

            // update user
            connectionId.setUser( newUserName );
            sender->setConnectionId( connectionId );

            // broadcast
            _broadcast( command, sender );
            break;
        }

        case ChatCommand::FileTransferAccepted:
        {
            if( chatCommandLineParser_.parse( command.arguments() ).hasOption( "-f" ) )
            {
                const File filename = chatCommandLineParser_.option( "-f" );

                // loop over fileTransfer clients
                bool found = false;
                for( auto&& iter = fileTransferClients_.begin(); iter != fileTransferClients_.end(); ++iter )
                {
                    if( iter.key() == filename && iter.value() == sender )
                    { found = true; break; }
                }

                if( !found ) fileTransferClients_.insertMulti( filename, sender );

            }

            // broadcast
            _broadcast( command, sender );

            break;
        }

        case ChatCommand::PingClient:
        {
            _broadcast( command, sender );
            break;
        }

        case ChatCommand::PongClient:
        {
            // update connection time from 'this' client
            if( !command.connectionId().connectionTime().isValid() )
            {
                auto connectionId( command.connectionId() );
                connectionId.setConnectionTime( sender->connectionId().connectionTime() );
                command.setConnectionId( connectionId );
            }

            _broadcast( command, sender );
            break;
        }

        case ChatCommand::FileTransferRequest:
        {

            const auto& recipients( command.recipients() );
            if( recipients.empty() )
            {
                _broadcast( command, sender );
                break;
            }

            // loop over recipients
            for( const auto& recipient:recipients )
            {
                    // try find a matching accepted clients
                bool found = false;
                for( const auto& client:acceptedClients_ )
                {
                    if( client->connectionId().user() == recipient )
                    {
                        found = true;
                        client->sendCommand( command );
                    }
                }

                // if not found, add to list of pending commands
                if( !found ) pendingCommands_[recipient].append( command );

            }

            break;
        }

        default:
        {
            _broadcast( command, sender );
            break;
        }

    }

    return;

}

//_____________________________________________________
void ServerInterface::_processAvatar( Avatar avatar )
{

    Debug::Throw( debugLevel_ )
        << "ServerInterface::_processAvatar -"
        << " user: " << avatar.connectionId().user()
        << endl;

    // find matching client
    ClientList::iterator iter( std::find_if(  connectedClients_.begin(), connectedClients_.end(), ChatClient::SameIdFTor( avatar.clientId() ) ) );
    Q_ASSERT( iter != connectedClients_.end() );

    auto sender = *iter;

    // save avatar locally
    AvatarImage avatarImage( avatar.connectionId().user(), avatar.image() );
    if( avatarImage.isNull() ) avatarImage.remove();
    else avatarImage.save();

    _broadcast( avatar, sender );

}

//_____________________________________________________
void ServerInterface::_broadcast(FileChunk fileChunk)
{

    Debug::Throw( debugLevel_ )
        << "ServerInterface::_broadcast -"
        << " filename: " << fileChunk.filename()
        << " flags: " << fileChunk.flags()
        << " size: " << File::sizeString( fileChunk.data().size() )
        << endl;

    // find all fileTransferClients that match filename
    const File filename( fileChunk.filename() );
    ClientList clients = fileTransferClients_.values( filename );

    // get recipients
    const QStringList& recipients( fileChunk.recipients() );
    for( const auto& client:clients )
    {
        if( recipients.empty() || recipients.contains( client->connectionId().user() ) )
        { client->sendFileChunk( fileChunk ); }
    }
}

//_____________________________________________________
void ServerInterface::_broadcast( Avatar avatar, ChatClient* sender ) const
{

    Debug::Throw() << "ServerInterface::_Broadcast - user: " << avatar.connectionId().user() << " (avatar)" << endl;

    // broadcast
    const auto& recipients( avatar.recipients() );
    if( !recipients.empty() )
    {

        // when recipients are specified, only loop through accepted clients
        for( const auto& client:acceptedClients_ )
        {

            if( client == sender ) continue;
            if( !recipients.contains( client->connectionId().user() ) ) continue;
            client->sendAvatar( avatar );

        }

    } else if( avatar.connectionId().isValid() ) {

        // when recipients are specified, only loop through accepted clients
        for( const auto& client:acceptedClients_ )
        {
            if( client == sender ) continue;
            if( client->connectionId() == avatar.connectionId() ) continue;
            client->sendAvatar( avatar );
        }

    } else {

        // loop over all registered clients
        for( const auto& client:acceptedClients_ )
        { if( client != sender ) client->sendAvatar( avatar ); }

    }

}
//_____________________________________________________
void ServerInterface::_broadcast( ChatCommand command, ChatClient* sender ) const
{

    Debug::Throw() << "ServerInterface::_Broadcast - user: " << command.connectionId().user() << " command: " << command.commandName() << endl;
    const auto& recipients( command.recipients() );

    if( !recipients.empty() )
    {

        // when recipients are specified, only loop through accepted clients
        for( const auto& client:acceptedClients_ )
        {
            if( client == sender ) continue;
            if( !recipients.contains( client->connectionId().user() ) ) continue;
            client->sendCommand( command );
        }

    } else if( command.connectionId().isValid() ) {

        // when recipients are specified, only loop through accepted clients
        for( const auto& client:acceptedClients_ )
        {
            if( client == sender ) continue;
            if( client->connectionId() == command.connectionId() ) continue;
            client->sendCommand( command );
        }

    } else {

        // loop over all registered clients
        for( const auto& client:acceptedClients_ )
        { if( client != sender ) client->sendCommand( command ); }

    }

}

//______________________________________
void ServerInterface::_updateConfiguration( void )
{
    Debug::Throw( "ServerInterface::_updateConfiguration.\n" );

    // hostName
    if( !parser_.hasOption( "--chat-host" ) )
    {
        QHostInfo hostInfo( QHostInfo::fromName( XmlOptions::get().raw( "CHAT_HOST" ) ) );
        if( !hostInfo.addresses().isEmpty() ) host_ = hostInfo.addresses().front();
    }

    // port
    if( !parser_.hasOption( "--chat-port" ) )
    { port_ = XmlOptions::get().get<unsigned int>( "CHAT_PORT" ); }

    // compression
    messages_.setUseCompression( XmlOptions::get().get<bool>( "COMPRESS_HISTORY" ) );

    // history
    const File oldHistoryFile( messages_.filename() );
    const File historyFile( XmlOptions::get().raw("HISTORY_FILE") );
    if( !oldHistoryFile.isEmpty() && oldHistoryFile != historyFile )
    { ChatMessageMigration( oldHistoryFile ).migrate( historyFile ); }

    messages_.setFilename( File( XmlOptions::get().raw("HISTORY_FILE") ) );

    // auto save timeout
    autoSaveTimeout_ = 1000*XmlOptions::get().get<int>( "AUTOSAVE_TIMEOUT" );

    // full save timeout
    fullSaveTimeout_ = 1000*3600*XmlOptions::get().get<int>( "FULLSAVE_TIMEOUT" );

    if( fullSaveTimer_.isActive() ) fullSaveTimer_.stop();
    if( fullSaveTimeout_ > 0 ) fullSaveTimer_.start( fullSaveTimeout_, this );

    return;

}

//______________________________________
void ServerInterface::_saveConfiguration( void )
{
    Debug::Throw( "ServerInterface::_saveConfiguration.\n" );
    writeMessages();
}

//_____________________________________________________
void ServerInterface::_sendPendingCommands( ChatClient* client )
{
    auto iter = pendingCommands_.find( client->connectionId().user() );
    if( iter == pendingCommands_.end() ) return;

    for( const auto& command:iter.value() )
    { client->sendCommand( command ); }

    pendingCommands_.erase( iter );
}
