#ifndef ServerInterface_h
#define ServerInterface_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ChatClient.h"
#include "ChatCommandLineParser.h"
#include "ChatMessageSet.h"
#include "FileChunk.h"
#include "XmlDocument.h"

#include <QBasicTimer>
#include <QHostAddress>
#include <QTcpServer>
#include <QTimerEvent>

class ServerInterface: public QObject, public Counter
{

    //* Qt meta object declaration
    Q_OBJECT

    public:

    //* constructor
    ServerInterface( QObject*, const CommandLineParser& = CommandLineParser() );

    //*@name accessors
    //@{

    const ChatMessage::Set& messages( void ) const
    { return messages_.messages(); }

    //@}

    //*@name modifiers
    //@{

    //* initialize
    void initialize( void );

    //* host
    void setHost( QHostAddress value )
    { host_ = value; }

    //* port
    void setPort( int value )
    { port_ = value; }

    //* write messages
    bool writeMessages( void );

    //@}

    protected Q_SLOTS:

    //* a new connection is granted
    virtual void _newConnection( void );

    //* client disconnected
    virtual void _clientDisconnected( void );

    //* command processing
    virtual void _processCommand( ChatCommand );

    //* avatar processing
    virtual void _processAvatar( Avatar );

    //* file chunk processing
    virtual void _broadcast( FileChunk );

    protected:

    //* timer event
    void timerEvent( QTimerEvent* );

    //* used to retrieve pair with matching client
    using ClientPair = QPair<File,ChatClient*>;
    class SameStateFTor: public ChatClient::SameStateFTor
    {
        public:

        //* constructor
        SameStateFTor( QAbstractSocket::SocketState state ):
        ChatClient::SameStateFTor( state )
        {}

        //* predicate
        using ChatClient::SameStateFTor::operator ();

        //* predicate
        bool operator() ( ClientPair pair ) const
        { return ChatClient::SameStateFTor::operator () (pair.second); }

    };

    //* used to retrieve pair with matching client
    class SameClientFTor
    {
        public:

        //* constructor
        SameClientFTor( ChatClient* client ):
            client_( client )
        {}

        //* predicate
        bool operator() ( ChatClient* client ) const
        { return client == client_; }

        private:

        //* prediction
        ChatClient* client_;
    };

    //* used to retrieve pair with matching client
    class SameConnectionIdFTor
    {
        public:

        //* constructor
        SameConnectionIdFTor( ConnectionId connectionId ):
            connectionId_( connectionId )
        {}

        //* predicate
        bool operator() ( ChatClient* client ) const
        { return client->connectionId() == connectionId_; }

        private:

        //* prediction
        ConnectionId connectionId_;
    };

    //* process command
    void _processCommand( ChatCommand, ChatClient* );

    //* broadcast a message to all registered clients but the sender (if valid)
    virtual void _broadcast( ChatCommand, ChatClient* = nullptr ) const;

    //* broadcast an avatar to all registered clients but the sender (if valid)
    virtual void _broadcast( Avatar, ChatClient* = nullptr ) const;

    private Q_SLOTS:

    //* update configuration
    void _updateConfiguration( void );

    //* save configuration
    void _saveConfiguration( void );

    private:

    //* send pending commands to a given contact
    void _sendPendingCommands( ChatClient* );

    //* command line parser (passed at initialization)
    CommandLineParser parser_;

    //* debug level
    int debugLevel_ = 1;

    //* host
    QHostAddress host_ = QHostAddress::Null;

    //* port
    int port_ = 0;

    //* Server
    QTcpServer* server_ = nullptr;

    //* chat command line parser
    ChatCommandLineParser chatCommandLineParser_;

    //* chat message list
    ChatMessageSet messages_;

    //* connected clients
    using ClientList = QList< ChatClient* >;
    ClientList connectedClients_;

    //* accepted clients
    ClientList acceptedClients_;

    //* map file transfer clients and filename
    using ClientMap = QHash<File,ChatClient*>;
    ClientMap fileTransferClients_;

    //* pending commands
    /** this stores commands that must be sent to a given contact,
    once it becomes online */
    using CommandHash = QHash<QString, ChatCommand::List>;
    CommandHash pendingCommands_;

    //*@name autosave
    //@{

    QBasicTimer autoSaveTimer_;
    int autoSaveTimeout_ = 0;

    QBasicTimer fullSaveTimer_;
    int fullSaveTimeout_ = 0;

    //@}

};

#endif
