/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "SoundNotifications.h"

#include "Debug.h"
#include "SoundNames.h"

#include <QFile>
#include <QThread>
#include <QMutex>

#if WITH_PULSEAUDIO
#include <pulse/simple.h>
#include <pulse/error.h>
#endif

//____________________________________________________________
class SoundNotificationsP:public QThread, public Counter
{

    public:

    //* constructor
    SoundNotificationsP( QObject* = nullptr );

    //* destructor
    ~SoundNotificationsP( void );

    //* is supported
    bool isSupported( void ) const;

    //* play a file
    void setFilename( QString );

    protected:

    //* Check files validity. Post a ValidFileEvent when finished
    virtual void run( void );

    private:

    //* filename to play
    QString filename_;

    #if WITH_PULSEAUDIO
    pa_simple* pulseAudioHandle_ = nullptr;
    #endif

};

//____________________________________________________________
SoundNotificationsP::SoundNotificationsP( QObject* parent ):
    QThread( parent ),
    Counter( "SoundNotificationsP" )
{
    #if WITH_PULSEAUDIO

    /* The Sample format to use */
    pa_sample_spec sampleSpec;
    sampleSpec.format = PA_SAMPLE_S16LE;
    sampleSpec.rate = 16000;
    sampleSpec.channels = 1;

    int error;
    pulseAudioHandle_ = pa_simple_new( nullptr, "Homechat", PA_STREAM_PLAYBACK, nullptr, "SoundNotifications", &sampleSpec, nullptr, nullptr, &error);
    if( !pulseAudioHandle_ )
    {
        Debug::Throw(0)
            << "SoundNotificationsP::SoundNotificationsP -"
            << " unable to initialize pulseAudio: "
            << pa_strerror(error)
            << endl;
    }
    #endif
}

//____________________________________________________________
SoundNotificationsP::~SoundNotificationsP( void )
{
    #if WITH_PULSEAUDIO
    if( pulseAudioHandle_ ) pa_simple_free( pulseAudioHandle_ );
    #endif
}

//____________________________________________________________
bool SoundNotificationsP::isSupported( void ) const
{
    #if WITH_PULSEAUDIO
    return pulseAudioHandle_;
    #else
    return false;
    #endif
}

//____________________________________________________________
void SoundNotificationsP::setFilename( QString filename )
{ filename_ = filename; }

//____________________________________________________________
void SoundNotificationsP::run( void )
{

    if( filename_.isEmpty() ) return;

    #if WITH_PULSEAUDIO
    QFile file( filename_ );
    if( !file.open( QIODevice::ReadOnly ) )
    {
        Debug::Throw(0) << "SoundNotificationsP::play - error opening " << filename_ << endl;
        return;
    }

    // read
    char buffer[1024];
    forever
    {

        // read some data
        const qint64 read = file.read( buffer, sizeof( buffer ) );
        if( read < 0 )
        {
            Debug::Throw(0) << "SoundNotifications::play - error reading " << filename_ << endl;
            return;
        } else if( read == 0 ) break;

        // play
        int error;
        if( pa_simple_write( pulseAudioHandle_, buffer, read, &error) < 0 )
        {
            Debug::Throw(0)
                << "SoundNotifications::play - error playing " << filename_ << ": "
                << pa_strerror(error)
                << endl;
            return;
        }

    }

    // Make sure that every single sample was played
    int error;
    if( pa_simple_drain( pulseAudioHandle_, &error) < 0)
    { Debug::Throw(0) << "SoundNotifications::play - draining failed: " << pa_strerror(error) << endl; }

    #endif

    return;

}

//____________________________________________________________
SoundNotifications::SoundNotifications( QObject* parent ):
    QObject( parent ),
    Counter( "SoundNotifications" ),
    d( new SoundNotificationsP(this) )
{ connect( d, SIGNAL(finished()), this, SLOT(_threadFinished()) ); }

//____________________________________________________________
SoundNotifications::~SoundNotifications( void )
{ delete d; }

//____________________________________________________________
bool SoundNotifications::isSupported( void ) const
{ return d->isSupported(); }

//____________________________________________________________
void SoundNotifications::play( SoundNotifications::Type type )
{

    // check if supported
    if( !isSupported() ) return;

    // check against mask
    if( !(mask_ & (1<<type) ) ) return;

    // get input file from type
    QString filename;
    switch( type )
    {
        default:
        case None: break;
        case ContactOnline: filename = SoundNames::ContactOnline; break;
        case ContactOffline: filename = SoundNames::ContactOffline; break;
        case NewMessage: filename = SoundNames::NewMessage; break;
        case FileTransferRequest: filename = SoundNames::FileTransferRequest; break;
    }

    if( filename.isEmpty() ) return;
    if( d->isRunning() ) filenames_.append( filename );
    else {
        d->setFilename( filename );
        d->start();
    }

}

//____________________________________________________________
void SoundNotifications::_threadFinished( void )
{
    if( !filenames_.empty() )
    {
        d->setFilename( filenames_.front() );
        filenames_.removeFirst();
        d->start();
    }
}
