/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "SshInterface.h"

#include "InformationDialog.h"
#include "Singleton.h"
#include "SshLoginDialog.h"
#include "XmlOptions.h"

#include <QHostAddress>

namespace Ssh
{

    //____________________________________________
    Interface::Interface( QObject* parent, const CommandLineParser& parser ):
        QObject( parent ),
        Counter( "Ssh::Interface" ),
        parser_( parser ),
        connection_( new Connection( this ) )
    {

        // connections
        connect( connection_, SIGNAL(attributesChanged()), this, SLOT(_updateAttributes()) );
        connect( connection_, SIGNAL(connected()), this, SIGNAL(connected()) );
        connect( connection_, SIGNAL(loginFailed()), this, SLOT(_notifyLoginFailed()));
        connect( connection_, SIGNAL(error(QString)), this, SIGNAL(connected()));
        connect( connection_, SIGNAL(error(QString)), this, SIGNAL(error(QString)));

        connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
        _updateConfiguration();

    }

    //_____________________________________________________
    void Interface::connect( void )
    {
        Debug::Throw( "Ssh::Interface::connect.\n" );

        // assign connection attributes
        connection_->setAttributes( attributes_ );

        // first create tunnels
        /* if fails, send connected signal nonetheless, to give a chance
        to direct TCP connection */
        if( !connection_->createTunnels() )
        {
            emit connected();
            return;
        }

        // connection and handshake
        if( !connection_->connect() )
        {
            Debug::Throw(0) << "Ssh::Interface::connect - connection failed" << endl;
            return;
        }

        // authentication
        if( !connection_->authenticate() )
        {
            Debug::Throw(0) << "Ssh::Interface::connect - authentication failed" << endl;
            return;
        }

    }

    //_____________________________________________________
    void Interface::disconnect( void )
    {
        Debug::Throw( "Ssh::Interface::disconnect.\n" );
        connection_->disconnect();
    }

    //______________________________________
    void Interface::_updateAttributes( void )
    {
        Debug::Throw( "Ssh::Interface::_updateAttributes.\n" );
        attributes_ = connection_->attributes();

        // retrieve connection attributes
        XmlOptions::get().setRaw( "SSH_HOST", attributes_.host() );
        XmlOptions::get().setRaw( "SSH_USERNAME", attributes_.userName() );
        XmlOptions::get().set<bool>( "SSH_REMEMBER_PASSWORD", attributes_.rememberPassword() );

        if( attributes_.rememberPassword() )
        {

            XmlOptions::get().setRaw( "SSH_PASSWORD", attributes_.password() );

        } else {

            XmlOptions::get().setRaw( "SSH_PASSWORD", QString() );

        }

    }

    //______________________________________
    void Interface::_notifyLoginFailed( void )
    {
        Debug::Throw( "Ssh::Interface::_notifyLoginFailed.\n" );

        // notify login failed
        InformationDialog dialog( nullptr, tr( "Login failed" ) );
        dialog.setWindowTitle( "SSH login - Homechat" );
        dialog.exec();

    }

    //______________________________________
    void Interface::_updateConfiguration( void )
    {
        Debug::Throw( "Interface::_updateConfiguration.\n" );

        // connection attributes
        attributes_.setHost( XmlOptions::get().raw( "SSH_HOST" ) );
        attributes_.setPort( XmlOptions::get().get<int>( "SSH_PORT" ) );
        attributes_.setUserName( XmlOptions::get().raw( "SSH_USERNAME" ) );

        attributes_.setRememberPassword( XmlOptions::get().get<bool>( "SSH_REMEMBER_PASSWORD" ) );
        if( attributes_.rememberPassword() )
        {

            if( XmlOptions::get().contains( "SSH_PASSWORD" ) )
            { attributes_.setPassword( XmlOptions::get().raw( "SSH_PASSWORD" ) ); }

        } else {

            attributes_.setPassword( QString() );
        }

        // tunnel attributes
        // update host and port from options
        TunnelAttributes tunnelAttributes;
        if( parser_.hasOption( "--chat-host" ) ) tunnelAttributes.setHost( parser_.option( "--chat-host" ) );
        else tunnelAttributes.setHost( XmlOptions::get().raw( "CHAT_HOST" ) );

        // tunnel port
        if( parser_.hasOption( "--chat-port" ) )
        {

            const int port( parser_.option( "--chat-port" ).toInt() );
            tunnelAttributes.setLocalPort( port );
            tunnelAttributes.setRemotePort( port );

        } else {

            const int port( XmlOptions::get().get<unsigned int>( "CHAT_PORT" ) );
            tunnelAttributes.setLocalPort( port );
            tunnelAttributes.setRemotePort( port );

        }

        // add to connection attributes
        attributes_.addTunnel( tunnelAttributes );

        return;
    }

}
