/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "SshSynchronizationDialog.h"

#include "SshConnectionWidget.h"
#include "SshFileTransferObject.h"

#include "IconNames.h"
#include "CustomDialog.h"
#include "GridLayout.h"
#include "IconEngine.h"
#include "LineEditor.h"
#include "LogWidget.h"
#include "QuestionDialog.h"

#include "Singleton.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QDialogButtonBox>
#include <QDomDocument>
#include <QDomElement>
#include <QGroupBox>
#include <QLabel>
#include <QToolButton>

namespace Ssh
{

    namespace Private
    {

        namespace Xml
        {
            const QString ConnectionAttributesList( "cal" );
            const QString ConnectionAttributes( "ca" );
            const QString ConnectionName( "cn" );
            const QString Host( "h" );
            const QString Port( "p" );
            const QString UserName( "u" );
            const QString Password( "pwd" );
            const QString RemotePath( "rpth" );
        };

        //________________________________________________________
        class XmlConnectionAttributes: public ConnectionAttributes
        {

            public:

            //* default constructor
            XmlConnectionAttributes( void )
            {}

            //* constructor from connection attributes
            XmlConnectionAttributes( ConnectionAttributes& other, File path ):
                ConnectionAttributes( other ),
                remotePath_( path )
            {}

            //* constructor from dom
            XmlConnectionAttributes( const QDomElement& element )
            {

                setRememberPassword( true );
                QDomNamedNodeMap attributes( element.attributes() );
                for( int i=0; i<attributes.count(); i++ )
                {
                    QDomAttr attribute( attributes.item( i ).toAttr() );
                    if( attribute.isNull() ) continue;
                    if( attribute.name() == Xml::ConnectionName ) setName( attribute.value() );
                    else if( attribute.name() == Xml::Host ) setHost( attribute.value() );
                    else if( attribute.name() == Xml::Port ) setPort( attribute.value().toInt() );
                    else if( attribute.name() == Xml::UserName ) setUserName( attribute.value() );
                    else if( attribute.name() == Xml::Password ) setPassword( attribute.value() );
                    else if( attribute.name() == Xml::RemotePath ) remotePath_ = File( attribute.value() );
                }
            }

            //*@name accessors
            //@{

            //* create dom element
            QDomElement domElement( QDomDocument& document ) const
            {

                QDomElement out( document.createElement( Xml::ConnectionAttributes ) );
                out.setAttribute( Xml::ConnectionName, name() );
                out.setAttribute( Xml::Host, host() );
                out.setAttribute( Xml::Port, QString::number( port() ) );
                out.setAttribute( Xml::UserName, userName() );
                out.setAttribute( Xml::Password, password() );
                out.setAttribute( Xml::RemotePath, remotePath_ );
                return out;
            }

            //* remote path
            File remotePath( void ) const
            { return remotePath_; }

            //*@}

            private:

            //* remote path
            File remotePath_;

        };

        //________________________________________________________
        class EditConnectionNameDialog: public CustomDialog
        {
            Q_OBJECT

            public:

            //* constructor
            EditConnectionNameDialog( QWidget* parent ):
                CustomDialog( parent, OkButton|CancelButton|Separator )
            {
                setWindowTitle( tr( "Connection name - Homechat synchronize" ) );
                mainLayout().addWidget( editor_ = new LineEditor( this ) );
            }

            //*@name accessors
            //@{

            //* name
            QString connectionName( void ) const { return editor_->text(); }

            //@}

            //*@name modifier
            //@{

            //* set connection name
            void setConnectionName( QString value )
            { editor_->setText( value ); }

            private:

            //* editor
            LineEditor* editor_;

        };

        //________________________________________________________
        class Container: public QGroupBox, public Counter
        {

            Q_OBJECT

            public:

            //* contructor
            Container( QString name, QWidget* parent ):
                QGroupBox( parent ),
                Counter( "Ssh::Private::Container" )
            {
                setLayout( new QVBoxLayout() );
                layout()->setMargin(5);
                layout()->setSpacing(5);

                auto gridLayout = new QGridLayout();
                gridLayout->setMargin(0);
                gridLayout->setSpacing(0);
                layout()->addItem( gridLayout );

                titleLabel_ = new QLabel( name, this );
                titleLabel_->setAlignment( Qt::AlignCenter );
                gridLayout->addWidget( titleLabel_, 0, 0, 1, 3 );

                {
                    auto button = new QToolButton( this );
                    button->setAutoRaise( true );
                    button->setIcon( IconEngine::get( IconNames::Edit ) );
                    button->setText( tr( "Edit" ) );
                    button->setToolTip( tr( "Change connection name" ) );
                    connect( button, SIGNAL(clicked()), SLOT(_rename()) );
                    gridLayout->addWidget( button, 0, 1, 1, 1 );
                }

                {
                    auto button = new QToolButton( this );
                    button->setAutoRaise( true );
                    button->setIcon( IconEngine::get( IconNames::DialogClose ) );
                    button->setText( tr( "Close" ) );
                    button->setToolTip( tr( "Delete this remote connection" ) );
                    connect( button, SIGNAL(clicked()), SLOT(_close()) );
                    gridLayout->addWidget( button, 0, 2, 1, 1 );
                }
            }

            //*@name accessors
            //@{

            QString connectionName( void ) const { return titleLabel_->text(); }

            //@}

            protected Q_SLOTS:

            //* rename
            void _rename( void )
            {

                EditConnectionNameDialog dialog( this );
                dialog.setConnectionName( connectionName() );
                if( !dialog.exec() ) return;

                // update title label and connection widget
                auto newName( dialog.connectionName() );
                if( newName != connectionName() )
                {
                    titleLabel_->setText( newName );
                    auto widget = findChild<ConnectionWidget*>();
                    widget->setName( newName );
                }

            }

            //* close
            void _close( void )
            {

                // ask for confirmation
                QuestionDialog dialog( this, tr( "Delete the remote connection named \"%1\" ?" ).arg( connectionName() ) );
                dialog.okButton().setText( tr( "Delete" ) );
                dialog.okButton().setIcon( IconEngine::get( IconNames::Delete ) );
                if( !dialog.exec() ) return;

                QWidget* window( this->window() );
                hide();
                deleteLater();
                window->adjustSize();

            }

            private:

            QLabel* titleLabel_ = nullptr;

        };

        //________________________________________________________
        class LocalConnectionWidget: public ConnectionWidget
        {

            Q_OBJECT

            public:

            //* constructor
            LocalConnectionWidget( QWidget* parent ):
                ConnectionWidget( parent )
            {
                hideSavePassword();
                GridLayout* layout( qobject_cast<GridLayout*>( this->layout() ) );
                layout->addWidget( new QLabel( tr( "Remote path:" ), this ) );
                layout->addWidget( remotePathEditor_ = new LineEditor( this ), 4, 1, 1, 3 );
            }

            //*@name accessors
            //@{

            //* remote path
            File remotePath( void ) const
            { return File( remotePathEditor_->text() ); }

            //@}

            //*@name modifiers
            //@{

            //* remote path
            void setRemotePath( File remotePath )
            { remotePathEditor_->setText( remotePath ); }

            //@}

            private:

            //* remote path
            LineEditor* remotePathEditor_;

        };

    }

    //________________________________________________________
    SynchronizationDialog::SynchronizationDialog( QWidget* parent ):
        BaseDialog( parent ),
        Counter( "Ssh::SynchronizationDialog" )
    {

        setOptionName( "SYNCHRONIZATION_DIALOG" );

        setLayout( new QVBoxLayout() );
        layout()->setMargin(10);
        layout()->setSpacing(5);

        layout()->addItem( layout_ = new QVBoxLayout() );
        layout_->setMargin(0);
        layout_->setSpacing(5);

        // add connection button
        auto hLayout = new QHBoxLayout();
        layout()->addItem( hLayout );
        hLayout->setMargin(0);
        hLayout->setSpacing(5);
        auto button = new QPushButton( IconEngine::get( IconNames::Add ), tr( "Add remote connection" ), this );
        hLayout->addStretch(1);
        hLayout->addWidget( button );

        connect( button, SIGNAL(clicked()), SLOT(_addConnection()) );

        // log widget
        layout()->addWidget( logWidget_ = new LogWidget( this ) );
        logWidget_->setOptionName( "SYNCHRONIZATION_LOG" );

        // progress bar
        layout()->addWidget( progressBar_ = new QProgressBar( this ) );

        progressBar_->setTextVisible( false );
        progressBar_->setMinimum(0);
        progressBar_->setMaximum(100);

        // buttons
        hLayout = new QHBoxLayout();
        hLayout->setMargin(0);
        hLayout->setSpacing(5);
        layout()->addItem( hLayout );

        synchronizeButton_ = new QPushButton( IconEngine::get( IconNames::Synchronize ), tr( "Synchronize" ), this );
        synchronizeButton_->setEnabled( false );
        connect( synchronizeButton_, SIGNAL(clicked()), SLOT(_synchronize()) );

        abortButton_ = new QPushButton( IconEngine::get( IconNames::DialogCancel ), tr( "Abort" ), this );
        abortButton_->setEnabled( false );
        connect( abortButton_, SIGNAL(clicked()), SLOT(_abort()) );

        closeButton_ = new QPushButton( IconEngine::get( IconNames::DialogClose ), tr( "Close" ), this );
        closeButton_->setEnabled( true );
        connect( closeButton_, SIGNAL(clicked()), SLOT(close()) );

        hLayout->addStretch(1);
        hLayout->addWidget(synchronizeButton_);
        hLayout->addWidget(abortButton_);
        hLayout->addWidget(closeButton_);

    }

    //________________________________________________________
    void SynchronizationDialog::addConnection( ConnectionAttributes attributes, File remotePath )
    {

        auto container = new Private::Container( attributes.name(), this );
        layout_->addWidget( container );

        auto connectionWidget = new Private::LocalConnectionWidget( container );
        container->layout()->addWidget( connectionWidget );

        connectionWidget->setAttributes( attributes );
        connectionWidget->setRemotePath( remotePath );

        connect( connectionWidget, SIGNAL(destroyed()), SLOT(_connectionWidgetDestroyed()) );
        connectionWidget->show();
        container->show();
        adjustSize();

        _updateSynchronizationButton();

        return;

    }

    //________________________________________________________
    void SynchronizationDialog::readConnectionAttributes( void )
    {

        // input file
        const File file( XmlOptions::get().raw( "RC_FILE" ) );

        // parse the file
        XmlDocument document;
        QFile qfile( file );
        if( !document.setContent( &qfile ) )
        {
            Debug::Throw() << document.error() << endl;
            return;
        }

        // look for relevant element
        auto topNodes = document.elementsByTagName( Private::Xml::ConnectionAttributesList );
        if( topNodes.isEmpty() )
        {
            Debug::Throw() <<  "SynchronizationDialog::readConnectionAttributes - top node not found" << endl;
            return;
        }

        for(QDomNode node = topNodes.at(0).firstChild(); !node.isNull(); node = node.nextSibling() )
        {
            auto element = node.toElement();
            if( element.isNull() ) continue;

            // special options
            if( element.tagName() == Private::Xml::ConnectionAttributes )
            {
                Private::XmlConnectionAttributes attributes( element );
                addConnection( attributes, attributes.remotePath() );
            }

        }

    }

    //________________________________________________________
    void SynchronizationDialog::writeConnectionAttributes( void )
    {
        // input file
        const File file( XmlOptions::get().raw( "RC_FILE" ) );

        // create document and read
        XmlDocument document;
        {
            QFile qtfile( file );
            document.setContent( &qtfile );
        }

        // create top element
        QDomElement top = document.createElement( Private::Xml::ConnectionAttributesList );

        auto connectionWidgets( findChildren<Private::LocalConnectionWidget*>() );
        for( const auto& widget:connectionWidgets )
        {
            // get relevant information from widget
            auto attributes( widget->attributes() );
            auto remotePath( widget->remotePath() );

            top.appendChild( Private::XmlConnectionAttributes( attributes, remotePath ).domElement( document ) );

        }

        document.replaceChild( top );
        QFile qfile( file );
        if( !qfile.open( QIODevice::WriteOnly ) ) return;
        qfile.write( document.toByteArray() );

    }

    //________________________________________________________________
    void SynchronizationDialog::closeEvent( QCloseEvent *event )
    {
        Debug::Throw( "SynchronizationDialog::closeEvent.\n" );
        event->accept();
        writeConnectionAttributes();
    }

    //________________________________________________________
    void SynchronizationDialog::_addConnection( void )
    {

        Private::EditConnectionNameDialog dialog( this );
        dialog.setConnectionName( tr( "New Connection" ) );
        if( !dialog.exec() ) return;

        // generate dummy connection attributes
        ConnectionAttributes attributes;
        attributes.setName( dialog.connectionName() );
        attributes.setHost( "localhost" );
        attributes.setPort( 22 );
        attributes.setUserName( Util::user() );
        attributes.setRememberPassword( true );

        File remotePath = File( ".config/homechat" ).addPath( Util::home() );

        addConnection( attributes, remotePath );

    }

    //________________________________________________________
    bool SynchronizationDialog::_synchronize( void )
    {

        class AbortLocker
        {
            public:

            //* constructor
            AbortLocker( SynchronizationDialog* parent ):
                parent_( parent )
            {
                parent_->abortButton_->setEnabled( true );
                parent_->aborted_ = false;

                parent_->synchronizeButton_->setEnabled( false );
                parent_->closeButton_->setEnabled( false );
            }

            //* destructor
            ~AbortLocker( void )
            {
                parent_->abortButton_->setEnabled( false );
                parent_->aborted_ = false;

                parent_->synchronizeButton_->setEnabled( true );
                parent_->closeButton_->setEnabled( true );
            }

            private:

            //* parent
            SynchronizationDialog* parent_;

        };

        AbortLocker locker(this);

        // copy all messages from remote and merge
        ChatMessageSet allMessages;
        auto connectionWidgets( findChildren<Private::LocalConnectionWidget*>() );
        for( const auto& widget:connectionWidgets )
        {
            // get relevant information from widget
            auto attributes( widget->attributes() );
            auto remotePath( widget->remotePath() );
            auto name( attributes.name() );

            // generate local path
            File localPath = File( QString( "homechat_%1" ).arg( name ) ).addPath( Util::tmp() );

            // get messages from remote connection
            auto messagePair( _copyFromRemote( attributes, localPath, remotePath ) );
            auto success = messagePair.second;
            if( !success ) return false;

            // merge to full list
            auto messages = messagePair.first.messages();
            for( const auto& message:messages )
            { allMessages.append( message ); }

            // check abortion flag
            if( aborted_ ) return false;

        }

        // write all merged messages to temporary path
        File localPath = File( "homechat_merged" ).addPath( Util::tmp() );

        // create local directory
        localPath.removeRecursive();
        if( !localPath.createDirectory() )
        {

            QString buffer( tr( "Failed to create directory %1" ).arg( localPath ) );
            logWidget_->append( buffer+"\n" );

            synchronizeButton_->setEnabled( true );
            closeButton_->setEnabled( true );
            return false;

        } else {

            QString buffer( tr( "Merging all messages to %1" ).arg( localPath ) );
            logWidget_->append( buffer+"\n" );

        }

        const bool useCompression = XmlOptions::get().get<bool>( "COMPRESS_HISTORY" );
        const File dbFile = File( XmlOptions::get().raw( "HISTORY_FILE" ) ).localName();
        allMessages.setUseCompression( useCompression );
        allMessages.setFilename( dbFile.addPath( localPath ) );
        allMessages.write();
        allMessages.mergeMessages();

        // copy merged messages remotely
        for( const auto& widget:connectionWidgets )
        {
            // get relevant information from widget
            auto attributes( widget->attributes() );
            auto remotePath( widget->remotePath() );
            auto name( attributes.name() );

            bool success = _copyToRemote( attributes, localPath, remotePath );
            if( !success || aborted_ ) return false;

        }

        logWidget_->append( "Done\n" );

        return true;

    }

    //__________________________________________
    void SynchronizationDialog::_abort( void )
    {
        Debug::Throw( "SynchronizationDialog::_abort.\n" );
        aborted_ = true;
        logWidget_->append( "Aborted\n" );
    }

    //__________________________________________
    void SynchronizationDialog::_append( QString value )
    {  logWidget_->append( value + "\n" ); }

    //__________________________________________
    void SynchronizationDialog::_appendTransfer( qint64 size, qint64 transferred )
    { progressBar_->setValue( 100*transferred/size ); }

    //__________________________________________
    void SynchronizationDialog::_connectionWidgetDestroyed( void )
    { synchronizeButton_->setEnabled( findChildren<ConnectionWidget*>().size() > 2 ); }

    //__________________________________________
    void SynchronizationDialog::_updateSynchronizationButton( void )
    { synchronizeButton_->setEnabled( findChildren<ConnectionWidget*>().size() >= 2 ); }

    //__________________________________________
    bool SynchronizationDialog::_copyToLocal( Connection& connection, File remoteFile, File localPath )
    {

        // try connect to remote for reading
        FileTransferObject object( this, remoteFile );
        // connect( &object, SIGNAL(debug(QString)), SLOT(_append(QString)) );
        connect( &object, SIGNAL(error(QString)), SLOT(_append(QString)) );
        connect( &object, SIGNAL(transferred(qint64,qint64)), SLOT(_appendTransfer(qint64,qint64)) );

        QString buffer( tr( "Transferring %1" ).arg( remoteFile ) );
        logWidget_->append( buffer + "\n" );

        object.connect( connection.session(), QIODevice::ReadOnly );
        const bool connected = object.waitForConnected( 5000 );
        if( !connected )
        {
            Debug::Throw(0) << "Remote file: " << remoteFile << " not found." << endl;
            return false;
        }

        // read file from remote and stored into local file
        const File localFile = remoteFile.localName().addPath( localPath );
        object.fromRemote( connection.session(), localFile );
        const bool transferred = object.waitForCompleted( 5000 );
        if( !transferred )
        {
            Debug::Throw(0) << "Transfer failed" << endl;
            return false;
        }

        return true;
    }

    //__________________________________________
    bool SynchronizationDialog::_copyToRemote( Connection& connection, File localFile, File remotePath )
    {

        // try connect to remote for reading
        File remoteFile( localFile.localName().addPath( remotePath ) );
        FileTransferObject object( this, remoteFile );
        // connect( &object, SIGNAL(debug(QString)), SLOT(_append(QString)) );
        connect( &object, SIGNAL(error(QString)), SLOT(_append(QString)) );
        connect( &object, SIGNAL(transferred(qint64,qint64)), SLOT(_appendTransfer(qint64,qint64)) );

        QString buffer( tr( "Transferring %1" ).arg( localFile ) );
        logWidget_->append( buffer + "\n" );

        object.setFileSize( localFile.fileSize() );
        object.connect( connection.session(), QIODevice::WriteOnly );
        const bool connected = object.waitForConnected( 5000 );
        if( !connected )
        {
            Debug::Throw(0) << "Cannot write to remote file " << remoteFile << endl;
            return false;
        }

        // read file from remote and stored into local file
        object.toRemote( connection.session(), localFile );
        const bool transferred = object.waitForCompleted( 5000 );
        if( !transferred )
        {
            Debug::Throw(0) << "Transfer failed" << endl;
            return false;
        }

        return true;
    }

    //__________________________________________
    SynchronizationDialog::ChatMessagePair SynchronizationDialog::_copyFromRemote( ConnectionAttributes attributes, File localPath, File remotePath )
    {

        ChatMessageSet messages;
        const bool useCompression = XmlOptions::get().get<bool>( "COMPRESS_HISTORY" );
        messages.setUseCompression( useCompression );
        messages.setDebugLevel(0);

        // generate remote filenames and read
        const File dbFile = File( XmlOptions::get().raw( "HISTORY_FILE" ) ).localName();

        Connection connection( this );

        // connect to remote host
        QString buffer( tr( "connecting to %1@%2" ).arg( attributes.userName() ).arg( attributes.host() ) );
        logWidget_->append( buffer+"\n" );

        connection.setAttributes( attributes );
        connection.createTunnels();
        connection.connect();
        connection.authenticate();
        const bool success = connection.waitForConnected( 5000 );
        if( success )
        {

            logWidget_->append(  tr( "Connected\n" ) );

        } else {

            logWidget_->append(  tr( "Connection failed\n" ) );
            return std::make_pair( messages,false );

        }

        // create local directory
        localPath.removeRecursive();
        if( !localPath.createDirectory() )
        {
            QString buffer( tr( "Failed to create directory %1" ).arg( localPath ) );
            logWidget_->append( buffer+"\n" );
            return std::make_pair( messages,false );
        }

        // generate list of files to be transferred

        File::List files;
        for( QDate date = QDate::currentDate(); date.year() >= 2015; date = date.addMonths( -1 ) )
        { files.append(  File( dbFile + date.toString( "-yyyy-MM" ) ).addPath( remotePath ) ); }

        files.append( File( dbFile + (".new" ) ).addPath( remotePath ) );
        files.append( File( dbFile ).addPath( remotePath ) );

        for( const auto& file:files )
        {
            _copyToLocal( connection, file, localPath );
            if( aborted_ ) return std::make_pair( messages,false );
        }

        messages.setFilename( dbFile.addPath( localPath ) );
        messages.read();

        return std::make_pair( messages, true );

    }

    //__________________________________________
    bool SynchronizationDialog::_copyToRemote( ConnectionAttributes attributes, File localPath, File remotePath )
    {

        Ssh::Connection connection( this );

        QString buffer( tr( "connecting to %1@%2" ).arg( attributes.userName() ).arg( attributes.host() ) );
        logWidget_->append( buffer+"\n" );

        // connect to remote host
        connection.setAttributes( attributes );
        connection.createTunnels();
        connection.connect();
        connection.authenticate();
        const bool success = connection.waitForConnected( 5000 );
        if( success )
        {

            logWidget_->append(  tr( "Connected\n" ) );

        } else {

            logWidget_->append(  tr( "Connection failed\n" ) );
            return false;

        }

        // generate remote filenames and read
        const File dbFile = File( XmlOptions::get().raw( "HISTORY_FILE" ) ).localName();
        File::List files;
        for( QDate date = QDate::currentDate(); date.year() >= 2015; date = date.addMonths( -1 ) )
        {
            File file = File(  dbFile + date.toString( "-yyyy-MM" ) ).addPath( localPath );
            if( file.exists() ) files.append( file );
        }

        File file = File(  dbFile + ".new" ).addPath( localPath );
        if( file.exists() ) files.append( file );

        file = dbFile.addPath( localPath );
        if( file.exists() ) files.append( file );

        //
        for( const auto& file:files )
        {
            _copyToRemote( connection, file, remotePath );
            if( aborted_ ) return false;
        }

        return true;
    }

}

#include "SshSynchronizationDialog.moc"
