/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "SystemTrayIcon.h"

#include "ClientApplication.h"
#include "IconEngine.h"
#include "IconSize.h"
#include "Singleton.h"
#include "SystemTrayMenu.h"
#include "SystemTrayToolTipWidget.h"

#include <QApplication>
#include <QDesktopWidget>

//______________________________________________________
SystemTrayIcon::SystemTrayIcon( QObject* parent ):
    QSystemTrayIcon( parent ),
    Counter( "SystemTrayIcon" )
{

    Debug::Throw( "SystemTrayIcon::SystemTrayIcon.\n" );

    // store icons
    connectedIcon_ = IconEngine::get( "homechat-systray-connected" );
    disconnectedIcon_ = IconEngine::get( "homechat-systray-disconnected" );
    pendingMessagesIcon_ = IconEngine::get( "homechat-systray-attention" );

    // use application icon
    setIcon( disconnectedIcon_ );

    // context menu
    contextMenu_ = new SystemTrayMenu();
    setContextMenu( contextMenu_ );

    // tooltip widget
    toolTipWidget_ = new SystemTrayToolTipWidget( nullptr );
    toolTipWidget_->setPreferredPosition( SystemTrayToolTipWidget::Top );

    connect( this, SIGNAL(activated(QSystemTrayIcon::ActivationReason)), this, SLOT(_activated(QSystemTrayIcon::ActivationReason)) );

}

//______________________________________________________
SystemTrayIcon::~SystemTrayIcon( void )
{
    contextMenu_->deleteLater();
    toolTipWidget_->deleteLater();
}

//______________________________________________________
void SystemTrayIcon::updatePendingMessages( bool hasPendingMessages )
{
    setIcon( hasPendingMessages ? pendingMessagesIcon_:connectedIcon_ );
    if( hasPendingMessages ) messageCount_++;
    else messageCount_ = 0;

    _updateToolTip();

}

//______________________________________________________
void SystemTrayIcon::setConnectionId( ConnectionId connectionId )
{

    // update and re-add
    connectionId_ = connectionId;

    // update list
    const int i = contacts_.indexOf( connectionId );
    if( i >= 0 ) contacts_.replace( i, connectionId );
    else contacts_.append( connectionId );

}

//______________________________________________________
void SystemTrayIcon::setContacts( const ConnectionId::List& contacts )
{
    contacts_ = contacts;
    if( connectionId_.isValid() )
    {
        const int i = contacts_.indexOf( connectionId_ );
        if( i >= 0 ) contacts_.replace( i, connectionId_ );
        else contacts_.append( connectionId_ );
    }
}

//_________________________________________________
void SystemTrayIcon::updateContact( ConnectionId oldConnectionId, ConnectionId connectionId )
{

    Debug::Throw() << "SystemTrayIcon::updateContact -"
        << " old: " << oldConnectionId
        << " new: " << connectionId
        << endl;

    // first update old contact from model
    // and update model
    const int i = contacts_.indexOf( oldConnectionId );
    if( i >= 0 ) oldConnectionId = contacts_.at(i);

    // store whether userState or connection state has Changed
    const bool connectionStateChanged( oldConnectionId.connectionState() != connectionId.connectionState() );

    // update internal list
    if( i>= 0 ) contacts_.replace( i, connectionId );
    else contacts_.append( connectionId );

    if( oldConnectionId == connectionId_ )
    {
        connectionId_ = connectionId;

        // update icon
        if( connectionStateChanged )
        {
            switch( connectionId.connectionState() )
            {
                case ConnectionId::Disconnected:
                {
                    // update icon
                    setIcon( disconnectedIcon_ );

                    // also mark all other clients as offline
                    for( int i = 0; i < contacts_.size(); ++i )
                    {

                        if( contacts_[i] != connectionId_ )
                        { contacts_[i].setConnectionState( ConnectionId::Disconnected ); }

                    }

                    break;
                }

                case ConnectionId::AwaitingReply:
                setIcon( disconnectedIcon_ );
                break;

                default:
                case ConnectionId::Connected:
                setIcon( connectedIcon_ );
                break;

            }

        }

    }

    // update tooltip
    _updateToolTip();

}

//_________________________________________________
void SystemTrayIcon::removeContact( ConnectionId contact )
{
    Debug::Throw( "SystemTrayIcon::removeContact.\n" );
    contacts_.removeOne( contact );
}

//_______________________________________________
bool SystemTrayIcon::event(QEvent *event)
{

    if (event->type() == QEvent::ToolTip )
    {
        const int offset( 10 );

        _updateToolTip();
        toolTipWidget_->setIndexRect( geometry().adjusted( -offset, -offset, offset, offset ) );
        toolTipWidget_->adjustSize();
        toolTipWidget_->show();
    }

    return QSystemTrayIcon::event(event);

}

//______________________________________________________
void SystemTrayIcon::_activated( QSystemTrayIcon::ActivationReason reason )
{
    switch( reason )
    {

        default: break;
        case Context: _showContextMenu(); break;
    }

}

//__________________________________________________________________
void SystemTrayIcon::_updateToolTip( void )
{
    Debug::Throw( "SystemTrayIcon::_updateToolTip.\n" );

    toolTipWidget_->setPixmap(
        static_cast<const ClientApplication*>( Singleton::get().application() )->applicationIcon().pixmap( IconSize( IconSize::Maximum ) ) );

    toolTipWidget_->setTitle( tr( "Homechat" ) );
    toolTipWidget_->updateText( contacts_, messageCount_ );

    return;

}

//______________________________________________________
void SystemTrayIcon::_showContextMenu( void ) const
{

    Debug::Throw( "SystemTrayIcon::_showContextMenu.\n" );

    // icon geometry
    QRect geometry( this->geometry() );
    if( !geometry.isValid() ) geometry = QRect( QCursor::pos(), QSize(0,0) );

    // menu size
    contextMenu_->adjustSize();
    const QSize size( contextMenu_->size() );

    // desktop size
    QDesktopWidget* desktop( qApp->desktop() );
    QRect desktopGeometry( desktop->availableGeometry( desktop->screenNumber( contextMenu_ ) ) );

    // detect the side of the systray icon by comparing geometry and desktop geometry
    enum DesktopSide
    {
        Left,
        Top,
        Right,
        Bottom
    };

    DesktopSide desktopSide( Bottom );
    if( geometry.bottom() < desktopGeometry.top() ) desktopSide = Top;
    else if( geometry.right() < desktopGeometry.left() ) desktopSide = Left;
    else if( geometry.left() > desktopGeometry.right() ) desktopSide = Right;

    // define menu geometry
    QPoint menuPosition;
    switch( desktopSide )
    {
        case Left:
        {
            menuPosition.setX( desktopGeometry.left() );
            menuPosition.setY( qMin( geometry.top(), desktopGeometry.bottom() - size.height() ) );
            break;
        }

        case Top:
        {
            menuPosition.setY( desktopGeometry.top() );
            menuPosition.setX( qMin( geometry.left(), desktopGeometry.right() - size.width() ) );
            break;
        }

        case Right:
        {
            menuPosition.setX( desktopGeometry.right() - size.width() );
            menuPosition.setY( qMin( geometry.top(), desktopGeometry.bottom() - size.height() ) );
            break;
        }

        default:
        case Bottom:
        {
            menuPosition.setY( desktopGeometry.bottom() - size.height() );
            menuPosition.setX( qMin( geometry.left(), desktopGeometry.right() - size.width() ) );
            break;
        }

    }

    // move menu
    contextMenu_->exec( menuPosition );

}
