/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ChatMessageListDialog.h"
#include "ChatMessageSet.h"
#include "ChatMessageMigration.h"
#include "Debug.h"
#include "DefaultOptions.h"
#include "PixmapEngine.h"
#include "ResourceMigration.h"
#include "ServerApplication.h"
#include "ServerConfigurationDialog.h"
#include "Singleton.h"
#include "Util.h"

#include <QApplication>
#include <QCoreApplication>
#include <QString>

//__________________________________________
//* main function
int main (int argc, char *argv[])
{
    // options
    installDefaultOptions();

    // non recordable options
    XmlOptions::get().set( "OLD_RC_FILE", Option( File(".homechat-serverrc").addPath(Util::home()), Option::None ), true );
    XmlOptions::get().set( "RC_FILE", Option( File("homechat-serverrc").addPath(Util::config()), Option::None ), true );

    // migrate old rc files
    File oldRCFile( XmlOptions::get().raw( "OLD_RC_FILE" ) );
    File rcFile( XmlOptions::get().raw( "RC_FILE" ) );
    ResourceMigration( oldRCFile ).migrate( rcFile );

    // assign and read
    XmlOptions::setFile( rcFile );
    XmlOptions::read();

    // migrate hitory
    File oldHistoryFile( XmlOptions::get().raw( "OLD_HISTORY_FILE" ) );
    File historyFile( XmlOptions::get().raw( "HISTORY_FILE" ) );
    ChatMessageMigration( oldHistoryFile ).migrate( historyFile );

    // debug level
    Debug::setLevel( XmlOptions::get().get<int>("DEBUG_LEVEL") );
    if( Debug::level() ) Debug::Throw() << XmlOptions::get() << endl;

    // singleton application is deleted before QApplication
    const CommandLineArguments arguments( argc, argv );
    ServerApplication singleton( arguments );
    Singleton::get().setApplication( &singleton );

    const CommandLineParser parser( singleton.commandLineParser( arguments ) );
    if( parser.hasFlag( "--configure" ) )
    {

        QTextStream( stdout ) << "running in configuration mode" << endl;

        // resources
        Q_INIT_RESOURCE( basePixmaps );
        Q_INIT_RESOURCE( pixmaps );

        // QCore application
        QApplication application( argc, argv );

        ServerConfigurationDialog dialog( nullptr );
        dialog.show();

        application.exec();
        return 0;

    } else if( parser.hasFlag( "--message-list" ) ) {

        QTextStream( stdout ) << "running in message list edition mode" << endl;

        // resources
        Q_INIT_RESOURCE( basePixmaps );
        Q_INIT_RESOURCE( pixmaps );

        // QCore application
        QApplication application( argc, argv );

        const int debugLevel( parser.hasFlag( "--verbose" ) ? 0:1 );
        ChatMessageListDialog dialog( nullptr, debugLevel );

        QIcon icon;
        icon.addPixmap( PixmapEngine::get( ":/homechat.png" ) );
        icon.addPixmap( PixmapEngine::get( ":/homechat-large.png" ) );
        dialog.setWindowIcon( icon );

        dialog.show();

        application.exec();
        return 0;

    } else if( parser.hasFlag( "--uncompress" ) ) {

        QTextStream( stdout ) << "uncompressing message list" << endl;

        ChatMessageSet messages;
        messages.setDebugLevel( parser.hasFlag( "--verbose" ) ? 0:1 );
        messages.setFilename( File( XmlOptions::get().raw("HISTORY_FILE") ) );
        messages.setUseCompression( XmlOptions::get().get<bool>("COMPRESS_HISTORY" ) );
        messages.uncompress();

        XmlOptions::get().set<bool>("COMPRESS_HISTORY", false );


    } else if( parser.hasFlag( "--compress" ) ) {

        QTextStream( stdout ) << "compressing message list" << endl;

        ChatMessageSet messages;
        messages.setDebugLevel( parser.hasFlag( "--verbose" ) ? 0:1 );
        messages.setFilename( File( XmlOptions::get().raw("HISTORY_FILE") ) );
        messages.setUseCompression( XmlOptions::get().get<bool>("COMPRESS_HISTORY" ) );
        messages.compress();

        XmlOptions::get().set<bool>("COMPRESS_HISTORY", true );

    } else {

        // QCore application
        QCoreApplication application( argc, argv );

        // initialize and run
        if( singleton.initApplicationManager() )
        {  application.exec(); }

    }

    return 0;

}
